@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	20 December 1989
- * Last Edited:	18 November 1993
+ * Last Edited:	15 August 1994
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -46,7 +46,6 @@
 
 #include <stdio.h>
 #include <ctype.h>
-#include <netdb.h>
 #include <errno.h>
 extern int errno;		/* just in case */
 #include <signal.h>
@@ -104,8 +103,9 @@ DRIVER bezerkdriver = {
 
 				/* prototype stream */
 MAILSTREAM bezerkproto = {&bezerkdriver};
-				/* standard driver's prototype */
-MAILSTREAM *mailstd_proto = &bezerkproto;
+
+				/* driver parameters */
+static long bezerk_fromwidget = T;
 
 /* Berkeley mail validate mailbox
  * Accepts: mailbox name
@@ -126,29 +126,24 @@ DRIVER *bezerk_valid (char *name)
 
 int bezerk_isvalid (char *name,char *tmp)
 {
-  int i,fd,ti,zn;
+  int i,fd,zn;
   int ret = NIL;
-  char *s = tmp,*t;
+  char *s = tmp,*t,file[MAILTMPLEN];
   struct stat sbuf;
-  struct hostent *host_name;
-  if (!lhostn) {		/* have local host yet? */
-    gethostname(tmp,MAILTMPLEN);/* get local host name */
-    lhostn = cpystr ((host_name = gethostbyname (tmp)) ?
-		     host_name->h_name : tmp);
-  }
-				/* INBOX is always accepted */
-  if (!strcmp (ucase (strcpy (tmp,name)),"INBOX")) return T;
-				/* if file, get its status */
+  time_t tp[2];
+  errno = EINVAL;		/* assume invalid argument */
+				/* must be non-empty file */
   if ((*name != '{') && !((*name == '*') && (name[1] == '{')) &&
-      (stat (bezerk_file (tmp,name),&sbuf) == 0) &&
-      ((fd = open (tmp,O_RDONLY,NIL)) >= 0)) {
-    if (sbuf.st_size == 0) { 	/* allow empty file if not .TxT */
-      if ((i = strlen (tmp)) < 4 || strcmp (tmp + i - 4 ,".TxT"))
-	return LONGT;
-    }
-    else if ((read (fd,tmp,MAILTMPLEN-1) >= 0) &&
-	     (tmp[0] == 'F') && VALID (s,t,ti,zn)) ret = T;
+      (!stat (dummy_file (file,name),&sbuf)) &&
+      ((fd = open (file,O_RDONLY,NIL)) >= 0)) {
+    if ((sbuf.st_size > 0) && (read (fd,tmp,MAILTMPLEN-1) >= 0))
+      VALID (s,t,ret,zn);
+    tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
+    tp[1] = sbuf.st_mtime;
+    utime (file,tp);		/* set the times */
     close (fd);			/* close the file */
+				/* no error if file empty */
+    if (!(ret || sbuf.st_size)) errno = 0;
   }
   return ret;			/* return what we should */
 }
@@ -161,8 +156,18 @@ int bezerk_isvalid (char *name,char *tmp)
 
 void *bezerk_parameters (long function,void *value)
 {
-  fatal ("Invalid bezerk_parameters function");
-  return NIL;
+  switch ((int) function) {
+  case SET_FROMWIDGET:
+    bezerk_fromwidget = (long) value;
+    break;
+  case GET_FROMWIDGET:
+    value = (void *) bezerk_fromwidget;
+    break;
+  default:
+    value = NIL;		/* error case */
+    break;
+  }
+  return value;
 }
 
 /* Berkeley mail find list of mailboxes
@@ -217,7 +222,7 @@ void bezerk_find_all_bboards (MAILSTREAM *stream,char *pat)
 long bezerk_subscribe (MAILSTREAM *stream,char *mailbox)
 {
   char tmp[MAILTMPLEN];
-  return sm_subscribe (bezerk_file (tmp,mailbox));
+  return sm_subscribe (dummy_file (tmp,mailbox));
 }
 
 
@@ -230,7 +235,7 @@ long bezerk_subscribe (MAILSTREAM *stream,char *mailbox)
 long bezerk_unsubscribe (MAILSTREAM *stream,char *mailbox)
 {
   char tmp[MAILTMPLEN];
-  return sm_unsubscribe (bezerk_file (tmp,mailbox));
+  return sm_unsubscribe (dummy_file (tmp,mailbox));
 }
 
 
@@ -265,20 +270,7 @@ long bezerk_unsubscribe_bboard (MAILSTREAM *stream,char *mailbox)
 
 long bezerk_create (MAILSTREAM *stream,char *mailbox)
 {
-  char tmp[MAILTMPLEN];
-  int i,fd;
-				/* must not be a ".TxT" file */
-  if ((i = strlen (mailbox)) > 4 && !strcmp (mailbox + i - 4 ,".TxT")) {
-    mm_log ("Can't create mailbox: name must not end with .TxT",ERROR);
-    return NIL;
-  }
-  if ((fd = open (bezerk_file (tmp,mailbox),O_WRONLY|O_CREAT|O_EXCL,0600))<0) {
-    sprintf (tmp,"Can't create mailbox %s: %s",mailbox,strerror (errno));
-    mm_log (tmp,ERROR);
-    return NIL;
-  }
-  close (fd);			/* close the file */
-  return T;			/* return success */
+  return dummy_create (stream,mailbox);
 }
 
 
@@ -306,8 +298,8 @@ long bezerk_rename (MAILSTREAM *stream,char *old,char *new)
   char tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN],lockx[MAILTMPLEN];
   int fd,ld;
 				/* get the c-client lock */
-  if ((ld = open (lockname (lock,bezerk_file (file,old)),O_RDWR|O_CREAT,0666))
-      < 0) {
+  if ((ld = open (lockname (lock,dummy_file (file,old)),O_RDWR|O_CREAT,
+		  (int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL))) < 0) {
     sprintf (tmp,"Can't get lock for mailbox %s: %s",old,strerror (errno));
     mm_log (tmp,ERROR);
     return NIL;
@@ -320,13 +312,13 @@ long bezerk_rename (MAILSTREAM *stream,char *old,char *new)
     return NIL;
   }
 				/* lock out non c-client applications */
-  if ((fd = bezerk_lock (file,O_RDONLY,S_IREAD|S_IWRITE,lockx,LOCK_EX)) < 0) {
+  if ((fd = bezerk_lock (file,O_RDWR,S_IREAD|S_IWRITE,lockx,LOCK_EX)) < 0) {
     sprintf (tmp,"Can't lock mailbox %s: %s",old,strerror (errno));
     mm_log (tmp,ERROR);
     return NIL;
   }
 				/* do the rename or delete operation */
-  if (new ? rename (file,bezerk_file (tmp,new)) : unlink (file)) {
+  if (new ? rename (file,dummy_file (tmp,new)) : unlink (file)) {
     sprintf (tmp,"Can't %s mailbox %s: %s",new ? "rename" : "delete",old,
 	     strerror (errno));
     mm_log (tmp,ERROR);
@@ -361,7 +353,7 @@ MAILSTREAM *bezerk_open (MAILSTREAM *stream)
   }
   stream->local = fs_get (sizeof (BEZERKLOCAL));
 				/* canonicalize the stream mailbox name */
-  bezerk_file (tmp,stream->mailbox);
+  dummy_file (tmp,stream->mailbox);
 				/* force readonly if bboard */
   if (*stream->mailbox == '*') stream->readonly = T;
   else {			/* canonicalize name */
@@ -392,15 +384,16 @@ MAILSTREAM *bezerk_open (MAILSTREAM *stream)
 				/* make lock for read/write access */
   if (!stream->readonly) while (retry) {
 				/* get a new file handle each time */
-    if ((fd = open (LOCAL->lname,O_RDWR|O_CREAT,0666)) < 0)
+    if ((fd = open (LOCAL->lname,O_RDWR|O_CREAT,
+		    (int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL))) < 0)
       mm_log ("Can't open mailbox lock, access is readonly",WARN);
 				/* can get the lock? */
     if (flock (fd,LOCK_EX|LOCK_NB)) {
       if (retry-- == KODRETRY) {/* no, first time through? */
 				/* yes, get other process' PID */
 	if (!fstat (fd,&sbuf) && (i = min (sbuf.st_size,MAILTMPLEN)) &&
-	    (read (fd,tmp,i) == i) && (i = atol (tmp))) {
-	  kill (i,SIGUSR2);	/* send the Kiss Of Death */
+	    (read (fd,tmp,i) == i) && !(tmp[i] = 0) && (i = atol (tmp))) {
+	  kill ((int) i,SIGUSR2);
 	  sprintf (tmp,"Trying to get mailbox lock from process %ld",i);
 	  mm_log (tmp,WARN);
 	}
@@ -415,7 +408,8 @@ MAILSTREAM *bezerk_open (MAILSTREAM *stream)
     }
     else {			/* got the lock, nobody else can alter state */
       LOCAL->ld = fd;		/* note lock's fd */
-      chmod (LOCAL->lname,0666);/* make sure mode OK (don't use fchmod()) */
+				/* make sure mode OK (don't use fchmod()) */
+      chmod (LOCAL->lname,(int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL));
       if (stream->silent) i = 0;/* silent streams won't accept KOD */
       else {			/* note our PID in the lock */
 	sprintf (tmp,"%d",getpid ());
@@ -539,8 +533,8 @@ ENVELOPE *bezerk_fetchstructure (MAILSTREAM *stream,long msgno,BODY **body)
     }
     INIT (&bs,mail_string,(void *) m->body,m->bodysize);
 				/* parse envelope and body */
-    rfc822_parse_msg (env,body ? b : NIL,m->header,m->headersize,&bs,lhostn,
-		      LOCAL->buf);
+    rfc822_parse_msg (env,body ? b : NIL,m->header,m->headersize,&bs,
+		      mylocalhost (),LOCAL->buf);
   }
   if (body) *body = *b;		/* return the body */
   return *env;			/* return the envelope */
@@ -958,50 +952,105 @@ long bezerk_move (MAILSTREAM *stream,char *sequence,char *mailbox)
 /* Berkeley mail append message from stringstruct
  * Accepts: MAIL stream
  *	    destination mailbox
+ *	    initial flags
+ *	    internal date
  *	    stringstruct of messages to append
  * Returns: T if append successful, else NIL
  */
 
 #define BUFLEN 8*MAILTMPLEN
 
-long bezerk_append (MAILSTREAM *stream,char *mailbox,STRING *message)
+long bezerk_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
+		    STRING *message)
 {
   struct stat sbuf;
-  int fd,i;
-  char buf[BUFLEN],lock[MAILTMPLEN];
+  int i,fd,ti,zn;
+  char *s,*x,buf[BUFLEN],file[MAILTMPLEN],lock[MAILTMPLEN];
+  time_t tp[2];
+  MESSAGECACHE elt;
   char c = '\n';
-  long ok = T;
-  unsigned long t = time (0);
-  long size = SIZE (message);
-  if ((fd = bezerk_lock (bezerk_file (buf,mailbox),O_WRONLY|O_APPEND|O_CREAT,
+  long j,n,ok = T;
+  time_t t = time (0);
+  unsigned long size = SIZE (message);
+  short f = bezerk_getflags (stream,flags);
+				/* parse date if given */
+  if (date && !mail_parse_date (&elt,date)) {
+    mm_log ("Bad date in append",ERROR);
+    return NIL;
+  }
+				/* make sure valid mailbox */
+  if (!bezerk_isvalid (mailbox,buf)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    if (strcmp (ucase (strcpy (buf,mailbox)),"INBOX")) {
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+      return NIL;
+    }
+    else break;
+  case 0:			/* merely empty file? */
+    break;
+  case EINVAL:
+    sprintf (buf,"Invalid Berkeley-format mailbox name: %s",mailbox);
+    mm_log (buf,ERROR);
+    return NIL;
+  default:
+    sprintf (buf,"Not a Berkeley-format mailbox: %s",mailbox);
+    mm_log (buf,ERROR);
+    return NIL;
+  }
+  if ((fd = bezerk_lock (dummy_file (file,mailbox),O_WRONLY|O_APPEND|O_CREAT,
 			 S_IREAD|S_IWRITE,lock,LOCK_EX)) < 0) {
     sprintf (buf,"Can't open append mailbox: %s",strerror (errno));
     mm_log (buf,ERROR);
     return NIL;
   }
+
   mm_critical (stream);		/* go critical */
   fstat (fd,&sbuf);		/* get current file size */
-  sprintf (buf,"From %s@%s %s",myusername (),lhostn,ctime (&t));
+  sprintf (buf,"From %s@%s ",myusername (),mylocalhost ());
+				/* write the date given */
+  if (date) mail_cdate (buf + strlen (buf),&elt);
+  else strcat (buf,ctime (&t));	/* otherwise write the time now */
+  sprintf (buf + strlen (buf),"Status: %sO\nX-Status: %s%s%s\n",
+	   f&fSEEN ? "R" : "",f&fDELETED ? "D" : "",
+	   f&fFLAGGED ? "F" : "",f&fANSWERED ? "A" : "");
   i = strlen (buf);		/* initial buffer space used */
   while (ok && size--) {	/* copy text, tossing out CR's */
-    if (c == '\n') {		/* if at start of line */
-      t = GETPOS (message);	/* prepend a broket if needed */
+				/* if at start of line */
+    if ((c == '\n') && (size > 5)) {
+      n = GETPOS (message);	/* prepend a broket if needed */
       if ((SNX (message) == 'F') && (SNX (message) == 'r') &&
 	  (SNX (message) == 'o') && (SNX (message) == 'm') &&
-	  (SNX (message) == ' ')) ok = bezerk_append_putc (fd,buf,&i,'>');
-      SETPOS (message,t);	/* restore position as needed */
+	  (SNX (message) == ' ')) {
+				/* always write widget if unconditional */
+	if (bezerk_fromwidget) ok = bezerk_append_putc (fd,buf,&i,'>');
+	else {			/* hairier test, count length of this line */
+	  for (j = 6; (j < size) && (SNX (message) != '\n'); j++);
+	  if (j < size) {	/* copy line */
+	    SETPOS (message,n);	/* restore position */
+	    x = s = (char *) fs_get (j + 1);
+	    while (j--) if ((c = SNX (message)) != '\015') *x++ = c;
+	    *x = '\0';		/* tie off line */
+	    VALID (s,x,ti,zn);	/* see if looks like need a widget */
+	    if (ti) ok = bezerk_append_putc (fd,buf,&i,'>');
+	    fs_give ((void **) &s);
+	  }
+	}
+      }
+      SETPOS (message,n);	/* restore position as needed */
     }
 				/* copy another character */
     if ((c = SNX (message)) != '\015') ok = bezerk_append_putc (fd,buf,&i,c);
   }
 				/* write trailing newline */
   if (ok) ok = bezerk_append_putc (fd,buf,&i,'\n');
-  if (!(ok && (ok = (write (fd,buf,i) >= 0)))) {
+  if (!(ok && (ok = (write (fd,buf,i) >= 0)) && (ok = !fsync (fd)))) {
     sprintf (buf,"Message append failed: %s",strerror (errno));
     mm_log (buf,ERROR);
     ftruncate (fd,sbuf.st_size);
   }
-  fsync (fd);			/* force out the update */
+  tp[0] = sbuf.st_atime;	/* preserve atime */
+  tp[1] = time (0);		/* set mtime to now */
+  utime (file,tp);		/* set the times */
   bezerk_unlock (fd,NIL,lock);	/* unlock and close mailbox */
   mm_nocritical (stream);	/* release critical */
   return ok;			/* return success */
@@ -1065,43 +1114,6 @@ void bezerk_abort (MAILSTREAM *stream)
   }
 }
 
-/* Berkeley mail generate file string
- * Accepts: temporary buffer to write into
- *	    mailbox name string
- * Returns: local file string
- */
-
-
-char *bezerk_file (char *dst,char *name)
-{
-  struct passwd *pw;
-  char *s,*t,tmp[MAILTMPLEN];
-  switch (*name) {
-  case '*':			/* bboard? */
-    sprintf (tmp,"~ftp/%s",(name[1] == '/') ? name+2 : name+1);
-    dst = bezerk_file (dst,tmp);/* recurse to get result */
-    break;
-  case '/':			/* absolute file path */
-    strcpy (dst,name);		/* copy the mailbox name */
-    break;
-  case '~':			/* home directory */
-    if (name[1] == '/') t = myhomedir ();
-    else {
-      strcpy (tmp,name + 1);	/* copy user name */
-      if (s = strchr (tmp,'/')) *s = '\0';
-      t = ((pw = getpwnam (tmp)) && pw->pw_dir) ? pw->pw_dir : "/NOSUCHUSER";
-    }
-    sprintf (dst,"%s%s",t,(s = strchr (name,'/')) ? s : "");
-    break;
-  default:			/* other name */
-    if (strcmp (ucase (strcpy (dst,name)),"INBOX"))
-      sprintf (dst,"%s/%s",myhomedir (),name);
-				/* INBOX becomes mail spool directory file */
-    else sprintf (dst,MAILFILE,myusername ());
-  }
-  return dst;
-}
-
 /* Berkeley open and lock mailbox
  * Accepts: file name to open/lock
  *	    file open mode
@@ -1117,7 +1129,7 @@ int bezerk_lock (char *file,int flags,int mode,char *lock,int op)
   struct timeval tp;
   struct stat sb;
 				/* build lock filename */
-  strcat (bezerk_file (lock,file),".lock");
+  strcat (dummy_file (lock,file),".lock");
   do {				/* until OK or out of tries */
     gettimeofday (&tp,NIL);	/* get the time now */
 #ifdef NFSKLUDGE
@@ -1129,7 +1141,8 @@ int bezerk_lock (char *file,int flags,int mode,char *lock,int op)
     j = strlen (tmp);		/* append local host name */
     gethostname (tmp + j,(MAILTMPLEN - j) - 1);
 				/* try to get hitching-post file */
-    if ((ld = open (tmp,O_WRONLY|O_CREAT|O_EXCL,0666)) < 0) {
+    if ((ld = open (tmp,O_WRONLY|O_CREAT|O_EXCL,
+		    (int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL))) < 0) {
 				/* prot fail & non-ex, don't use lock files */
       if ((errno == EACCES) && (stat (tmp,&sb))) *lock = '\0';
       else {			/* otherwise something strange is going on */
@@ -1140,7 +1153,8 @@ int bezerk_lock (char *file,int flags,int mode,char *lock,int op)
       }
     }
     else {			/* got a hitching-post */
-      chmod (tmp,0666);		/* make sure others can break the lock */
+				/* make sure others can break the lock */
+      chmod (tmp,(int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL));
       close (ld);		/* close the hitching-post */
       link (tmp,lock);		/* tie hitching-post to lock, ignore failure */
       stat (tmp,&sb);		/* get its data */
@@ -1158,23 +1172,27 @@ int bezerk_lock (char *file,int flags,int mode,char *lock,int op)
    * idea -- that's what IMAP is for -- but some people insist upon losing...
    */
 				/* try to get the lock */
-    if ((ld = open (lock,O_WRONLY|O_CREAT|O_EXCL,0666)) < 0) switch (errno) {
-    case EEXIST:		/* if extant and old, grab it for ourselves */
-      if ((!stat (lock,&sb)) && tp.tv_sec > sb.st_ctime + LOCKTIMEOUT * 60)
-	ld = open (lock,O_WRONLY|O_CREAT,0666);
-      break;
-    case EACCES:		/* protection fail, ignore if non-ex or old */
-      if (stat (lock,&sb) || (tp.tv_sec > sb.st_ctime + LOCKTIMEOUT * 60))
-	*lock = '\0';		/* assume no world write mail spool dir */
-      break;
-    default:			/* some other failure */
-      sprintf (tmp,"Error creating %s: %s",lock,strerror (errno));
-      mm_log (tmp,WARN);	/* this is probably not good */
-      *lock = '\0';		/* don't use lock files */
-      break;
-    }
+    if ((ld = open (lock,O_WRONLY|O_CREAT|O_EXCL,
+		    (int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL))) < 0)
+      switch (errno) {
+      case EEXIST:		/* if extant and old, grab it for ourselves */
+	if ((!stat (lock,&sb)) && tp.tv_sec > sb.st_ctime + LOCKTIMEOUT * 60)
+	  ld = open (lock,O_WRONLY|O_CREAT,
+		     (int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL));
+	break;
+      case EACCES:		/* protection fail, ignore if non-ex or old */
+	if (stat (lock,&sb) || (tp.tv_sec > sb.st_ctime + LOCKTIMEOUT * 60))
+	  *lock = '\0';		/* assume no world write mail spool dir */
+	break;
+      default:			/* some other failure */
+	sprintf (tmp,"Error creating %s: %s",lock,strerror (errno));
+	mm_log (tmp,WARN);	/* this is probably not good */
+	*lock = '\0';		/* don't use lock files */
+	break;
+      }
     if (ld >= 0) {		/* if made a lock file */
-      chmod (lock,0666);	/* make sure others can break the lock */
+				/* make sure others can break the lock */
+      chmod (lock,(int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL));
       close (ld);		/* close the lock file */
     }
 #endif
@@ -1206,17 +1224,15 @@ int bezerk_lock (char *file,int flags,int mode,char *lock,int op)
 void bezerk_unlock (int fd,MAILSTREAM *stream,char *lock)
 {
   struct stat sbuf;
-  struct timeval tp[2];
+  time_t tp[2];
   fstat (fd,&sbuf);		/* get file times */
 				/* if stream and csh would think new mail */
   if (stream && (sbuf.st_atime <= sbuf.st_mtime)) {
-    gettimeofday (&tp[0],NIL);	/* set atime to now */
+    tp[0] = time (0);		/* set atime to now */
 				/* set mtime to (now - 1) if necessary */
-    tp[1].tv_sec = tp[0].tv_sec > sbuf.st_mtime ? sbuf.st_mtime :
-      tp[0].tv_sec - 1;
-    tp[1].tv_usec = 0;		/* oh well */
+    tp[1] = tp[0] > sbuf.st_mtime ? sbuf.st_mtime : tp[0] - 1;
 				/* set the times, note change */
-    if (!utimes (LOCAL->name,tp)) LOCAL->filetime = tp[1].tv_sec;
+    if (!utime (LOCAL->name,tp)) LOCAL->filetime = tp[1];
   }
   flock (fd,LOCK_UN);		/* release flock'ers */
   close (fd);			/* close the file */
@@ -1240,7 +1256,7 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
   int ti = 0,zn = 0;
   int first = T;
   long nmsgs = stream->nmsgs;
-  long recent = stream->recent;
+  long newcnt = 0;
   struct stat sbuf;
   MESSAGECACHE *elt;
   FILECACHE *m = NIL,*n = NIL;
@@ -1248,16 +1264,9 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
 				/* open and lock mailbox (shared OK) */
   if ((fd = bezerk_lock (LOCAL->name,LOCAL->ld ? O_RDWR : O_RDONLY,NIL,
 			 lock,op)) < 0) {
-				/* failed, OK for non-ex file? */
-    if ((errno != ENOENT) || LOCAL->filesize) {
-      sprintf (LOCAL->buf,"Mailbox open failed, aborted: %s",strerror (errno));
-      mm_log (LOCAL->buf,ERROR);
-      bezerk_abort (stream);
-    }
-    else {			/* this is to allow for non-ex INBOX */
-      mail_exists (stream,0);	/* make sure upper level sees this as empty */
-      mail_recent (stream,0);
-    }
+    sprintf (LOCAL->buf,"Mailbox open failed, aborted: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    bezerk_abort (stream);
     mail_unlock (stream);
     return -1;
   }
@@ -1303,7 +1312,9 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
       delta -= i;		/* account for data read in */
 				/* validate newly-appended data */
       if (first) {		/* only do this first time! */
-	if (!((*s =='F') && VALID (s,t,ti,zn))) {
+	if (*s == '\n') s++;	/* skip spurious newline */
+	VALID (s,t,ti,zn);	/* see new data is valid */
+        if (!ti) {		/* invalid data? */
 	  mm_log ("Mailbox format invalidated (consult an expert), aborted",
 		  ERROR);
 	  bezerk_unlock (fd,stream,lock);
@@ -1311,7 +1322,7 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
 	  mail_unlock (stream);
 	  return -1;
 	}
-	first = NIL;		/* don't do this again */
+	first = NIL;		/* never do this again */
       }
 
 				/* found end of message or end of data? */
@@ -1352,13 +1363,14 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
   }
 				/* expand the primary cache */
   (*mailcache) (stream,nmsgs,CH_SIZE);
-  if (LOCAL->msgs) {		/* have a cache yet? */
-    if (nmsgs >= LOCAL->cachesize)
-      fs_resize ((void **) &LOCAL->msgs,
-		 (LOCAL->cachesize += CACHEINCREMENT) * sizeof (FILECACHE *));
+  if (nmsgs>=LOCAL->cachesize) {/* need to expand cache? */
+				/* number of messages plus room to grow */
+    LOCAL->cachesize = nmsgs + CACHEINCREMENT;
+    if (LOCAL->msgs)		/* resize if already have a cache */
+      fs_resize ((void **) &LOCAL->msgs,LOCAL->cachesize*sizeof (FILECACHE *));
+    else LOCAL->msgs =		/* create new cache */
+      (FILECACHE **) fs_get (LOCAL->cachesize * sizeof (FILECACHE *));
   }
-  else LOCAL->msgs = (FILECACHE **)
-    fs_get ((LOCAL->cachesize = nmsgs + CACHEINCREMENT) *sizeof (FILECACHE *));
 
   if (LOCAL->buflen > CHUNK) {	/* maybe move where the buffer is in memory*/
     fs_give ((void **) &LOCAL->buf);
@@ -1371,16 +1383,22 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
      * if memory is corrupted.  Note that in the case of a totally empty
      * message, a newline is appended and counts adjusted.
      */
-    if (!((s = m->internal) && VALID (s,t,ti,zn)))
-      if (!(s && !strchr (s,'\n') && strcat (s,"\n") && VALID (s,t,ti,zn) &&
-	    m->headersize++))
-	fatal ("Bogus entry in new cache list");
+    ti = NIL;			/* valid header not found */
+    if (s = m->internal) VALID (s,t,ti,zn);
+    if (!ti) {			/* totally empty message? */
+      if (s && !strchr (s,'\n')) {
+	strcat (s,"\n");	/* append newline */
+        VALID (s,t,ti,zn);	/* try again */
+	m->headersize++;	/* adjust count */
+      }
+      if (!ti) fatal ("Bogus entry in new cache list");
+    }
 				/* pointer to message header */
     m->header = s = strchr (t++,'\n') + 1;
     m->headersize -= m->header - m->internal;
     m->body = NIL;		/* assume no body as yet */
     m->bodysize = 0;
-    recent++;			/* assume recent by default */
+    newcnt++;			/* assume recent by default */
     (elt = mail_elt (stream,i+1))->recent = T;
 				/* calculate initial Status/X-Status lines */
     bezerk_update_status (m->status,elt);
@@ -1439,8 +1457,8 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
 	  break;
 	case 'O':		/* message old */
 	  if (elt->recent) {	/* don't do this more than once! */
-	    elt->recent = NIL;
-	    recent--;		/* not recent any longer... */
+	    elt->recent = NIL;	/* not recent any longer... */
+	    newcnt--;
 	  }
 	  break;
 	case 'D':		/* message deleted */
@@ -1504,8 +1522,8 @@ int bezerk_parse (MAILSTREAM *stream,char *lock,int op)
   LOCAL->filesize = sbuf.st_size;
   LOCAL->filetime = sbuf.st_mtime;
   mail_exists (stream,nmsgs);	/* notify upper level of new mailbox size */
-  mail_recent (stream,recent);	/* and of change in recent messages */
-  if (recent) LOCAL->dirty = T;	/* mark dirty so O flags are set */
+  mail_recent (stream,stream->recent + newcnt);
+  if (newcnt) LOCAL->dirty = T;	/* mark dirty so O flags are set */
   return fd;			/* return the winnage */
 }
 
@@ -1530,20 +1548,39 @@ char *bezerk_eom (char *som,char *sod,long i)
   strcpy (wdtest.ch,"AAAA1234");/* constant for word testing */
   while ((s > som) && *s-- != '\n');
   if (wdtest.wd != 0x41414141) {/* not a 32-bit word machine? */
-    while (s = strstr (s,"\nFrom ")) if (s++ && VALID (s,t,ti,zn)) return s;
+    while (s = strstr (s,"\nFrom ")) {
+      s++;			/* skip newline */
+      VALID (s,t,ti,zn);	/* if found start of message... */
+      if (ti) return s;		/* return that pointer */
+    }
   }
+
   else {			/* can do it faster this way */
     register Word m = 0x0a0a0a0a;
-    while ((long) s & 3)	/* any characters before word boundary? */
-      if ((*s++ == '\n') && s[0] == 'F' && VALID (s,t,ti,zn)) return s;
+				/* any characters before word boundary? */
+    while ((long) s & 3) if (*s++ == '\n') {
+      VALID (s,t,ti,zn);
+      if (ti) return s;
+    }
     i = (sod + i) - s;		/* total number of tries */
     do {			/* fast search for newline */
       if (0x80808080 & (0x01010101 + (0x7f7f7f7f & ~(m ^ *(Word *) s)))) {
-				/* interesting word, check it closer */
-	if (*s++ == '\n' && s[0] == 'F' && VALID (s,t,ti,zn)) return s;
-	else if (*s++ == '\n' && s[0] == 'F' && VALID (s,t,ti,zn)) return s;
-	else if (*s++ == '\n' && s[0] == 'F' && VALID (s,t,ti,zn)) return s;
-	else if (*s++ == '\n' && s[0] == 'F' && VALID (s,t,ti,zn)) return s;
+	if (*s++ == '\n') {	/* interesting word, check it closer */
+	  VALID (s,t,ti,zn);
+	  if (ti) return s;
+	}
+	if (*s++ == '\n') {	/* second byte */
+	  VALID (s,t,ti,zn);
+	  if (ti) return s;
+	}
+	if (*s++ == '\n') {	/* third byte */
+	  VALID (s,t,ti,zn);
+	  if (ti) return s;
+	}
+	if (*s++ == '\n') {	/* final byte */
+	  VALID (s,t,ti,zn);
+	  if (ti) return s;
+	}
       }
       else s += 4;		/* try next word */
       i -= 4;			/* count a word checked */
@@ -1643,10 +1680,14 @@ void bezerk_save (MAILSTREAM *stream,int fd)
 	}
       }
     }
+    if (fsync (fd)) {		/* make sure the updates take */
+      sprintf (LOCAL->buf,"Unable to sync mailbox: %s",strerror (e = errno));
+      mm_log (LOCAL->buf,WARN);
+      mm_diskerror (stream,e,T);
+      retry = T;		/* oops */
+    }
   } while (retry);		/* repeat if need to try again */
-  fsync (fd);			/* make sure the disk has the update */
-				/* nuke any cruft after that */
-  ftruncate (fd,LOCAL->filesize);
+  ftruncate(fd,LOCAL->filesize);/* nuke any cruft after that */
   fsync (fd);			/* is this necessary? */
   fstat (fd,&sbuf);		/* now get updated file time */
   LOCAL->filetime = sbuf.st_mtime;
@@ -1662,24 +1703,30 @@ void bezerk_save (MAILSTREAM *stream,int fd)
 
 int bezerk_copy_messages (MAILSTREAM *stream,char *mailbox)
 {
-  char file[MAILTMPLEN];
-  char lock[MAILTMPLEN];
+  char file[MAILTMPLEN],lock[MAILTMPLEN];
   struct iovec iov[16];
   int fd,iovc;
   struct stat sbuf;
+  time_t tp[2];
   long i;
   int ok = T;
 				/* make sure valid mailbox */
-  if (!bezerk_isvalid (mailbox,file)) {
-    if (errno == ENOENT)	/* failed, was it no such file? */
-      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
-    else {
-      sprintf (LOCAL->buf,"Not a Berkeley-format mailbox: %s",mailbox);
-      mm_log (LOCAL->buf,ERROR);
-    }
+  if (!bezerk_isvalid (mailbox,file)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
+    return NIL;
+  case 0:			/* merely empty file? */
+    break;
+  case EINVAL:
+    sprintf (LOCAL->buf,"Invalid Berkeley-format mailbox name: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  default:
+    sprintf (LOCAL->buf,"Not a Berkeley-format mailbox: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
     return NIL;
   }
-  if ((fd = bezerk_lock (bezerk_file (file,mailbox),O_WRONLY|O_APPEND|O_CREAT,
+  if ((fd = bezerk_lock (dummy_file (file,mailbox),O_WRONLY|O_APPEND|O_CREAT,
 			 S_IREAD|S_IWRITE,lock,LOCK_EX)) < 0) {
     sprintf (LOCAL->buf,"Can't open destination mailbox: %s",strerror (errno));
     mm_log (LOCAL->buf,ERROR);
@@ -1703,7 +1750,15 @@ int bezerk_copy_messages (MAILSTREAM *stream,char *mailbox)
       }
     }
   }
-  fsync (fd);			/* force out the update */
+  if (fsync (fd)) {		/* force out the update */
+    sprintf (LOCAL->buf,"Message copy sync failed: %s",strerror (errno));
+    mm_log (LOCAL->buf,ERROR);
+    ftruncate (fd,sbuf.st_size);
+    ok = NIL;			/* oops */
+  }
+  tp[0] = sbuf.st_atime;	/* preserve atime */
+  tp[1] = time (0);		/* set mtime to now */
+  utime (file,tp);		/* set the times */
   bezerk_unlock (fd,NIL,lock);	/* unlock and close mailbox */
   mm_nocritical (stream);	/* release critical */
   return ok;			/* return whether or not succeeded */
@@ -1767,7 +1822,7 @@ void bezerk_update_status (char *status,MESSAGECACHE *elt)
 
 short bezerk_getflags (MAILSTREAM *stream,char *flag)
 {
-  char *t;
+  char *t,tmp[MAILTMPLEN],err[MAILTMPLEN];
   short f = 0;
   short i,j;
   if (flag && *flag) {		/* no-op if no flag string */
@@ -1777,11 +1832,11 @@ short bezerk_getflags (MAILSTREAM *stream,char *flag)
       return NIL;
     }
 				/* copy the flag string w/o list construct */
-    strncpy (LOCAL->buf,flag+i,(j = strlen (flag) - (2*i)));
-    LOCAL->buf[j] = '\0';
-    t = ucase (LOCAL->buf);	/* uppercase only from now on */
+    strncpy (tmp,flag+i,(j = strlen (flag) - (2*i)));
+    tmp[j] = '\0';
+    t = ucase (tmp);		/* uppercase only from now on */
 
-    while (*t) {		/* parse the flags */
+    while (t && *t) {		/* parse the flags */
       if (*t == '\\') {		/* system flag? */
 	switch (*++t) {		/* dispatch based on first character */
 	case 'S':		/* possible \Seen flag */
@@ -1809,14 +1864,12 @@ short bezerk_getflags (MAILSTREAM *stream,char *flag)
 	}
 				/* add flag to flags list */
 	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
-	else {			/* bitch about bogus flag */
-	  mm_log ("Unknown system flag",ERROR);
-	  return NIL;
-	}
       }
       else {			/* no user flags yet */
-	mm_log ("Unknown flag",ERROR);
-	return NIL;
+	t = strtok (t," ");	/* isolate flag name */
+	sprintf (err,"Unknown flag: %.80s",t);
+	t = strtok (NIL," ");	/* get next flag */
+	mm_log (err,ERROR);
       }
     }
   }
@@ -2052,8 +2105,8 @@ search_t bezerk_search_string (search_t f,char **d,long *n)
       *n = strtol (c+1,&c,10);	/* get its length */
       if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
 	  *n > strlen (*d = c)) return NIL;
-      c[*n] = '\255';		/* write new delimiter */
-      strtok (c,"\255");	/* reset the strtok mechanism */
+      c[*n] = DELIM;		/* write new delimiter */
+      strtok (c,DELMS);		/* reset the strtok mechanism */
       break;
     default:			/* atomic string */
       *n = strlen (*d = strtok (c," "));
