@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	24 June 1992
- * Last Edited:	14 September 1993
+ * Last Edited:	22 August 1994
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -91,8 +91,6 @@ DRIVER dawzdriver = {
 
 				/* prototype stream */
 MAILSTREAM dawzproto = {&dawzdriver};
-				/* standard driver's prototype */
-MAILSTREAM *mailstd_proto = &dawzproto;
 
 /* Dawz mail validate mailbox
  * Accepts: mailbox name
@@ -115,12 +113,13 @@ long dawz_isvalid (char *name)
   int fd;
   char *s,tmp[MAILTMPLEN];
   struct stat sbuf;
+  errno = NIL;			/* zap error */
 				/* INBOX is always accepted */
   if (!strcmp (ucase (strcpy (tmp,name)),"INBOX")) return T;
 				/* if file, get its status */
-  if (*name != '{' && dawz_file (tmp,name) && (stat (tmp,&sbuf) == 0)) {
+  if (*name != '{' && mailboxfile (tmp,name) && (stat (tmp,&sbuf) == 0)) {
 				/* allow empty file */
-    if (sbuf.st_size == 0) return T;
+    if (sbuf.st_size == 0) return LONGT;
     if ((fd = open (tmp,O_BINARY|O_RDONLY,NIL)) >= 0 && read (fd,tmp,64) >= 0){
       close (fd);		/* close the file */
       if ((s = strchr (tmp,'\015')) && s[1] == '\012') {
@@ -144,7 +143,7 @@ long dawz_isvalid (char *name)
 
 void *dawz_parameters (long function,void *value)
 {
-  fatal ("Invalid dawz_parameters function");
+  return NIL;
 }
 
 /* Dawz mail find list of mailboxes
@@ -154,34 +153,7 @@ void *dawz_parameters (long function,void *value)
 
 void dawz_find (MAILSTREAM *stream,char *pat)
 {
-  void *sdb = NIL;
-  struct find_t f;
-  char *s,tmp[MAILTMPLEN],file[MAILTMPLEN];
-  char *t = sm_read (&sdb);
-  int i = 0;
-  if (t) do if ((*t != '{') && strcmp (t,"INBOX") && pmatch (t,pat) &&
-		dawz_isvalid (t)) mm_mailbox (t);
-  while (t = sm_read (&sdb));	/* read subscription database */
-  else {			/* no subscriptions, do a directory */
-				/* directory specified in pattern? */
-    if ((s = strrchr (pat,'\\')) || (*(s = pat + 1) == ':')) {
-      strncpy (file,pat,i = (++s) - pat);
-      file[i] = '\0';		/* tie off prefix */
-    }
-				/* make fully-qualified file name */
-    if (!dawz_file (tmp,pat)) return;
-				/* tie off directory name */
-    if (s = strrchr (tmp,'\\')) *s = '\0';
-    else tmp[2] = '\0';
-    sprintf (tmp + strlen (tmp),"\\*%s",DEFEXT);
-				/* loop through matching files */
-    if (!_dos_findfirst (tmp,_A_NORMAL,&f)) do {
-				/* suppress extension */
-      if (s = strchr (f.name,'.')) *s = '\0';
-      strcpy (file + i,f.name);	/* build file name */
-      if (pmatch (file,pat)) mm_mailbox (file);
-    } while (!_dos_findnext (&f));
-  }
+  if (stream) dummy_find (stream,pat);
 }
 
 
@@ -194,7 +166,8 @@ void dawz_find_bboards (MAILSTREAM *stream,char *pat)
 {
   /* Always a no-op */
 }
-
+
+
 /* Dawz mail find list of all mailboxes
  * Accepts: mail stream
  *	    pattern to search
@@ -202,27 +175,7 @@ void dawz_find_bboards (MAILSTREAM *stream,char *pat)
 
 void dawz_find_all (MAILSTREAM *stream,char *pat)
 {
-  struct find_t f;
-  char *s,tmp[MAILTMPLEN],file[MAILTMPLEN];
-  int i = 0;
-				/* directory specified in pattern? */
-  if ((s = strrchr (pat,'\\')) || (*(s = pat + 1) == ':')) {
-    strncpy (file,pat,i = (++s) - pat);
-    file[i] = '\0';		/* tie off prefix */
-  }
-				/* make fully-qualified file name */
-  if (!dawz_file (tmp,pat)) return;
-				/* tie off directory name */
-  if (s = strrchr (tmp,'\\')) *s = '\0';
-  else tmp[2] = '\0';
-  sprintf (tmp + strlen (tmp),"\\*%s",DEFEXT);
-				/* loop through matching files */
-  if (!_dos_findfirst (tmp,_A_NORMAL,&f)) do {
-				/* suppress extension */
-    if (s = strchr (f.name,'.')) *s = '\0';
-    strcpy (file + i,f.name);	/* build file name */
-    if (pmatch (file,pat) && dawz_isvalid (file)) mm_mailbox (file);
-  } while (!_dos_findnext (&f));
+  if (stream) dummy_find_all (stream,pat);
 }
 
 /* Dawz mail subscribe to mailbox
@@ -234,7 +187,7 @@ void dawz_find_all (MAILSTREAM *stream,char *pat)
 long dawz_subscribe (MAILSTREAM *stream,char *mailbox)
 {
   char tmp[MAILTMPLEN];
-  if (dawz_file (tmp,mailbox)) return sm_subscribe (mailbox);
+  if (mailboxfile (tmp,mailbox)) return sm_subscribe (mailbox);
   else return dawz_badname (tmp,mailbox);
 }
 
@@ -248,7 +201,7 @@ long dawz_subscribe (MAILSTREAM *stream,char *mailbox)
 long dawz_unsubscribe (MAILSTREAM *stream,char *mailbox)
 {
   char tmp[MAILTMPLEN];
-  if (dawz_file (tmp,mailbox)) return sm_unsubscribe (mailbox);
+  if (mailboxfile (tmp,mailbox)) return sm_unsubscribe (mailbox);
   else return dawz_badname (tmp,mailbox);
 }
 
@@ -274,7 +227,7 @@ long dawz_create (MAILSTREAM *stream,char *mailbox)
 {
   char tmp[MAILTMPLEN];
   int fd;
-  if (!dawz_file (tmp,mailbox)) return dawz_badname (tmp,mailbox);
+  if (!mailboxfile (tmp,mailbox)) return dawz_badname (tmp,mailbox);
   if ((fd = open (tmp,O_WRONLY|O_CREAT|O_EXCL,S_IREAD|S_IWRITE)) < 0) {
 				/* failed */
     sprintf (tmp,"Can't create mailbox %s: %s",mailbox,strerror (errno));
@@ -309,8 +262,8 @@ long dawz_rename (MAILSTREAM *stream,char *old,char *new)
 {
   char tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN],lockx[MAILTMPLEN];
 				/* make file name */
-  if (!dawz_file (file,old)) return dawz_badname (tmp,old);
-  if (!dawz_file (tmp,new)) return dawz_badname (tmp,new);
+  if (!mailboxfile (file,old)) return dawz_badname (tmp,old);
+  if (new && !mailboxfile (tmp,new)) return dawz_badname (tmp,new);
 				/* do the rename or delete operation */
   if (new ? rename (file,tmp) : unlink (file)) {
     sprintf (tmp,"Can't %s mailbox %s: %s",new ? "rename" : "delete",old,
@@ -344,7 +297,7 @@ MAILSTREAM *dawz_open (MAILSTREAM *stream)
     if (stream->flagstring) fs_give ((void **) &stream->flagstring);
     for (i = 0; i < NUSERFLAGS; ++i) stream->user_flags[i] = NIL;
   }
-  if (!dawz_file (tmp,stream->mailbox))
+  if (!mailboxfile (tmp,stream->mailbox))
     return (MAILSTREAM *) dawz_badname (tmp,stream->mailbox);
   if (((fd = open (tmp,O_BINARY|(stream->readonly ? O_RDONLY:O_RDWR),NIL)) < 0)
       && (strcmp (ucase (stream->mailbox),"INBOX") ||
@@ -354,8 +307,6 @@ MAILSTREAM *dawz_open (MAILSTREAM *stream)
     mm_log (tmp,ERROR);
     return NIL;
   }
-				/* hokey default for local host */
-  if (!lhostn) lhostn = cpystr ("localhost");
   stream->local = fs_get (sizeof (DAWZLOCAL));
 				/* canonicalize the stream mailbox name */
   fs_give ((void **) &stream->mailbox);
@@ -366,8 +317,8 @@ MAILSTREAM *dawz_open (MAILSTREAM *stream)
   stream->sequence++;		/* bump sequence number */
 				/* parse mailbox */
   stream->nmsgs = stream->recent = 0;
-  if (dawz_ping (stream) && !stream->nmsgs)
-    mm_log ("Mailbox is empty",(long) NIL);
+  if (!dawz_ping (stream)) return NIL;
+  if (!stream->nmsgs) mm_log ("Mailbox is empty",(long) NIL);
   return stream;		/* return stream to caller */
 }
 
@@ -529,7 +480,7 @@ ENVELOPE *dawz_fetchstructure (MAILSTREAM *stream,long msgno,BODY **body)
       d.pos = hdrpos + hdrsize;
       INIT (&bs,dawz_string,(void *) &d,textsize);
 				/* parse envelope and body */
-      rfc822_parse_msg (env,body ? b : NIL,hdr,hdrsize,&bs,lhostn,tmp);
+      rfc822_parse_msg (env,body ? b : NIL,hdr,hdrsize,&bs,mylocalhost (),tmp);
     }
     fs_give ((void **) &tmp);
     fs_give ((void **) &hdr);
@@ -977,25 +928,40 @@ long dawz_move (MAILSTREAM *stream,char *sequence,char *mailbox)
  * Returns: T if append successful, else NIL
  */
 
-long dawz_append (MAILSTREAM *stream,char *mailbox,STRING *message)
+long dawz_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
+		  STRING *message)
 {
-  int fd;
   struct stat sbuf;
+  int fd,zone;
   char tmp[MAILTMPLEN];
+  MESSAGECACHE elt;
   time_t ti;
   struct tm *t;
   long i;
   long size = SIZE (message);
+  short f = dawz_getflags (stream,flags);
+  if (date) {			/* want to preserve date? */
+				/* yes, parse date into an elt */
+    if (!mail_parse_date (&elt,date)) {
+      mm_log ("Bad date in append",ERROR);
+      return NIL;
+    }
+  }
   tzset ();			/* initialize timezone stuff */
-				/* make sure valid mailbox */
-  if (!dawz_isvalid (mailbox)) {
-    sprintf (tmp,"Not a Dawz-format mailbox: %s",mailbox);
-    mm_log (tmp,ERROR);
+  if (!dawz_isvalid (mailbox)) {/* make sure valid mailbox */
+    if (errno == ENOENT)
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",
+		 (long) NIL);
+    else if (mailboxfile (tmp,mailbox)) {
+      sprintf (tmp,"Not a Dawz-format mailbox: %s",mailbox);
+      mm_log (tmp,ERROR);
+    }
+    else dawz_badname (tmp,mailbox);
     return NIL;
   }
-  if (!dawz_file (tmp,mailbox)) return dawz_badname (tmp,mailbox);
 				/* open the destination */
-  if ((fd = open (tmp,O_BINARY|O_WRONLY|O_APPEND|O_CREAT,S_IREAD|S_IWRITE))<0){
+  if ((fd = open (mailboxfile (tmp,mailbox),
+		  O_BINARY|O_WRONLY|O_APPEND|O_CREAT,S_IREAD|S_IWRITE)) < 0) {
     sprintf (tmp,"Can't open append mailbox: %s",strerror (errno));
     mm_log (tmp,ERROR);
     return NIL;
@@ -1004,9 +970,13 @@ long dawz_append (MAILSTREAM *stream,char *mailbox,STRING *message)
   fstat (fd,&sbuf);		/* get current file size */
   ti = time (0);		/* get time now */
   t = localtime (&ti);		/* output local time */
-  sprintf (tmp,"%2d-%s-%d %02d:%02d:%02d-%s,%ld;000000000000\015\012",
-	   t->tm_mday,months[t->tm_mon],t->tm_year+1900,
-	   t->tm_hour,t->tm_min,t->tm_sec,tzname[t->tm_isdst],size);
+  zone = -(((int)timezone)/ 60);/* get timezone from TZ environment stuff */
+  if (date) mail_date(tmp,&elt);/* use date if given */
+  else sprintf (tmp,"%2d-%s-%d %02d:%02d:%02d %+03d%02d",
+		t->tm_mday,months[t->tm_mon],t->tm_year+1900,
+		t->tm_hour,t->tm_min,t->tm_sec,zone/60,abs (zone) % 60);
+				/* add remainder of header */
+  sprintf (tmp + strlen (tmp),",%ld;0000000000%02o\015\012",size,f);
 
 				/* write header */
   if (write (fd,tmp,strlen (tmp)) < 0) {
@@ -1045,24 +1015,6 @@ void dawz_gc (MAILSTREAM *stream,long gcflags)
 /* Internal routines */
 
 
-/* Dawz mail generate file string
- * Accepts: temporary buffer to write into
- *	    mailbox name string
- * Returns: local file string
- */
-
-char *dawz_file (char *dst,char *name)
-{
-  char *s;
-				/* forbid extensions of filename */
-  if (strchr ((s = strrchr (name,'\\')) ? s : name,'.')) return NIL;
-				/* absolute path name? */
-  if ((*name == '\\') || (name[1] == ':')) sprintf (dst,"%s%s",name,DEFEXT);
-  else sprintf (dst,"%s\\%s%s",myhomedir (),name,DEFEXT);
-  return ucase (dst);
-}
-
-
 /* Return bad file name error message
  * Accepts: temporary buffer
  *	    file name
@@ -1128,7 +1080,6 @@ long dawz_getflags (MAILSTREAM *stream,char *flag)
       if (!i) {			/* didn't find a matching flag? */
 	sprintf (key,"Unknown flag: %.80s",t);
 	mm_log (key,ERROR);
-	return NIL;		/* return no system flags */
       }
 				/* parse next flag */
     } while (t = strtok (NIL," "));
@@ -1224,10 +1175,22 @@ long dawz_copy_messages (MAILSTREAM *stream,char *mailbox)
   struct stat sbuf;
   MESSAGECACHE *elt;
   unsigned long i,j,k;
+  long ret = LONGT;
   int fd;
-				/* got file? */
-  if (!dawz_file (tmp,mailbox)) return dawz_badname (tmp,mailbox);
-  if ((fd = open (tmp,O_BINARY|O_WRONLY|O_APPEND|O_CREAT,S_IREAD|S_IWRITE))<0){
+  if (!dawz_isvalid (mailbox)) {/* make sure valid mailbox */
+    if (errno == ENOENT)
+      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",
+		 (long) NIL);
+    else if (mailboxfile (tmp,mailbox)) {
+      sprintf (tmp,"Not a Dawz-format mailbox: %s",mailbox);
+      mm_log (tmp,ERROR);
+    }
+    else dawz_badname (tmp,mailbox);
+    return NIL;
+  }
+				/* open the destination */
+  if ((fd = open (mailboxfile (tmp,mailbox),
+		  O_BINARY|O_WRONLY|O_APPEND|O_CREAT,S_IREAD|S_IWRITE)) < 0) {
     sprintf (tmp,"Unable to open copy mailbox: %s",strerror (errno));
     mm_log (tmp,ERROR);
     return NIL;
@@ -1235,36 +1198,27 @@ long dawz_copy_messages (MAILSTREAM *stream,char *mailbox)
   mm_critical (stream);		/* go critical */
   fstat (fd,&sbuf);		/* get current file size */
 				/* for each requested message */
-  for (i = 1; i <= stream->nmsgs; i++)
+  for (i = 1; ret && (i <= stream->nmsgs); i++)
     if ((elt = mail_elt (stream,i))->sequence) {
-      lseek (LOCAL->fd,dawz_header (stream,i,&k),SEEK_SET);
-      j = 0;			/* mark first time through */
+      lseek (LOCAL->fd,elt->data1,SEEK_SET);
+				/* number of bytes to copy */
+      k = (elt->data2 >> 24) + elt->rfc822_size;
       do {			/* read from source position */
-	if (j) {		/* get another message chunk */
-	  k = min (j,(unsigned long) MAILTMPLEN);
-	  read (LOCAL->fd,tmp,(unsigned int) k);
-	}
-	else {			/* first time through */
-	  mail_date (tmp,elt);	/* make a header */
-	  sprintf (tmp + strlen (tmp),",%d;000000000000\015\012",
-		   elt->rfc822_size);
-	  k = strlen (tmp);	/* size of header string */
-				/* amount of data to write subsequently */
-	  j = elt->rfc822_size + k;
-	}
-	if (write (fd,tmp,(unsigned int) k) < 0) {
+	j = min (k,(long) MAILTMPLEN);
+	read (LOCAL->fd,tmp,(unsigned int) j);
+	if (write (fd,tmp,(unsigned int) j) < 0) {
 	  sprintf (tmp,"Unable to write message: %s",strerror (errno));
 	  mm_log (tmp,ERROR);
 	  chsize (fd,sbuf.st_size);
-	  close (fd);		/* punt */
-	  mm_nocritical (stream);
-	  return NIL;
+	  j = k;
+	  ret = NIL;		/* note error */
+	  break;
 	}
-      } while (j -= k);		/* until done */
+      } while (k -= j);		/* until done */
     }
   close (fd);			/* close the file */
   mm_nocritical (stream);	/* release critical */
-  return T;
+  return ret;
 }
 
 /* Dawz update status string
@@ -1400,11 +1354,28 @@ char dawz_search_since (MAILSTREAM *stream,long msgno,char *d,long n)
   MESSAGECACHE *elt = mail_elt (stream,msgno);
   return (char) (((elt->year << 9) + (elt->month << 5) + elt->day) >= n);
 }
-
+
+#define BUFLEN 4*MAILTMPLEN
 
 char dawz_search_body (MAILSTREAM *stream,long msgno,char *d,long n)
 {
-  return NIL;			/* need code here */
+  char tmp[BUFLEN];
+  unsigned long bufsize,hdrsize;
+  unsigned long curpos = dawz_header (stream,msgno,&hdrsize);
+  unsigned long textsize = mail_elt (stream,msgno)->rfc822_size - hdrsize;
+				/* get to header position */
+  lseek (LOCAL->fd,curpos += hdrsize,SEEK_SET);
+  while (textsize) {
+    bufsize = min (textsize,(unsigned long) BUFLEN);
+    read (LOCAL->fd,tmp,(unsigned int) bufsize);
+    if (search (tmp,bufsize,d,n)) return T;
+				/* backtrack by pattern size if not at end */
+    if (bufsize != textsize) bufsize -= n;
+    textsize -= bufsize;	/* this many bytes handled */
+    curpos += bufsize;		/* advance to that point */
+    lseek (LOCAL->fd,curpos,SEEK_SET);
+  }
+  return NIL;			/* not found */
 }
 
 
@@ -1414,10 +1385,25 @@ char dawz_search_subject (MAILSTREAM *stream,long msgno,char *d,long n)
   return s ? search (s,(long) strlen (s),d,n) : NIL;
 }
 
-
 char dawz_search_text (MAILSTREAM *stream,long msgno,char *d,long n)
 {
-  return NIL;			/* need code here */
+  char tmp[BUFLEN];
+  unsigned long bufsize,hdrsize;
+  unsigned long curpos = dawz_header (stream,msgno,&hdrsize);
+  unsigned long textsize = mail_elt (stream,msgno)->rfc822_size;
+				/* get to header position */
+  lseek (LOCAL->fd,curpos,SEEK_SET);
+  while (textsize) {
+    bufsize = min (textsize,(unsigned long) BUFLEN);
+    read (LOCAL->fd,tmp,(unsigned int) bufsize);
+    if (search (tmp,bufsize,d,n)) return T;
+				/* backtrack by pattern size if not at end */
+    if (bufsize != textsize) bufsize -= n;
+    textsize -= bufsize;	/* this many bytes handled */
+    curpos += bufsize;		/* advance to that point */
+    lseek (LOCAL->fd,curpos,SEEK_SET);
+  }
+  return NIL;			/* not found */
 }
 
 char dawz_search_bcc (MAILSTREAM *stream,long msgno,char *d,long n)
@@ -1515,8 +1501,8 @@ search_t dawz_search_string (search_t f,char **d,long *n)
       *n = strtol (c+1,&c,10);	/* get its length */
       if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
 	  *n > strlen (*d = c)) return NIL;
-      c[*n] = '\255';		/* write new delimiter */
-      strtok (c,"\255");	/* reset the strtok mechanism */
+      c[*n] = DELIM;		/* write new delimiter */
+      strtok (c,DELMS);		/* reset the strtok mechanism */
       break;
     default:			/* atomic string */
       *n = strlen (*d = strtok (c," "));
