@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	22 November 1989
- * Last Edited:	21 October 1993
+ * Last Edited:	21 August 1994
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -45,11 +45,10 @@
 /* c-client global data */
 
 DRIVER *maildrivers = NIL;	/* list of mail drivers */
-DRIVER *inboxdriver = NIL;	/* INBOX mail driver */
-char *lhostn = NIL;		/* local host name */
 mailgets_t mailgets = NIL;	/* pointer to alternate gets function */
 				/* mail cache manipulation function */
 mailcache_t mailcache = mm_cache;
+tcptimeout_t tcptimeout = NIL;	/* TCP timeout handler routine */
 
 /* Default limited get string
  * Accepts: readin function pointer
@@ -233,8 +232,7 @@ void mail_link (DRIVER *driver)
   *d = driver;			/* put driver at the end */
   driver->next = NIL;		/* this driver is the end of the list */
 }
-
-
+
 /* Mail manipulate driver parameters
  * Accepts: mail stream
  *	    function code
@@ -242,9 +240,49 @@ void mail_link (DRIVER *driver)
  * Returns: function-dependent return value
  */
 
+				/* TCP timeouts, in seconds */
+static long mail_opentimeout = 75;
+static long mail_readtimeout = 0;
+static long mail_writetimeout = 0;
+static long mail_closetimeout = 0;
+
 void *mail_parameters (MAILSTREAM *stream,long function,void *value)
 {
-  return stream->dtb ? (stream->dtb->parameters) (function,value) : NIL;
+  void *r,*ret = NIL;
+  DRIVER *d = maildrivers;
+  switch ((int) function) {
+  case GET_OPENTIMEOUT:
+    ret = (void *) mail_opentimeout;
+    break;
+  case SET_OPENTIMEOUT:
+    mail_opentimeout = (long) value;
+    break;
+  case GET_READTIMEOUT:
+    ret = (void *) mail_readtimeout;
+    break;
+  case SET_READTIMEOUT:
+    mail_readtimeout = (long) value;
+    break;
+  case GET_WRITETIMEOUT:
+    ret = (void *) mail_writetimeout;
+    break;
+  case SET_WRITETIMEOUT:
+    mail_writetimeout = (long) value;
+    break;
+  case GET_CLOSETIMEOUT:
+    ret = (void *) mail_closetimeout;
+    break;
+  case SET_CLOSETIMEOUT:
+    mail_closetimeout = (long) value;
+    break;
+  default:
+    if (stream && stream->dtb)	/* if have a stream, do it for that stream */
+      return (*stream->dtb->parameters)(function,value);
+    else do if (r = (d->parameters) (function,value)) ret = r;
+    while (d = d->next);	/* until at the end */
+    break;
+  }
+  return ret;
 }
 
 /* Mail find list of subscribed mailboxes
@@ -257,8 +295,7 @@ void mail_find (MAILSTREAM *stream,char *pat)
   DRIVER *d = maildrivers;
 				/* if have a stream, do it for that stream */
   if (stream && stream->dtb) (*stream->dtb->find) (stream,pat);
-				/* otherwise do for all DTB's */
-  else do (d->find) (NIL,pat);
+  else do (d->find) (NIL,pat);	/* otherwise do for all DTB's */
   while (d = d->next);		/* until at the end */
 }
 
@@ -311,27 +348,39 @@ void mail_find_all_bboard (MAILSTREAM *stream,char *pat)
  * Accepts: MAIL stream
  *	    mailbox name
  *	    purpose string for error message
- *	    flag indicating this is not an open
  * Return: driver factory on success, NIL on failure
  */
 
-DRIVER *mail_valid (MAILSTREAM *stream,char *mailbox,char *purpose,long nopen)
+DRIVER *mail_valid (MAILSTREAM *stream,char *mailbox,char *purpose)
 {
-  char tmp[MAILTMPLEN];
+  char *s,tmp[MAILTMPLEN];
   DRIVER *factory;
   for (factory = maildrivers; factory && !(*factory->valid) (mailbox);
        factory = factory->next);
-				/* must match stream and not be dummy */
-  if (factory && ((stream && (stream->dtb != factory)) ||
-		  (nopen && !strcmp (factory->name,"dummy")))) factory = NIL;
+				/* must match stream if not dummy */
+  if (factory && stream && (stream->dtb != factory))
+    factory = strcmp (factory->name,"dummy") ? NIL : stream->dtb;
   if (!factory && purpose) {	/* if want an error message */
-    sprintf (tmp,"Can't %s %s: no such mailbox",purpose,mailbox);
+    switch (*mailbox) {		/* error depends upon first character */
+    case '*':			/* bboard */
+      if (mailbox[1] != '{') {	/* but only if local */
+	s = "no such bboard";
+	break;
+      }
+				/* otherwise drop into remote */
+    case '{':			/* remote specification */
+      s = "invalid remote specification";
+      break;
+    default:			/* others */
+      s = "no such mailbox";
+      break;
+    }
+    sprintf (tmp,"Can't %s %s: %s",purpose,mailbox,s);
     mm_log (tmp,ERROR);
   }
   return factory;		/* return driver factory */
 }
-
-
+
 /* Mail validate network mailbox name
  * Accepts: mailbox name
  *	    mailbox driver to validate against
@@ -343,13 +392,15 @@ DRIVER *mail_valid (MAILSTREAM *stream,char *mailbox,char *purpose,long nopen)
 DRIVER *mail_valid_net (char *name,DRIVER *drv,char *host,char *mailbox)
 {
   NETMBX mb;
-  if (!mail_valid_net_parse (name,&mb) || strcmp (mb.service,drv->name))
-    return NIL;
+  if (!mail_valid_net_parse (name,&mb) || 
+      (strcmp (mb.service,"imap") ? strcmp (mb.service,drv->name) :
+       strncmp (drv->name,"imap",4))) return NIL;
   if (host) strcpy (host,mb.host);
   if (mailbox) strcpy (mailbox,mb.mailbox);
   return drv;
 }
-
+
+
 /* Mail validate network mailbox name
  * Accepts: mailbox name
  *	    NETMBX structure to return values
@@ -362,7 +413,8 @@ long mail_valid_net_parse (char *name,NETMBX *mb)
   char c,*s,*t,*v;
   mb->port = 0;			/* initialize structure */
   *mb->host = *mb->mailbox = *mb->service = '\0';
-  mb->anoflag = NIL;		/* not (yet) anonymous */
+				/* init flags */
+  mb->anoflag = mb->dbgflag = NIL;
 				/* check if bboard */
   if (mb->bbdflag = (*name == '*') ? T : NIL) name++;
 				/* have host specification? */
@@ -403,7 +455,10 @@ long mail_valid_net_parse (char *name,NETMBX *mb)
       }
       else {			/* non-argument switch */
 	if (!strcmp (s,"anonymous")) mb->anoflag = T;
+	else if (!strcmp (s,"bboard")) mb->bbdflag = T;
+	else if (!strcmp (s,"debug")) mb->dbgflag = T;
 	else if (!strcmp (s,"imap") || !strcmp (s,"imap2") ||
+		 !strcmp (s,"imap4") || !strcmp (s,"pop3") ||
 		 !strcmp (s,"nntp")) {
 	  if (*mb->service) return NIL;
 	  else strcpy (mb->service,s);
@@ -418,8 +473,7 @@ long mail_valid_net_parse (char *name,NETMBX *mb)
 				/* default mailbox name */
   if (!*mb->mailbox) strcpy (mb->mailbox,mb->bbdflag ? "general" : "INBOX");
 				/* default service name */
-  if (!(*mb->service && strcmp (mb->service,"imap")))
-    strcpy (mb->service,"imap2");
+  if (!*mb->service) strcpy (mb->service,"imap");
   return T;
 }
 
@@ -431,7 +485,7 @@ long mail_valid_net_parse (char *name,NETMBX *mb)
 
 long mail_subscribe (MAILSTREAM *stream,char *mailbox)
 {
-  DRIVER *factory = mail_valid (stream,mailbox,"subscribe to mailbox",LONGT);
+  DRIVER *factory = mail_valid (stream,mailbox,"subscribe to mailbox");
   return factory ? (*factory->subscribe) (stream,mailbox) : NIL;
 }
 
@@ -444,7 +498,7 @@ long mail_subscribe (MAILSTREAM *stream,char *mailbox)
 
 long mail_unsubscribe (MAILSTREAM *stream,char *mailbox)
 {
- DRIVER *factory = mail_valid (stream,mailbox,"unsubscribe to mailbox",LONGT);
+ DRIVER *factory = mail_valid (stream,mailbox,"unsubscribe to mailbox");
   return factory ? (*factory->unsubscribe) (stream,mailbox) : NIL;
 }
 
@@ -460,7 +514,7 @@ long mail_subscribe_bboard (MAILSTREAM *stream,char *mailbox)
   char tmp[MAILTMPLEN];
   DRIVER *factory;
   sprintf (tmp,"*%s",mailbox);
-  return (factory = mail_valid (stream,tmp,"subscribe to bboard",LONGT)) ?
+  return (factory = mail_valid (stream,tmp,"subscribe to bboard")) ?
     (*factory->subscribe_bboard) (stream,mailbox) : NIL;
 }
 
@@ -476,7 +530,7 @@ long mail_unsubscribe_bboard (MAILSTREAM *stream,char *mailbox)
   char tmp[MAILTMPLEN];
   DRIVER *factory;
   sprintf (tmp,"*%s",mailbox);
-  return (factory = mail_valid (stream,tmp,"unsubscribe to bboard",LONGT)) ?
+  return (factory = mail_valid (stream,tmp,"unsubscribe to bboard")) ?
     (*factory->unsubscribe_bboard) (stream,mailbox) : NIL;
 }
 
@@ -488,22 +542,20 @@ long mail_unsubscribe_bboard (MAILSTREAM *stream,char *mailbox)
 
 long mail_create (MAILSTREAM *stream,char *mailbox)
 {
-  int localp = *mailbox != '{';
+  /* A local mailbox is one that is not qualified as being a remote or a
+   * namespace mailbox.  Any remote or namespace mailbox driver must check
+   * for itself whether or not the mailbox already exists. */
+  int localp = (*mailbox != '{') && (*mailbox != '#');
   char tmp[MAILTMPLEN];
-  if (!stream) {		/* guess at driver if stream not specified */
-    if (localp) {		/* if local */
-      if (!(stream = mail_open (NIL,"INBOX",OP_PROTOTYPE)))
-	stream = mailstd_proto;	/* default to standard if no INBOX */
-    }
-    else stream = mail_open (NIL,mailbox,OP_PROTOTYPE);
-  }
-  if (!stream) {		/* damn, still no prototype! */
+				/* guess at driver if stream not specified */
+  if (!(stream || (stream = localp ? default_proto () :
+		   mail_open (NIL,mailbox,OP_PROTOTYPE)))) {
     sprintf (tmp,"Can't create mailbox %s: indeterminate format",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   }
 				/* must not already exist if local */
-  if (localp && mail_valid (stream,mailbox,NIL,LONGT)) {
+  if (localp && mail_valid (stream,mailbox,NIL)) {
     sprintf (tmp,"Can't create mailbox %s: mailbox already exists",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
@@ -519,7 +571,7 @@ long mail_create (MAILSTREAM *stream,char *mailbox)
 
 long mail_delete (MAILSTREAM *stream,char *mailbox)
 {
-  DRIVER *factory = mail_valid (stream,mailbox,"delete mailbox",LONGT);
+  DRIVER *factory = mail_valid (stream,mailbox,"delete mailbox");
   return factory ? (*factory->delete) (stream,mailbox) : NIL;
 }
 
@@ -534,8 +586,8 @@ long mail_delete (MAILSTREAM *stream,char *mailbox)
 long mail_rename (MAILSTREAM *stream,char *old,char *new)
 {
   char tmp[MAILTMPLEN];
-  DRIVER *factory = mail_valid (stream,old,"rename mailbox",LONGT);
-  if ((*old != '{') && mail_valid (NIL,new,NIL,LONGT)) {
+  DRIVER *factory = mail_valid (stream,old,"rename mailbox");
+  if ((*old != '{') && (*old != '#') && mail_valid (NIL,new,NIL)) {
     sprintf (tmp,"Can't rename to mailbox %s: mailbox already exists",new);
     mm_log (tmp,ERROR);
     return NIL;
@@ -552,8 +604,8 @@ long mail_rename (MAILSTREAM *stream,char *old,char *new)
 
 MAILSTREAM *mail_open (MAILSTREAM *stream,char *name,long options)
 {
-  DRIVER *factory = mail_valid (NIL,name,(!(stream && stream->silent)) ?
-				"open mailbox" : NIL,(long) NIL);
+  DRIVER *factory = mail_valid (NIL,name,(options & OP_SILENT) ?
+				NIL : "open mailbox");
   if (factory) {		/* must have a factory */
     if (!stream) {		/* instantiate stream if none to recycle */
       if (options & OP_PROTOTYPE) return (*factory->open) (NIL);
@@ -916,15 +968,33 @@ long mail_move (MAILSTREAM *stream,char *sequence,char *mailbox)
 /* Mail append message string
  * Accepts: mail stream
  *	    destination mailbox
+ *	    initial flags
+ *	    message internal date
  *	    stringstruct of message to append
  * Returns: T on success, NIL on failure
  */
 
 long mail_append (MAILSTREAM *stream,char *mailbox,STRING *message)
 {
-  DRIVER *factory = mail_valid (stream,mailbox,"append to mailbox",(long) NIL);
+				/* compatibility jacket */
+  return mail_append_full (stream,mailbox,NIL,NIL,message);
+}
+
+
+long mail_append_full (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
+		       STRING *message)
+{
+  DRIVER *factory = mail_valid (stream,mailbox,NIL);
+  if (!factory) {		/* got a driver to use? */
+    if (!stream &&		/* ask default for TRYCREATE if no stream */
+	(*default_proto ()->dtb->append) (stream,mailbox,flags,date,message))
+      fatal ("Append validity confusion");
+				/* now generate error message */
+    mail_valid (stream,mailbox,"append to mailbox");
+    return NIL;			/* return failure */
+  }
 				/* do the driver's action */
-  return factory ? (factory->append) (stream,mailbox,message) : NIL;
+  return (factory->append) (stream,mailbox,flags,date,message);
 }
 
 /* Mail garbage collect stream
@@ -1145,15 +1215,22 @@ long mail_parse_date (MESSAGECACHE *elt,char *s)
       case (('G'-'A')*1024)+(('M'-'A')*32)+'T'-'A':
       case 'Z': elt->zhours = 0; break;
 
-      case (('L'-'A')*1024) + (('C'-'A')*32)+'L'-'A':
-	/* I knew I was going to regret that idiotic `LCL' stuff someday!! */
+      default:			/* assume local otherwise */
 	tn = time (0);		/* time now... */
 	t = localtime (&tn);	/* get local minutes since midnight */
 	m = t->tm_hour * 60 + t->tm_min;
+	ms = t->tm_yday;	/* note Julian day */
 	t = gmtime (&tn);	/* minus UTC minutes since midnight */
-				/* add a day if too far past */
-	if ((m -= t->tm_hour * 60 + t->tm_min) < -720) m += 1440;
-	if (m > 720) m -= 1440;	/* subtract a day if too far ahead */
+	m -= t->tm_hour * 60 + t->tm_min;
+	/* ms can be one of:
+	 *  36x  local time is December 31, UTC is January 1, offset -24 hours
+	 *    1  local time is 1 day ahead of UTC, offset +24 hours
+	 *    0  local time is same day as UTC, no offset
+	 *   -1  local time is 1 day behind UTC, offset -24 hours
+	 * -36x  local time is January 1, UTC is December 31, offset +24 hours
+	 */
+	if (ms -= t->tm_yday)	/* correct offset if different Julian day */
+	  m += ((ms < 0) == (abs (ms) == 1)) ? -24*60 : 24*60;
 	if (m < 0) {		/* occidental? */
 	  m = abs (m);		/* yup, make positive number */
 	  elt->zoccident = 1;	/* and note west of UTC */
@@ -1161,8 +1238,6 @@ long mail_parse_date (MESSAGECACHE *elt,char *s)
 	elt->zhours = m / 60;	/* now break into hours and minutes */
 	elt->zminutes = m % 60;
 	break;
-      default:			/* unknown time zone name */
-	return NIL;
       }
       elt->zminutes = 0;	/* never a fractional hour */
       break;
@@ -1180,6 +1255,10 @@ long mail_parse_date (MESSAGECACHE *elt,char *s)
       return NIL;
     }
   }
+  else {			/* make sure all time fields zero */
+    elt->hours = elt->minutes = elt->seconds = elt->zhours = elt->zminutes =
+      elt->zoccident = 0;
+  }
   return T;
 }
 
