@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	25 January 1993
- * Last Edited:	21 November 1993
+ * Last Edited:	21 August 1994
  *
- * Copyright 1993 by the University of Washington.
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -93,6 +93,9 @@ DRIVER nntpdriver = {
 
 				/* prototype stream */
 MAILSTREAM nntpproto = {&nntpdriver};
+
+				/* driver parameters */
+static char *nntp_newsrc = NIL;	/* news directory name */
 
 /* NNTP mail validate mailbox
  * Accepts: mailbox name
@@ -114,8 +117,24 @@ DRIVER *nntp_valid (char *name)
 
 void *nntp_parameters (long function,void *value)
 {
-  fatal ("Invalid nntp_parameters function");
-  return NIL;
+  switch ((int) function) {
+  case SET_NEWSRC:
+    if (nntp_newsrc) fs_give ((void **) &nntp_newsrc);
+    nntp_newsrc = cpystr ((char *) value);
+    break;
+  case GET_NEWSRC:
+    if (!nntp_newsrc) {		/* set news directory name if not defined */
+      char tmp[MAILTMPLEN];
+      sprintf (tmp,"%s\\NEWSRC",myhomedir ());
+      nntp_newsrc = cpystr (tmp);
+    }
+    value = (void *) nntp_newsrc;
+    break;
+  default:
+    value = NIL;		/* error case */
+    break;
+  }
+  return value;
 }
 
 /* NNTP mail find list of mailboxes
@@ -319,8 +338,6 @@ MAILSTREAM *nntp_mopen (MAILSTREAM *stream)
 				/* return prototype for OP_PROTOTYPE call */
   if (!stream) return &nntpproto;
   mail_valid_net_parse (stream->mailbox,&mb);
-				/* hokey default for local host */
-  if (!lhostn) lhostn = cpystr ("localhost");
   if (!*mb.mailbox) strcpy (mb.mailbox,"general");
   if (LOCAL) {			/* if recycle stream, see if changing hosts */
     if (strcmp (lcase (mb.host),lcase (strcpy (tmp,LOCAL->host)))) {
@@ -337,24 +354,15 @@ MAILSTREAM *nntp_mopen (MAILSTREAM *stream)
     stream->dtb = &nntpdriver;/* reattach this driver */
     mail_free_cache (stream);	/* clean up cache */
   }
+				/* in case /debug switch given */
+  if (mb.dbgflag) stream->debug = T;
 
-				/* open NNTP now if not already open */
-  if (!nstream && (tcpstream = tcp_open (mb.host,mb.port ?
-					 (long) mb.port : NNTPTCPPORT))) {
-    nstream = (SMTPSTREAM *) fs_get (sizeof (SMTPSTREAM));
-    nstream->tcpstream = tcpstream;
-    nstream->debug = stream->debug;
-    nstream->reply = NIL;
-				/* get NNTP greeting */
-    if (((i = smtp_reply (nstream)) == NNTPGREET) || (i == NNTPGREETNOPOST)) {
-      mm_log (nstream->reply + 4,(long) NIL);
-      smtp_send (nstream,"MODE","READER");
-    }
-    else {			/* oops */
-      mm_log (nstream->reply,ERROR);
-      smtp_close (nstream);	/* punt stream */
-      nstream = NIL;
-    }
+  if (!nstream) {		/* open NNTP now if not already open */
+    char *hostlist[2];
+    hostlist[0] = strcpy (tmp,mb.host);
+    if (mb.port) sprintf (tmp + strlen (tmp),":%ld",mb.port);
+    hostlist[1] = NIL;
+    nstream = nntp_open (hostlist,OP_READONLY+(stream->debug ? OP_DEBUG:NIL));
   }
   if (nstream) {		/* now try to open newsgroup */
     if ((!stream->halfopen) &&	/* open the newsgroup if not halfopen */
@@ -380,46 +388,33 @@ MAILSTREAM *nntp_mopen (MAILSTREAM *stream)
     LOCAL->tmp = cpystr (tmp);	/* build name of tmp file we use */
     stream->sequence++;		/* bump sequence number */
     stream->readonly = T;	/* make sure higher level knows readonly */
-    if (stream->halfopen) {	/* no map or state for half-open */
-      LOCAL->number = NIL;
-      LOCAL->seen = NIL;
-    }
+    LOCAL->number = NIL;
+    LOCAL->seen = NIL;
 
-    else {			/* try to get list of valid numbers */
-      sprintf (tmp,"%ld-%ld",i,j);
-      if ((smtp_send (nstream,"LISTGROUP",mb.mailbox) == NNTPGOK) ||
-	  (smtp_send (nstream,"XHDR Date",tmp) == NNTPHEAD)) {
-				/* create number map */
-	LOCAL->number = (unsigned long *) fs_get(nmsgs*sizeof (unsigned long));
-				/* initialize c-client/NNTP map */
-	for (i = 0; (i<nmsgs) && (s = tcp_getline (nstream->tcpstream)); ++i) {
-	  LOCAL->number[i] = atol (s);
-	  fs_give ((void **) &s);
-	}
-				/* get and flush the dot-line */
-	if ((s = tcp_getline (nstream->tcpstream)) && (*s == '.'))
-	  fs_give ((void **) &s);
-	else {			/* lose */
-	  mm_log ("NNTP article listing protocol failure",ERROR);
-	  nntp_close (stream);	/* do close action */
+    if (!stream->halfopen) {	/* if not half-open */
+      if (nmsgs) {		/* find what messages exist */
+	LOCAL->number = (unsigned long *) fs_get (nmsgs*sizeof(unsigned long));
+	sprintf (tmp,"%ld-%ld",i,j);
+	if ((smtp_send (nstream,"LISTGROUP",mb.mailbox) == NNTPGOK) ||
+	    (smtp_send (nstream,"XHDR Date",tmp) == NNTPHEAD)) {
+	  for (i = 0; (s = tcp_getline (nstream->tcpstream)) && strcmp (s,".");
+	       i++) {		/* initialize c-client/NNTP map */
+	    if (i < nmsgs) LOCAL->number[i] = atol (s);
+	    fs_give ((void **) &s);
+	  }
+	  if (s) fs_give ((void **) &s);
+	  if (i != nmsgs) {	/* found holes? */
+	    nmsgs = i;		/* set new size */
+	    fs_resize ((void **) &LOCAL->number,nmsgs*sizeof (unsigned long));
+	  }
 	}
+				/* assume c-client/NNTP map is entire range */
+	else for (k = 0; k < nmsgs; ++k) LOCAL->number[k] = i + k;
+				/* create state flag cache */
+	LOCAL->seen = (char *) fs_get (nmsgs * sizeof (char));
+				/* initialize state */
+	for (i = 0; i < nmsgs; ++i) LOCAL->seen[i] = NIL;
       }
-      else {			/* a vanilla NNTP server, barf */
-				/* any holes in sequence? */
-	if (nmsgs != (k = (j - i) + 1)) {
-	  sprintf (tmp,"[HOLES] %ld non-existant message(s)",k-nmsgs);
-	  mm_notify (stream,tmp,(long) NIL);
-	  nmsgs = k;		/* sure are, set new message count */
-	}
-				/* create number map */
-	LOCAL->number = (unsigned long *) fs_get(nmsgs*sizeof (unsigned long));
-				/* initialize c-client/NNTP map */
-	for (k = 0; k < nmsgs; ++k) LOCAL->number[k] = i + k;
-      }
-				/* create state list */
-      LOCAL->seen = (char *) fs_get (nmsgs * sizeof (char));
-  				/* initialize state */
-      for (i = 0; i < nmsgs; ++i) LOCAL->seen[i] = NIL;
 
       mail_exists(stream,nmsgs);/* notify upper level about mailbox size */
       i = 0;			/* nothing scanned yet */
@@ -434,7 +429,6 @@ MAILSTREAM *nntp_mopen (MAILSTREAM *stream)
       }
       if (s) {			/* newsgroup found? */
 	if (*s == ' ') s++;	/* skip whitespace */
-	if (c == '!') mm_log ("Not subscribed to that newsgroup",WARN);
 	while (*s && i < nmsgs){/* process until run out of messages or list */
 	  j = strtol (s,&s,10);	/* start of possible range */
 				/* other end of range */
@@ -598,8 +592,7 @@ ENVELOPE *nntp_fetchstructure (MAILSTREAM *stream,long msgno,BODY **body)
 				/* calculate message size */
     elt->rfc822_size = hdrsize + textsize;
 				/* parse envelope and body */
-    rfc822_parse_msg (env,body ? b : NIL,h,hdrsize,body ? &bs : NIL,
-		      lhostn,tmp);
+    rfc822_parse_msg (env,body ? b : NIL,h,hdrsize,body ? &bs:NIL,BADHOST,tmp);
     fs_give ((void **) &h);	/* don't need header any more */
 				/* clean up the file descriptor */
     if (body) close ((int) bs.data);
@@ -965,7 +958,7 @@ long nntp_ping (MAILSTREAM *stream)
   /* Kludge alert: SMTPSOFTFATAL is 421 which is used in NNTP to mean ``No
    * next article in this group''.  Hopefully, no NNTP server will choke on
    * a bogus command. */
-  return (smtp_send (LOCAL->nntpstream,"PING","PONG") != SMTPSOFTFATAL);
+  return (smtp_send (LOCAL->nntpstream,"NOOP",NIL) != SMTPSOFTFATAL);
 }
 
 
@@ -1048,7 +1041,8 @@ long nntp_move (MAILSTREAM *stream,char *sequence,char *mailbox)
  * Returns: T if append successful, else NIL
  */
 
-long nntp_append (MAILSTREAM *stream,char *mailbox,STRING *message)
+long nntp_append (MAILSTREAM *stream,char *mailbox,char *flags,char *date,
+		  STRING *message)
 {
   mm_log ("Append not valid for NNTP",ERROR);
   return NIL;
@@ -1077,14 +1071,14 @@ char *nntp_read_sdb (FILE **f)
 {
   int i;
   char *s,*t,tmp[MAILTMPLEN];
-  if (!*f) {			/* first time through? */
-    NEWSRC (tmp);		/* yes, make name of newsrc file and open */
-    if (!(*f = fopen (tmp,"r"))) {
-      char msg[MAILTMPLEN];
-      sprintf (msg,"No news state found, will create %s",tmp);
-      mm_log (msg,WARN);
-      return NIL;
-    }
+				/* if first time, open newsrc file */
+  if (!(*f || (*f = fopen (NEWSRC (tmp),"r")))) {
+    char msg[MAILTMPLEN];
+    sprintf (msg,"No news state found, will create %s",tmp);
+    mm_log (msg,WARN);
+    if ((i = open (tmp,O_WRONLY|O_APPEND|O_CREAT,S_IREAD|S_IWRITE)) >= 0)
+      close (i);
+    return NIL;
   }
 				/* read a line from the file */
   if (fgets (tmp,MAILTMPLEN,*f)) {
@@ -1116,12 +1110,18 @@ long nntp_update_sdb (char *name,char *data)
   int i = strlen (name);
   char c,*s,tmp[MAILTMPLEN],new[MAILTMPLEN];
   FILE *f = NIL,*of,*nf;
-  OLDNEWSRC (tmp);		/* make name of old newsrc file */
+				/* make name of old newsrc */
+  if ((s = strrchr (NEWSRC (tmp),'\\')) && (s = strrchr (tmp,'.')))
+    strcpy (s,".OLD");
+  else strcat (tmp,".OLD");
   if (!(of = fopen (tmp,"w"))) {/* open old newsrc */
     mm_log ("Can't create backup of news state",ERROR);
     return NIL;
   }
-  NEWNEWSRC (new);		/* make name of new newsrc file */
+				/* make name of new newsrc */
+  if ((s = strrchr (NEWSRC (new),'\\')) && (s = strrchr (new,'.')))
+    strcpy (s,".NEW");
+  else strcat (new,".NEW");
   if (!(nf = fopen (new,"w"))) {/* open new newsrc */
     mm_log ("Can't create new news state",ERROR);
     fclose (of);
@@ -1130,30 +1130,26 @@ long nntp_update_sdb (char *name,char *data)
 				/* process .newsrc file */
   while (s = nntp_read_sdb (&f)) {
     fprintf (of,"%s\n",s);	/* write to backup file */
-    if ((!strncmp (s,name,i)) && (((c = s[i]) == ':') || (c == '!'))) {
-      if (data) switch (*data) {/* first time we saw this entry... */
+    if (data && (!strncmp (s,name,i)) && (((c = s[i]) == ':') || (c == '!'))) {
+      switch (*data) {		/* first time we saw this entry... */
       case ':':			/* subscription request */
 	if (c == '!') c = ':';	/* subscribe if unsubscribed */
 	else {			/* complain if already subscribed */
 	  sprintf (tmp,"Already subscribed to newsgroup %s",name);
-	  mm_log (tmp,ERROR);
+	  mm_log (tmp,WARN);
 	}
-	data = s + i;		/* preserve old read state */
+	data = s + i + 1;	/* preserve old read state */
 	break;
       case '!':			/* unsubscription request? */
 	if (c == ':') c = '!';	/* unsubscribe if subscribed */
-	else {			/* complain if not subscribed */
-	  sprintf (tmp,"Not subscribed to newsgroup %s",name);
-	  mm_log (tmp,ERROR);
-	}
-	data = s + i;		/* preserve old read state */
+	data = s + i + 1;	/* preserve old read state */
 	break;
       default:			/* update read state */
 	break;
       }
 				/* write the new entry */
-	fprintf (nf,"%s%c %s\n",name,c,data);
-	data = NIL;		/* request satisfied */
+      fprintf (nf,"%s%c%s\n",name,c,data);
+      data = NIL;		/* request satisfied */
     }
     else fprintf (nf,"%s\n",s);	/* not the entry we want, write to new file */
     fs_give ((void **) &s);
@@ -1163,9 +1159,8 @@ long nntp_update_sdb (char *name,char *data)
   case ':':			/* subscription request */
     fprintf (nf,"%s: \n",name);
     break;
-  case '!':			/* unsubscription request? */
-    sprintf (tmp,"Not subscribed to newsgroup %s",name);
-    mm_log (tmp,ERROR);
+  case '!':			/* unsubscription request */
+    fprintf (nf,"%s! \n",name);
     break;
   default:			/* update read state */
     fprintf (nf,"%s: %s\n",name,data);
@@ -1173,8 +1168,7 @@ long nntp_update_sdb (char *name,char *data)
   }
   fclose (nf);			/* close new file */
   fclose (of);			/* close backup file */
-  NEWSRC (tmp);			/* make name of current file */
-  unlink (tmp);			/* remove the current file */
+  unlink (NEWSRC (tmp));	/* remove the current file */
   if (rename (new,tmp)) {	/* rename new database to current */
     mm_log ("Can't update news state",ERROR);
     return NIL;
@@ -1191,8 +1185,7 @@ long nntp_update_sdb (char *name,char *data)
 
 short nntp_getflags (MAILSTREAM *stream,char *flag)
 {
-  char tmp[MAILTMPLEN];
-  char *t;
+  char *t,tmp[MAILTMPLEN],err[MAILTMPLEN];
   short f = 0;
   short i,j;
   if (flag && *flag) {		/* no-op if no flag string */
@@ -1206,7 +1199,7 @@ short nntp_getflags (MAILSTREAM *stream,char *flag)
     tmp[j] = '\0';
     t = ucase (tmp);		/* uppercase only from now on */
 
-    while (*t) {		/* parse the flags */
+    while (t && *t) {		/* parse the flags */
       if (*t == '\\') {		/* system flag? */
 	switch (*++t) {		/* dispatch based on first character */
 	case 'S':		/* possible \Seen flag */
@@ -1234,14 +1227,12 @@ short nntp_getflags (MAILSTREAM *stream,char *flag)
 	}
 				/* add flag to flags list */
 	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
-	else {			/* bitch about bogus flag */
-	  mm_log ("Unknown system flag",ERROR);
-	  return NIL;
-	}
       }
       else {			/* no user flags yet */
-	mm_log ("Unknown flag",ERROR);
-	return NIL;
+	t = strtok (t," ");	/* isolate flag name */
+	sprintf (err,"Unknown flag: %.80s",t);
+	t = strtok (NIL," ");	/* get next flag */
+	mm_log (err,ERROR);
       }
     }
   }
@@ -1365,11 +1356,29 @@ unsigned long nntp_msgdate (MAILSTREAM *stream,long msgno)
   if (!elt->day) nntp_fetchstructure (stream,msgno,NIL);
   return (long) (elt->year << 9) + (elt->month << 5) + elt->day;
 }
-
+
+#define BUFLEN 4*MAILTMPLEN
 
 char nntp_search_body (MAILSTREAM *stream,long msgno,char *d,long n)
 {
-  return NIL;			/* need code here */
+  char tmp[BUFLEN];
+  unsigned long bufsize,textsize,curpos = 0;
+				/* get the text */
+  sprintf (tmp,"%ld",LOCAL->number[msgno - 1]);
+  if (smtp_send (LOCAL->nntpstream,"BODY",tmp) != NNTPBODY) return NIL;
+  nntp_slurp (stream,&textsize);
+  while (textsize) {
+    bufsize = min (textsize,(unsigned long) BUFLEN);
+    read (LOCAL->fd,tmp,(unsigned int) bufsize);
+    if (search (tmp,bufsize,d,n)) return T;
+				/* backtrack by pattern size if not at end */
+    if (bufsize != textsize) bufsize -= n;
+    textsize -= bufsize;	/* this many bytes handled */
+    curpos += bufsize;		/* advance to that point */
+    lseek (LOCAL->fd,curpos,SEEK_SET);
+  }
+  close (LOCAL->fd);		/* clean up the file descriptor */
+  return NIL;			/* not found */
 }
 
 
@@ -1382,7 +1391,24 @@ char nntp_search_subject (MAILSTREAM *stream,long msgno,char *d,long n)
 
 char nntp_search_text (MAILSTREAM *stream,long msgno,char *d,long n)
 {
-  return NIL;			/* need code here */
+  char tmp[BUFLEN];
+  unsigned long bufsize,textsize,curpos = 0;
+				/* get the text */
+  sprintf (tmp,"%ld",LOCAL->number[msgno - 1]);
+  if (smtp_send (LOCAL->nntpstream,"ARTICLE",tmp) != NNTPARTICLE) return NIL;
+  nntp_slurp (stream,&textsize);
+  while (textsize) {
+    bufsize = min (textsize,(unsigned long) BUFLEN);
+    read (LOCAL->fd,tmp,(unsigned int) bufsize);
+    if (search (tmp,bufsize,d,n)) return T;
+				/* backtrack by pattern size if not at end */
+    if (bufsize != textsize) bufsize -= n;
+    textsize -= bufsize;	/* this many bytes handled */
+    curpos += bufsize;		/* advance to that point */
+    lseek (LOCAL->fd,curpos,SEEK_SET);
+  }
+  close (LOCAL->fd);		/* clean up the file descriptor */
+  return NIL;			/* not found */
 }
 
 char nntp_search_bcc (MAILSTREAM *stream,long msgno,char *d,long n)
@@ -1479,8 +1505,8 @@ search_t nntp_search_string (search_t f,char **d,long *n)
       *n = strtol (c+1,&c,10);	/* get its length */
       if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
 	  *n > strlen (*d = c)) return NIL;
-      c[*n] = '\255';		/* write new delimiter */
-      strtok (c,"\255");	/* reset the strtok mechanism */
+      c[*n] = DELIM;		/* write new delimiter */
+      strtok (c,DELMS);		/* reset the strtok mechanism */
       break;
     default:			/* atomic string */
       *n = strlen (*d = strtok (c," "));
