@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	8 July 1988
- * Last Edited:	8 September 1993
+ * Last Edited:	2 August 1994
  *
  * Sponsorship:	The original version of this work was developed in the
  *		Symbolic Systems Resources Group of the Knowledge Systems
@@ -18,8 +18,8 @@
  *		by the Biomedical Research Technology Program of the National
  *		Institutes of Health under grant number RR-00785.
  *
- * Original version Copyright 1988 by The Leland Stanford Junior University.
- * Copyright 1993 by the University of Washington.
+ * Original version Copyright 1988 by The Leland Stanford Junior University
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -46,25 +46,43 @@
 #include <signal.h>
 #include "mail.h"
 #include "osdep.h"
-#if unix
-#include <pwd.h>
-#include <netdb.h>
-#endif
-#ifdef noErr
-#include <MacTCPCommonTypes.h>
-#include <AddressXlation.h>
-#endif
 #include "rfc822.h"
 #include "smtp.h"
 #include "nntp.h"
 #include "misc.h"
 
+/* Excellent reasons to hate ifdefs, and why my real code never uses them */
 
+#ifdef __MINT__
+# define UNIXLIKE 1
+# define MAC 1
+# define MACOS 0
+#else
+# if unix
+#  define UNIXLIKE 1
+#  define MACOS 0
+# else
+#  define USWPWD 0
+#  ifdef noErr
+#   define MAC 1
+#   define MACOS 1
+#  else
+#   define MAC 0
+#   define MACOS 0
+#  endif
+# endif
+#endif
+#if UNIXLIKE
+#include <pwd.h>
+#endif
+#if MAC
+#include <MacTCPCommonTypes.h>
+#include <AddressXlation.h>
+#endif
+
 char *curhst = NIL;		/* currently connected host */
 char *curusr = NIL;		/* current login user */
-char personalname[256];		/* user's personal name */
-extern DRIVER imapdriver,bezerkdriver,tenexdriver,mboxdriver,mhdriver,
- newsdriver,nntpdriver,philedriver,dummydriver,dawzdriver;
+char personalname[MAILTMPLEN];	/* user's personal name */
 
 static char *hostlist[] = {	/* SMTP server host list */
   "mailhost",
@@ -77,7 +95,7 @@ static char *newslist[] = {	/* Netnews server host list */
   NIL
 };
 
-void main  ();
+int main  ();
 void mm  ();
 void header  ();
 void display_body  ();
@@ -87,25 +105,24 @@ void smtptest  ();
 
 /* Main program - initialization */
 
-void main ()
+int main ()
 {
   MAILSTREAM *stream = NIL;
   char tmp[MAILTMPLEN];
   long debug;
-#ifdef noErr
+#if MACOS
   size_t *base = (size_t *) 0x000908;
 				/* increase stack size on a Mac */
   SetApplLimit ((Ptr) (*base - (size_t) 65535L));
 #endif
-#if unix
+#if UNIXLIKE
   char *name;
   char *suffix;
   struct passwd *pwd;
   struct hostent *host_name;
-  gethostname (tmp,MAILTMPLEN);	/* get local name */
-				/* get it in full form */
-  curhst = (host_name = gethostbyname (tmp)) ?
-    cpystr (host_name->h_name) : cpystr (tmp);
+#ifdef __MINT__
+  mint_setup ();
+#endif
 				/* get user name and passwd entry */
   if (name = (char *) getlogin ()) pwd = getpwnam (name);
   else {
@@ -120,25 +137,10 @@ void main ()
 #else
   prompt ("Personal name: ",personalname);
   curusr = cpystr ("somebody");
-  curhst = cpystr ("someplace");
 #endif
+  curhst = cpystr (mylocalhost ());
   puts ("MTest -- C client test program");
-  mail_link (&imapdriver);	/* link in IMAP driver */
-#ifdef MSDOS
-  mail_link (&dawzdriver);	/* link in dawz mail driver */
-  mail_link (&nntpdriver);	/* link in NNTP driver */
-  mail_link (&dummydriver);	/* finally dummy driver at the end */
-#endif
-#if unix
-  mail_link (&tenexdriver);	/* link in Tenex mail driver */
-  mail_link (&mhdriver);	/* link in mh mail driver */
-  mail_link (&mboxdriver);	/* link in mbox mail driver */
-  mail_link (&bezerkdriver);	/* link in Berkeley mail driver */
-  mail_link (&newsdriver);	/* link in netnews mail driver */
-  mail_link (&nntpdriver);	/* link in NNTP driver */
-  mail_link (&philedriver);	/* line in file driver */
-  mail_link (&dummydriver);	/* finally dummy driver at the end */
-#endif
+#include "linkage.c"
 				/* user wants protocol telemetry? */
   prompt ("Debug protocol (y/n)?",tmp);
   ucase (tmp);
@@ -154,9 +156,13 @@ void main ()
     else if (tmp[0]) stream = mail_open (stream,tmp,debug ? OP_DEBUG : NIL);
   } while (!stream && tmp[0]);
   mm (stream,debug);		/* run user interface if opened */
-#ifdef noErr
+#ifdef __MINT__
+  mint_cleanup ();
+#else
+# if MACOS
 				/* clean up resolver */
   if (resolveropen) CloseResolver ();
+# endif
 #endif
 }
 
@@ -168,7 +174,7 @@ void mm (stream,debug)
 	MAILSTREAM *stream;
 	long debug;
 {
-  char cmd[256],cmdx[128];
+  char cmd[MAILTMPLEN],cmdx[MAILTMPLEN];
   char *arg;
   long i;
   long last = 0;
@@ -343,7 +349,7 @@ void header (stream,msgno)
 	long msgno;
 {
   unsigned long i;
-  char tmp[256];
+  char tmp[MAILTMPLEN];
   char *t;
   MESSAGECACHE *cache = mail_elt (stream,msgno);
   mail_fetchstructure (stream,msgno,NIL);
@@ -382,7 +388,7 @@ void display_body (body,pfx,i)
 	char *pfx;
 	long i;
 {
-  char tmp[256];
+  char tmp[MAILTMPLEN];
   char *s = tmp;
   PARAMETER *par;
   PART *part;			/* multipart doesn't have a row to itself */
@@ -435,8 +441,9 @@ void status (stream)
   rfc822_date (date);
   puts (date);
   if (stream) {
-    if (stream->mailbox) printf ("Mailbox: %s, %ld messages, %ld recent\n",
-				 stream->mailbox,stream->nmsgs,stream->recent);
+    if (stream->mailbox)
+      printf (" %s mailbox: %s, %ld messages, %ld recent\n",
+	      stream->dtb->name,stream->mailbox,stream->nmsgs,stream->recent);
     else puts ("%No mailbox is open on this stream");
     if (stream->user_flags[0]) {
       printf ("Keywords: %s",stream->user_flags[0]);
@@ -578,7 +585,7 @@ long mm_diskerror (stream,errcode,serious)
 	long errcode;
 	long serious;
 {
-#if unix
+#if UNIXLIKE
   kill (getpid (),SIGSTOP);
 #else
   abort ();
@@ -603,10 +610,13 @@ void smtptest (debug)
   unsigned char *text = (unsigned char *) fs_get (8*MAILTMPLEN);
   ENVELOPE *msg = mail_newenvelope ();
   BODY *body = mail_newbody ();
-  sprintf (line,"%s <%s@%s>",personalname,curusr,curhst);
-  rfc822_parse_adrlist (&msg->from,line,curhst);
-  sprintf (line,"%s@%s",curusr,curhst);
-  rfc822_parse_adrlist (&msg->return_path,line,curhst);
+  msg->from = mail_newaddr ();
+  msg->from->personal = cpystr (personalname);
+  msg->from->mailbox = cpystr (curusr);
+  msg->from->host = cpystr (curhst);
+  msg->return_path = mail_newaddr ();
+  msg->return_path->mailbox = cpystr (curusr);
+  msg->return_path->host = cpystr (curhst);
   prompt ("To: ",line);
   rfc822_parse_adrlist (&msg->to,line,curhst);
   if (msg->to) {
