@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	4 September 1991
- * Last Edited:	18 November 1993
+ * Last Edited:	22 August 1994
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -35,10 +35,8 @@
 
 #include <stdio.h>
 #include <ctype.h>
-#include <netdb.h>
 #include <errno.h>
 extern int errno;		/* just in case */
-#include <sys/types.h>
 #include "mail.h"
 #include "osdep.h"
 #include <sys/file.h>
@@ -138,7 +136,6 @@ void *news_parameters (function,value)
 	long function;
 	void *value;
 {
-  fatal ("Invalid news_parameters function");
   return NIL;
 }
 
@@ -164,20 +161,21 @@ void news_find_bboards (stream,pat)
 	MAILSTREAM *stream;
 	char *pat;
 {
+  FILE *f = NIL;
   void *s = NIL;
-  char *t,*u,tmp[MAILTMPLEN],patx[MAILTMPLEN];
+  char *t,*u,patx[MAILTMPLEN];
   struct stat sbuf;
   lcase (strcpy (patx,pat));	/* make sure compare with lower case */
 				/* make sure have a local news spool */
   if (!(stat (NEWSSPOOL,&sbuf) || stat (ACTIVEFILE,&sbuf) ||
 	(stream && stream->anonymous))) {
 				/* check all newsgroups from .newsrc */
-    while (t = news_read (&s)) if (u = strchr (t,':')) {
-      *u = '\0';		/* tie off at end of name */
-      if (pmatch (lcase (t),patx)) {
-	sprintf (tmp,"*%s",t);
-	if (news_valid (tmp)) mm_bboard (t);
+    while (t = news_read_sdb (&f)) {
+      if (u = strchr (t,':')) {	/* subscribed newsgroup? */
+	*u = '\0';		/* tie off at end of name */
+	if (pmatch (t,patx)) mm_bboard (t);
       }
+      fs_give ((void **) &t);	/* discard the line */
     }
   }
   while (t = sm_read (&s))	/* get data from subscription manager */
@@ -251,7 +249,8 @@ long news_unsubscribe (stream,mailbox)
 {
   return NIL;			/* never valid for netnews */
 }
-
+
+
 /* Netnews mail subscribe to bboard
  * Accepts: mail stream
  *	    bboard to add to subscription list
@@ -262,52 +261,10 @@ long news_subscribe_bboard (stream,mailbox)
 	MAILSTREAM *stream;
 	char *mailbox;
 {
-  int fd,fdo;
-  char *s,*txt;
-  char tmp[MAILTMPLEN];
-  struct stat sbuf;
-  long ret = NIL;
-				/* open .newsrc file */
-  if ((fd = open (NEWSRC,O_RDWR|O_CREAT,0600)) < 0) {
-    mm_log ("Can't update news state",ERROR);
-    return NIL;
-  }
-  flock (fd,LOCK_EX);		/* wait for exclusive access */
-  fstat (fd,&sbuf);		/* get size of data */
-  read (fd,1 + (txt = (char *) fs_get (sbuf.st_size + 2)),sbuf.st_size);
-  txt[0] = '\n';		/* make searches easier */
-  txt[sbuf.st_size + 1] = '\0';	/* tie off string */
-				/* make backup file */
-  strcat (strcpy (tmp,myhomedir ()),"/.oldnewsrc");
-  if ((fdo = open (tmp,O_WRONLY|O_CREAT,0600)) >= 0) {
-    write (fdo,txt + 1,sbuf.st_size);
-    close (fdo);
-  }
-  sprintf (tmp,"\n%s:",mailbox);/* see if already subscribed */
-  if (strstr (txt,tmp)) {	/* well? */
-    sprintf (tmp,"Already subscribed to newsgroup %s",mailbox);
-    mm_log (tmp,ERROR);
-  }
-  else {
-    sprintf (tmp,"\n%s!",mailbox);
-    if (s = strstr (txt,tmp)) {	/* if unsubscribed, just change one byte */
-      lseek (fd,strchr (s,'!') - (txt + 1),L_SET);
-      write (fd,":",1);		/* now subscribed */
-    }
-    else {			/* write new entry at end of file */
-      lseek (fd,sbuf.st_size,L_SET);
-      sprintf (tmp,"%s:\n",mailbox);
-      write (fd,tmp,strlen (tmp));
-    }
-    fsync (fd);			/* drop changes */
-    ret = T;
-  }
-  flock (fd,LOCK_UN);		/* release lock */
-  close (fd);			/* close off file */
-  fs_give ((void **) &txt);	/* free database */
-  return ret;
+  return news_update_sdb (mailbox,":");
 }
-
+
+
 /* Netnews mail unsubscribe to bboard
  * Accepts: mail stream
  *	    bboard to delete from subscription list
@@ -318,42 +275,7 @@ long news_unsubscribe_bboard (stream,mailbox)
 	MAILSTREAM *stream;
 	char *mailbox;
 {
-  int fd,fdo;
-  char *s,*txt;
-  char tmp[MAILTMPLEN];
-  struct stat sbuf;
-  long ret = NIL;
-				/* open .newsrc file */
-  if ((fd = open (NEWSRC,O_RDWR|O_CREAT,0600)) < 0) {
-    mm_log ("Can't update news state",ERROR);
-    return NIL;
-  }
-  flock (fd,LOCK_EX);		/* wait for exclusive access */
-  fstat (fd,&sbuf);		/* get size of data */
-  read (fd,1 + (txt = (char *) fs_get (sbuf.st_size + 2)),sbuf.st_size);
-  txt[0] = '\n';		/* make searches easier */
-  txt[sbuf.st_size + 1] = '\0';	/* tie off string */
-				/* make backup file */
-  strcat (strcpy (tmp,myhomedir ()),"/.oldnewsrc");
-  if ((fdo = open (tmp,O_WRONLY|O_CREAT,0600)) >= 0) {
-    write (fdo,txt + 1,sbuf.st_size);
-    close (fdo);
-  }
-  sprintf (tmp,"\n%s:",mailbox);/* see if subscribed */
-  if (!(s = strstr (txt,tmp))) {/* well? */
-    sprintf (tmp,"Not subscribed to newsgroup %s",mailbox);
-    mm_log (tmp,ERROR);
-  }
-  else {			/* unsubscribe */
-    lseek (fd,strchr (s,':') - (txt + 1),L_SET);
-    write (fd,"!",1);		/* now unsubscribed */
-    fsync (fd);			/* drop changes */
-    ret = T;
-  }
-  flock (fd,LOCK_UN);		/* release lock */
-  close (fd);			/* close off file */
-  fs_give ((void **) &txt);	/* free database */
-  return ret;
+  return news_update_sdb (mailbox,"!");
 }
 
 /* Netnews mail create mailbox
@@ -409,8 +331,7 @@ MAILSTREAM *news_open (stream)
   long i,nmsgs,j,k;
   long recent = 0,unseen = 0;
   char c = NIL,*s,*t,tmp[MAILTMPLEN];
-  void *sdb = NIL;
-  struct hostent *host_name;
+  FILE *f = NIL;
   struct direct **names;
   				/* return prototype for OP_PROTOTYPE call */
   if (!stream) return &newsproto;
@@ -422,11 +343,6 @@ MAILSTREAM *news_open (stream)
   lcase (stream->mailbox);	/* build news directory name */
   sprintf (s = tmp,"%s/%s",NEWSSPOOL,stream->mailbox + 1);
   while (s = strchr (s,'.')) *s = '/';
-  if (!lhostn) {		/* have local host yet? */
-    gethostname(tmp,MAILTMPLEN);/* get local host name */
-    lhostn = cpystr ((host_name = gethostbyname (tmp)) ?
-			  host_name->h_name : tmp);
-  }
 				/* scan directory */
   if ((nmsgs = scandir (tmp,&names,news_select,news_numsort)) >= 0) {
     stream->local = fs_get (sizeof (NEWSLOCAL));
@@ -453,14 +369,16 @@ MAILSTREAM *news_open (stream)
 
     i = 0;			/* nothing scanned yet */
     if (!stream->anonymous) {	/* not if anonymous you don't */
-      while ((t = news_read (&sdb)) && (s = strpbrk (t,":!")) && (c = *s)) {
-	*s++ = '\0';		/* tie off newsgroup name, point to data */
-	if (strcmp (t,LOCAL->name)) s = NIL;
-	else break;		/* found it! */
+      while (t = news_read_sdb (&f)) {
+	if ((s = strpbrk (t,":!")) && (c = *s)) {
+	  *s++ = '\0';		/* tie off newsgroup name, point to data */
+	  if (strcmp (t,LOCAL->name)) s = NIL;
+	  else break;		/* found it! */
+	}
+	fs_give ((void **) &t);	/* give back this entry line */
       }
       if (s) {			/* newsgroup found? */
 	if (*s == ' ') s++;	/* skip whitespace */
-	if (c == '!') mm_log ("Not subscribed to that newsgroup",WARN);
 	while (*s && i < nmsgs){/* process until run out of messages or list */
 	  j = strtol (s,&s,10);	/* start of possible range */
 				/* other end of range */
@@ -483,7 +401,10 @@ MAILSTREAM *news_open (stream)
 	}
       }
       else mm_log ("No state for newsgroup found, reading as new",WARN);
-      if (t) fs_give (&sdb);	/* free up database if necessary */
+      if (t) {			/* need to free up cruft? */
+	fs_give ((void **) &t);	/* yes, give back newsrc entry */
+	fclose (f);		/* close the file */
+      }
     }
     if (unseen) {		/* report first unseen message */
       sprintf (tmp,"[UNSEEN] %ld is first unseen message",unseen);
@@ -621,7 +542,7 @@ ENVELOPE *news_fetchstructure (stream,msgno,body)
     t = LOCAL->body[msgno - 1] ? LOCAL->body[msgno - 1] : "";
     INIT (&bs,mail_string,(void *) t,strlen (t));
 				/* parse envelope and body */
-    rfc822_parse_msg (env,body ? b : NIL,h,strlen (h),&bs,lhostn,LOCAL->buf);
+    rfc822_parse_msg (env,body ? b : NIL,h,strlen (h),&bs,BADHOST,LOCAL->buf);
   }
   if (body) *body = *b;		/* return the body */
   return *env;			/* return the envelope */
@@ -932,62 +853,10 @@ long news_ping (stream)
 void news_check (stream)
 	MAILSTREAM *stream;
 {
-  int fd;
-  long i,j,k;
-  char *s;
-  char tmp[MAILTMPLEN];
-  struct stat sbuf;
-  struct iovec iov[3];
-				/* never do if anonymous or no updates */
-  if (stream->anonymous || !LOCAL->dirty) return;
-  *LOCAL->buf = '\n';		/* header to make for easier searches */
-				/* open .newsrc file */
-  if ((fd = open (NEWSRC,O_RDWR|O_CREAT,0600)) < 0) {
-    mm_log ("Can't update news state",ERROR);
-    return;
-  }
-  flock (fd,LOCK_EX);		/* wait for exclusive access */
-  fstat (fd,&sbuf);		/* get size of data */
-				/* ensure enough room */
-  if (sbuf.st_size >= (LOCAL->buflen + 1)) {
-				/* fs_resize does an unnecessary copy */
-    fs_give ((void **) &LOCAL->buf);
-    LOCAL->buf = (char *) fs_get ((LOCAL->buflen = sbuf.st_size + 1) + 1);
-  }
-  *LOCAL->buf = '\n';		/* slurp the silly thing in */
-  read (fd,iov[0].iov_base = LOCAL->buf + 1,iov[0].iov_len = sbuf.st_size);
-				/* tie off file */
-  LOCAL->buf[sbuf.st_size + 1] = '\0';
-				/* make backup file */
-  strcat (strcpy (tmp,myhomedir ()),"/.oldnewsrc");
-  if ((i = open (tmp,O_WRONLY|O_CREAT,0600)) >= 0) {
-    write (i,LOCAL->buf + 1,sbuf.st_size);
-    close (i);
-  }
-				/* find as subscribed newsgroup */
-  sprintf (tmp,"\n%s:",LOCAL->name);
-  if (s = strstr (LOCAL->buf,tmp)) s += strlen (tmp);
-  else {			/* find as unsubscribed newsgroup */
-    sprintf (tmp,"\n%s!",LOCAL->name);
-    if (s = strstr (LOCAL->buf,tmp)) s += strlen (tmp);
-  }
-  iov[2].iov_base = "";		/* dummy in case no third block */
-  iov[2].iov_len = 0;
-
-  if (s) {			/* found existing, calculate prefix length */
-    iov[0].iov_len = s - (LOCAL->buf + 1);
-    if (s = strchr (s+1,'\n')) {/* find suffix */
-      iov[2].iov_base = ++s;	/* suffix base and length */
-      iov[2].iov_len = sbuf.st_size - (s - (LOCAL->buf + 1));
-    }
-    s = tmp;			/* pointer to dump sequence numbers */
-  }
-  else {			/* not found, append as unsubscribed group */
-    sprintf (tmp,"%s!",LOCAL->name);
-    s = tmp + strlen (tmp);	/* point to end of string */
-  }
-  *s++ = ' ';			/* leading space */
-  *s = '\0';			/* go through list */
+  unsigned long i,j,k;
+  char *s,tmp[MAILTMPLEN];
+  if (!LOCAL->dirty) return;	/* never do if no updates */
+  *(s = tmp) = '\0';		/* initialize list */
   for (i = 0,j = 1,k = 0; i < stream->nmsgs; ++i) {
     if (LOCAL->seen[i]) {	/* seen message? */
       k = LOCAL->number[i];	/* this is the top of the current range */
@@ -997,25 +866,19 @@ void news_check (stream)
 				/* calculate end of range */
       if (k = LOCAL->number[i] - 1) {
 				/* dump range */
-	sprintf (s,(j == k) ? "%d," : "%d-%d,",j,k);
+	sprintf (s,(j == k) ? "%ld," : "%ld-%ld,",j,k);
 	s += strlen (s);	/* find end of string */
       }
       j = 0;			/* no more range in progress */
     }
   }
   if (j) {			/* dump trailing range */
-    sprintf (s,(j == k) ? "%d" : "%d-%d",j,k);
+    sprintf (s,(j == k) ? "%ld" : "%ld-%ld",j,k);
     s += strlen (s);		/* find end of string */
   }
   else if (s[-1] == ',') s--;	/* prepare to patch out any trailing comma */
-  *s++ = '\n';			/* trailing newline */
-  iov[1].iov_base = tmp;	/* this group text */
-  iov[1].iov_len = s - tmp;	/* length of the text */
-  lseek (fd,0,L_SET);		/* go to beginning of file */
-  writev (fd,iov,iov[2].iov_len ? 3 : 2);
-  ftruncate (fd,iov[0].iov_len + iov[1].iov_len + iov[2].iov_len);
-  flock (fd,LOCK_UN);		/* unlock the file */
-  close (fd);			/* flush .newsrc file */
+  *s++ = '\0';			/* tie off string */
+  news_update_sdb (LOCAL->name,tmp);
 }
 
 
@@ -1072,10 +935,12 @@ long news_move (stream,sequence,mailbox)
  * Returns: T if append successful, else NIL
  */
 
-long news_append (stream,mailbox,message)
+long news_append (stream,mailbox,flags,date,message)
 	MAILSTREAM *stream;
 	char *mailbox;
-	STRING *message;
+	char *flags;
+	char *date;
+	 		  STRING *message;
 {
   mm_log ("Append not valid for netnews",ERROR);
   return NIL;
@@ -1103,31 +968,117 @@ void news_gc (stream,gcflags)
 /* Internal routines */
 
 
-/* Read netnews database
- * Accepts: pointer to netnews database handle (handle NIL if first time)
- * Returns: character string for netnews database or NIL if done
- * Note - uses strtok() mechanism
+/* Read NEWS database
+ * Accepts: pointer to subscription database file handle (NIL if first time)
+ * Returns: line from the file
  */
 
-char *news_read (sdb)
-	void **sdb;
+char *news_read_sdb (f)
+	FILE **f;
 {
-  int fd;
+  int i;
   char *s,*t,tmp[MAILTMPLEN];
-  struct stat sbuf;
-  if (!*sdb) {			/* first time through? */
-    if ((fd = open (NEWSRC,O_RDONLY,NIL)) < 0) return NIL;
-    fstat (fd,&sbuf);		/* get file size and read data */
-    read (fd,s = (char *) (*sdb = fs_get (sbuf.st_size + 1)),sbuf.st_size);
-    close (fd);			/* close file */
-    s[sbuf.st_size] = '\0';	/* tie off string */
-    if (t = strtok (s,"\n")) return t;
+				/* if first time, open newsrc file */
+  if (!(*f || (*f = fopen (NEWSRC,"r")))) {
+    char msg[MAILTMPLEN];
+    sprintf (msg,"No news state found, will create %s",tmp);
+    mm_log (msg,WARN);
+    if ((i = open (tmp,O_WRONLY|O_APPEND|O_CREAT,
+		   (int) mail_parameters (NIL,GET_SUBPROTECTION,NIL))) >= 0)
+      close (i);
+    return NIL;
   }
-				/* subsequent times through database */
-  else if (t = strtok (NIL,"\n")) return t;
-  fs_give (sdb);		/* free database */
+				/* read a line from the file */
+  if (fgets (tmp,MAILTMPLEN,*f)) {
+    i = strlen (tmp);		/* how many characters we got */
+    if (tmp[i - 1] == '\n') {	/* got a complete line? */
+      tmp[i - 1] = '\0';	/* yes, tie off the line */
+      s = cpystr (tmp);		/* make return string */
+    }
+    else {			/* ugh, have to build from fragments */
+      t = news_read_sdb (f);	/* recuse to get remainder */
+      sprintf (s = (char *) fs_get (i + strlen (t) + 1),"%s%s",tmp,t);
+      fs_give ((void **) &t);	/* discard fragment */
+    }
+    return s;			/* return the string we made */
+  }
+  fclose (*f);			/* end of file, close file */
+  *f = NIL;			/* make sure caller knows */
   return NIL;			/* all done */
 }
+
+/* Update NEWS database
+ * Accepts: newsgroup name
+ *	    message data
+ * Returns: T if success, NIL if failure
+ */
+
+long news_update_sdb (name,data)
+	char *name;
+	char *data;
+{
+  int i = strlen (name);
+  char c,*s,tmp[MAILTMPLEN],new[MAILTMPLEN];
+  FILE *f = NIL,*of,*nf;
+  strcat (strcpy (tmp,myhomedir ()),"/.oldnewsrc");
+  if (!(of = fopen (tmp,"w"))) {/* open old newsrc */
+    mm_log ("Can't create backup of news state",ERROR);
+    return NIL;
+  }
+  strcat (strcpy (new,myhomedir ()),"/.newnewsrc");
+  if (!(nf = fopen (new,"w"))) {/* open new newsrc */
+    mm_log ("Can't create new news state",ERROR);
+    fclose (of);
+    return NIL;
+  }
+				/* process .newsrc file */
+  while (s = news_read_sdb (&f)) {
+    fprintf (of,"%s\n",s);	/* write to backup file */
+    if (data && (!strncmp (s,name,i)) && (((c = s[i]) == ':') || (c == '!'))) {
+      switch (*data) {		/* first time we saw this entry... */
+      case ':':			/* subscription request */
+	if (c == '!') c = ':';	/* subscribe if unsubscribed */
+	else {			/* complain if already subscribed */
+	  sprintf (tmp,"Already subscribed to newsgroup %s",name);
+	  mm_log (tmp,WARN);
+	}
+	data = s + i + 1;	/* preserve old read state */
+	break;
+      case '!':			/* unsubscription request? */
+	if (c == ':') c = '!';	/* unsubscribe if subscribed */
+	data = s + i + 1;	/* preserve old read state */
+	break;
+      default:			/* update read state */
+	break;
+      }
+				/* write the new entry */
+      fprintf (nf,"%s%c%s\n",name,c,data);
+      data = NIL;		/* request satisfied */
+    }
+    else fprintf (nf,"%s\n",s);	/* not the entry we want, write to new file */
+    fs_give ((void **) &s);
+  }
+
+  if (data) switch (*data) {	/* if didn't find it, make new entry */
+  case ':':			/* subscription request */
+    fprintf (nf,"%s: \n",name);
+    break;
+  case '!':			/* unsubscription request */
+    fprintf (nf,"%s! \n",name);
+    break;
+  default:			/* update read state */
+    fprintf (nf,"%s: %s\n",name,data);
+    break;
+  }
+  fclose (nf);			/* close new file */
+  fclose (of);			/* close backup file */
+  unlink (NEWSRC);		/* remove the current file, name in tmp */
+  if (rename (new,tmp)) {	/* rename new database to current */
+    mm_log ("Can't update news state",ERROR);
+    return NIL;
+  }
+  return LONGT;			/* return success */
+}
 
 
 /* Parse flag list
@@ -1140,7 +1091,7 @@ short news_getflags (stream,flag)
 	MAILSTREAM *stream;
 	char *flag;
 {
-  char *t;
+  char *t,tmp[MAILTMPLEN],err[MAILTMPLEN];
   short f = 0;
   short i,j;
   if (flag && *flag) {		/* no-op if no flag string */
@@ -1150,11 +1101,11 @@ short news_getflags (stream,flag)
       return NIL;
     }
 				/* copy the flag string w/o list construct */
-    strncpy (LOCAL->buf,flag+i,(j = strlen (flag) - (2*i)));
-    LOCAL->buf[j] = '\0';
-    t = ucase (LOCAL->buf);	/* uppercase only from now on */
+    strncpy (tmp,flag+i,(j = strlen (flag) - (2*i)));
+    tmp[j] = '\0';
+    t = ucase (tmp);		/* uppercase only from now on */
 
-    while (*t) {		/* parse the flags */
+    while (t && *t) {		/* parse the flags */
       if (*t == '\\') {		/* system flag? */
 	switch (*++t) {		/* dispatch based on first character */
 	case 'S':		/* possible \Seen flag */
@@ -1182,14 +1133,12 @@ short news_getflags (stream,flag)
 	}
 				/* add flag to flags list */
 	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
-	else {			/* bitch about bogus flag */
-	  mm_log ("Unknown system flag",ERROR);
-	  return NIL;
-	}
       }
       else {			/* no user flags yet */
-	mm_log ("Unknown flag",ERROR);
-	return NIL;
+	t = strtok (t," ");	/* isolate flag name */
+	sprintf (err,"Unknown flag: %.80s",t);
+	t = strtok (NIL," ");	/* get next flag */
+	mm_log (err,ERROR);
       }
     }
   }
@@ -1548,8 +1497,8 @@ search_t news_search_string (f,d,n)
       *n = strtol (c+1,&c,10);	/* get its length */
       if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
 	  *n > strlen (*d = c)) return NIL;
-      c[*n] = '\255';		/* write new delimiter */
-      strtok (c,"\255");	/* reset the strtok mechanism */
+      c[*n] = DELIM;		/* write new delimiter */
+      strtok (c,DELMS);		/* reset the strtok mechanism */
       break;
     default:			/* atomic string */
       *n = strlen (*d = strtok (c," "));
