@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	25 August 1993
- * Last Edited:	28 September 1993
+ * Last Edited:	15 August 1994
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -35,7 +35,6 @@
 
 #include <stdio.h>
 #include <ctype.h>
-#include <netdb.h>
 #include <errno.h>
 extern int errno;		/* just in case */
 #include <signal.h>
@@ -119,16 +118,11 @@ int phile_isvalid (name,tmp)
   int i,fd,ti,zn;
   struct stat sbuf;
   char *s = tmp,*t;
-  struct hostent *host_name;
-  if (!lhostn) {		/* have local host yet? */
-    gethostname(tmp,MAILTMPLEN);/* get local host name */
-    lhostn = cpystr ((host_name = gethostbyname (tmp)) ?
-		     host_name->h_name : tmp);
-  }
 				/* INBOX is never accepted */
   return (strcmp (ucase (strcpy (tmp,name)),"INBOX") &&
 	  (*name != '{') && !((*name == '*') && (name[1] == '{')) &&
-	  !stat (phile_file (tmp,name),&sbuf) && !(sbuf.st_mode & S_IFDIR));
+	  !stat (mailboxfile (tmp,name),&sbuf) && !(sbuf.st_mode & S_IFDIR) &&
+	  sbuf.st_size);
 }
 
 /* File manipulate driver parameters
@@ -141,7 +135,6 @@ void *phile_parameters (function,value)
 	long function;
 	void *value;
 {
-  fatal ("Invalid phile_parameters function");
   return NIL;
 }
 
@@ -207,7 +200,7 @@ long phile_subscribe (stream,mailbox)
 	char *mailbox;
 {
   char tmp[MAILTMPLEN];
-  return sm_subscribe (phile_file (tmp,mailbox));
+  return sm_subscribe (mailboxfile (tmp,mailbox));
 }
 
 
@@ -222,7 +215,7 @@ long phile_unsubscribe (stream,mailbox)
 	char *mailbox;
 {
   char tmp[MAILTMPLEN];
-  return sm_unsubscribe (phile_file (tmp,mailbox));
+  return sm_unsubscribe (mailboxfile (tmp,mailbox));
 }
 
 
@@ -267,7 +260,7 @@ long phile_create (stream,mailbox)
 	MAILSTREAM *stream;
 	char *mailbox;
 {
-  /* Never valid for Phile */
+  return dummy_create (stream,mailbox);
 }
 
 
@@ -281,7 +274,7 @@ long phile_delete (stream,mailbox)
 	MAILSTREAM *stream;
 	char *mailbox;
 {
-  /* Never valid for Phile */
+  return dummy_delete (stream,mailbox);
 }
 
 
@@ -297,7 +290,7 @@ long phile_rename (stream,old,new)
 	char *old;
 	char *new;
 {
-  /* Never valid for Phile */
+  return dummy_rename (stream,old,new);
 }
 
 /* File open
@@ -324,7 +317,7 @@ MAILSTREAM *phile_open (stream)
     mail_free_cache (stream);	/* clean up cache */
   }
 				/* canonicalize the stream mailbox name */
-  phile_file (tmp,stream->mailbox);
+  mailboxfile (tmp,stream->mailbox);
 				/* force readonly if bboard */
   if (*stream->mailbox == '*') stream->readonly = T;
   else {			/* canonicalize name */
@@ -349,7 +342,7 @@ MAILSTREAM *phile_open (stream)
   LOCAL->body = mail_newbody ();
   t = gmtime (&sbuf.st_mtime);
   elt->hours = t->tm_hour; elt->minutes = t->tm_min; elt->seconds = t->tm_sec;
-  elt->day = t->tm_mday; elt->month = t->tm_mon;
+  elt->day = t->tm_mday; elt->month = t->tm_mon + 1;
   elt->year = t->tm_year - (BASEYEAR - 1900);
   sprintf (tmp,"%s, %d %s %d %02d:%02d:%02d +0000",
 	   days[t->tm_wday],t->tm_mday,months[t->tm_mon],t->tm_year+1900,
@@ -362,7 +355,7 @@ MAILSTREAM *phile_open (stream)
   if (pw = getpwuid (sbuf.st_uid)) strcpy (tmp,pw->pw_name);
   else sprintf (tmp,"User-Number-%ld",(long) sbuf.st_uid);
   LOCAL->env->from->mailbox = cpystr (tmp);
-  LOCAL->env->from->host = cpystr (lhostn);
+  LOCAL->env->from->host = cpystr (mylocalhost ());
 				/* set subject to be mailbox name */
   LOCAL->env->subject = cpystr (stream->mailbox);
 				/* slurp the data */
@@ -797,10 +790,12 @@ long phile_move (stream,sequence,mailbox)
  * Returns: T if append successful, else NIL
  */
 
-long phile_append (stream,mailbox,message)
+long phile_append (stream,mailbox,flags,date,message)
 	MAILSTREAM *stream;
 	char *mailbox;
-	STRING *message;
+	char *flags;
+	char *date;
+	 		   STRING *message;
 {
   mm_log ("Append not valid for file",ERROR);
   return NIL;
@@ -822,42 +817,6 @@ void phile_gc (stream,gcflags)
 /* Internal routines */
 
 
-/* File generate file string
- * Accepts: temporary buffer to write into
- *	    mailbox name string
- * Returns: local file string
- */
-
-
-char *phile_file (dst,name)
-	char *dst;
-	char *name;
-{
-  struct passwd *pw;
-  char *s,*t,tmp[MAILTMPLEN];
-  switch (*name) {
-  case '*':			/* bboard? */
-    sprintf (tmp,"~ftp/%s",(name[1] == '/') ? name+2 : name+1);
-    dst = phile_file (dst,tmp);/* recurse to get result */
-    break;
-  case '/':			/* absolute file path */
-    strcpy (dst,name);		/* copy the mailbox name */
-    break;
-  case '~':			/* home directory */
-    if (name[1] == '/') t = myhomedir ();
-    else {
-      strcpy (tmp,name + 1);	/* copy user name */
-      if (s = strchr (tmp,'/')) *s = '\0';
-      t = ((pw = getpwnam (tmp)) && pw->pw_dir) ? pw->pw_dir : "/NOSUCHUSER";
-    }
-    sprintf (dst,"%s%s",t,(s = strchr (name,'/')) ? s : "");
-    break;
-  default:			/* other name */
-    sprintf (dst,"%s/%s",myhomedir (),name);
-  }
-  return dst;
-}
-
 /* Parse flag list
  * Accepts: MAIL stream
  *	    flag list as a character string
@@ -869,7 +828,7 @@ short phile_getflags (stream,flag)
 	MAILSTREAM *stream;
 	char *flag;
 {
-  char *t,tmp[MAILTMPLEN];
+  char *t,tmp[MAILTMPLEN],err[MAILTMPLEN];
   short f = 0;
   short i,j;
   if (flag && *flag) {		/* no-op if no flag string */
@@ -883,7 +842,7 @@ short phile_getflags (stream,flag)
     tmp[j] = '\0';
     t = ucase (tmp);		/* uppercase only from now on */
 
-    while (*t) {		/* parse the flags */
+    while (t && *t) {		/* parse the flags */
       if (*t == '\\') {		/* system flag? */
 	switch (*++t) {		/* dispatch based on first character */
 	case 'S':		/* possible \Seen flag */
@@ -911,14 +870,12 @@ short phile_getflags (stream,flag)
 	}
 				/* add flag to flags list */
 	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
-	else {			/* bitch about bogus flag */
-	  mm_log ("Unknown system flag",ERROR);
-	  return NIL;
-	}
       }
       else {			/* no user flags yet */
-	mm_log ("Unknown flag",ERROR);
-	return NIL;
+	t = strtok (t," ");	/* isolate flag name */
+	sprintf (err,"Unknown flag: %.80s",t);
+	t = strtok (NIL," ");	/* get next flag */
+	mm_log (err,ERROR);
       }
     }
   }
@@ -1254,8 +1211,8 @@ search_t phile_search_string (f,d,n)
       *n = strtol (c+1,&c,10);	/* get its length */
       if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
 	  *n > strlen (*d = c)) return NIL;
-      c[*n] = '\255';		/* write new delimiter */
-      strtok (c,"\255");	/* reset the strtok mechanism */
+      c[*n] = DELIM;		/* write new delimiter */
+      strtok (c,DELMS);		/* reset the strtok mechanism */
       break;
     default:			/* atomic string */
       *n = strlen (*d = strtok (c," "));
