@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	22 May 1990
- * Last Edited:	18 November 1993
+ * Last Edited:	15 August 1994
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -36,7 +36,6 @@
 
 #include <stdio.h>
 #include <ctype.h>
-#include <netdb.h>
 #include <errno.h>
 extern int errno;		/* just in case */
 #include "mail.h"
@@ -112,39 +111,39 @@ DRIVER *tenex_valid (name)
  * Returns: T if valid, NIL otherwise
  */
 
-long tenex_isvalid (name,tmp)
+int tenex_isvalid (name,tmp)
 	char *name;
 	char *tmp;
 {
   int i,fd;
-  char *s;
+  int ret = NIL;
+  char *s,file[MAILTMPLEN];
   struct stat sbuf;
-  struct hostent *host_name;
-  if (!lhostn) {		/* have local host yet? */
-    gethostname(tmp,MAILTMPLEN);/* get local host name */
-    lhostn = cpystr ((host_name = gethostbyname (tmp)) ?
-		     host_name->h_name : tmp);
-  }
+  time_t tp[2];
+  errno = EINVAL;		/* assume invalid argument */
 				/* if file, get its status */
   if ((*name != '{') && !((*name == '*') && (name[1] == '{')) &&
-      (stat (tenex_file (tmp,name),&sbuf) == 0)) {
-    if (sbuf.st_size != 0) {	/* if non-empty file */
-      if ((fd = open (tmp,O_RDONLY,NIL)) >= 0 && read (fd,tmp,64) >= 0) {
-	close (fd);		/* close the file */
-	if (s = strchr (tmp,'\n')) {
-	  *s = '\0';		/* tie off header */
+      (!stat (tenex_file (file,name),&sbuf)) &&
+      ((fd = open (file,O_RDONLY,NIL)) >= 0)) {
+    errno = 0;			/* clear error */
+    if (sbuf.st_size) {		/* allow if non-empty */
+      if ((read (fd,tmp,64) >= 0) && (s = strchr (tmp,'\n')) &&
+	  (s[-1] != '\015')) {
+	*s = '\0';		/* tie off header */
 				/* must begin with dd-mmm-yy" */
-	  if (((tmp[2] == '-' && tmp[6] == '-') ||
-	       (tmp[1] == '-' && tmp[5] == '-')) &&
-	      (s = strchr (tmp+20,',')) && strchr (s+2,';')) return LONGT;
-	}
+	ret = (((tmp[2] == '-' && tmp[6] == '-') ||
+		(tmp[1] == '-' && tmp[5] == '-')) &&
+	       (s = strchr (tmp+20,',')) && strchr (s+2,';')) ? T : NIL;
       }
+      tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
+      tp[1] = sbuf.st_mtime;
+      utime (file,tp);		/* set the times */
     }
-				/* allow empty if a ".TxT" file */
-    else if ((i = strlen (tmp)) > 4 && !strcmp (tmp + i - 4 ,".TxT"))
-      return LONGT;
+				/* allow empty file if INBOX */
+    else if (!strcmp (ucase (strcpy (tmp,name)),"INBOX")) ret = T;
+    close (fd);			/* close the file */
   }
-  return NIL;			/* failed miserably */
+  return ret;			/* return what we should */
 }
 
 
@@ -158,7 +157,6 @@ void *tenex_parameters (function,value)
 	long function;
 	void *value;
 {
-  fatal ("Invalid tenex_parameters function");
   return NIL;
 }
 
@@ -280,20 +278,7 @@ long tenex_create (stream,mailbox)
 	MAILSTREAM *stream;
 	char *mailbox;
 {
-  char tmp[MAILTMPLEN];
-  int i,fd;
-				/* must be a ".TxT" file */
-  if ((i = strlen (mailbox)) < 4 || strcmp (mailbox + i - 4 ,".TxT")) {
-    mm_log ("Can't create mailbox: name must end with .TxT",ERROR);
-    return NIL;
-  }
-  if ((fd = open (tenex_file (tmp,mailbox),O_WRONLY|O_CREAT|O_EXCL,0600)) < 0){
-    sprintf (tmp,"Can't create mailbox %s: %s",mailbox,strerror (errno));
-    mm_log (tmp,ERROR);
-    return NIL;
-  }
-  close (fd);			/* close the file */
-  return LONGT;			/* return success */
+  return dummy_create (stream,mailbox);
 }
 
 
@@ -325,7 +310,7 @@ long tenex_rename (stream,old,new)
   long ret = T;
   char tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN];
   int ld;
-  int fd = open (tenex_file (file,old),O_RDONLY,NIL);
+  int fd = open (tenex_file (file,old),O_RDWR,NIL);
 				/* lock out non c-client applications */
   if (fd < 0) {			/* open mailbox */
     sprintf (tmp,"Can't open mailbox %s: %s",old,strerror (errno));
@@ -366,53 +351,16 @@ long tenex_rename (stream,old,new)
 MAILSTREAM *tenex_open (stream)
 	MAILSTREAM *stream;
 {
-  long i;
   int fd,ld;
-  char *s,*t,*k;
   char tmp[MAILTMPLEN];
-  struct stat sbuf;
 				/* return prototype for OP_PROTOTYPE call */
-  if (!stream) return &tenexproto;
+  if (!stream) return user_flags (&tenexproto);
   if (LOCAL) {			/* close old file if stream being recycled */
     tenex_close (stream);	/* dump and save the changes */
     stream->dtb = &tenexdriver;	/* reattach this driver */
     mail_free_cache (stream);	/* clean up cache */
   }
-  else {			/* flush old flagstring and flags if any */
-    if (stream->flagstring) fs_give ((void **) &stream->flagstring);
-    for (i = 0; i < NUSERFLAGS; ++i) stream->user_flags[i] = NIL;
-				/* open .imaprc or .mminit file */
-    if ((fd = open (strcat (strcpy (tmp,myhomedir ()),"/.imaprc"),O_RDONLY,
-		    NIL)) < 0)
-      fd = open (strcat (strcpy (tmp,myhomedir ()),"/.mminit"),O_RDONLY,NIL);
-    if (fd >= 0) {		/* got an init file? */
-      fstat (fd,&sbuf);		/* yes, get size */
-      read (fd,(s = stream->flagstring = (char *) fs_get (sbuf.st_size + 1)),
-	    sbuf.st_size);	/* make readin area and read the file */
-      close (fd);		/* don't need the file any more */
-      s[sbuf.st_size] = '\0';	/* tie it off */
-				/* parse init file */
-      while (*s && (t = strchr (s,'\n'))) {
-	*t = '\0';		/* tie off line, find second space */
-	if ((k = strchr (s,' ')) && (k = strchr (++k,' '))) {
-	  *k = '\0';		/* tie off two words, is it "set keywords"? */
-	  if (!strcmp (lcase (s),"set keywords")) {
-				/* yes, get first keyword */
-	    k = strtok (++k,", ");
-				/* until parsed all keywords or empty */
-	    for (i = 0; k && i < NUSERFLAGS; ++i) {
-				/* set keyword, get next one */
-	      stream->user_flags[i] = k;
-	      k = strtok (NIL,", ");
-	    }
-	    break;		/* don't need to look at rest of file */
-	  }
-	}
-	s = ++t;		/* try next line */
-      }
-    }
-  }
-
+  user_flags (stream);		/* set up user flags */
 				/* force readonly if bboard */
   if (*stream->mailbox == '*') stream->readonly = T;
   if (stream->readonly ||
@@ -428,8 +376,10 @@ MAILSTREAM *tenex_open (stream)
     }
   }
   stream->local = fs_get (sizeof (TENEXLOCAL));
+  LOCAL->buf = (char *) fs_get (MAXMESSAGESIZE + 1);
+  LOCAL->buflen = MAXMESSAGESIZE;
 				/* note if an INBOX or not */
-  LOCAL->inbox = !strcmp (ucase (stream->mailbox),"INBOX");
+  LOCAL->inbox = !strcmp (ucase (strcpy (LOCAL->buf,stream->mailbox)),"INBOX");
   if (*stream->mailbox != '*') {/* canonicalize the stream mailbox name */
     fs_give ((void **) &stream->mailbox);
     stream->mailbox = cpystr (tmp);
@@ -442,8 +392,8 @@ MAILSTREAM *tenex_open (stream)
   flock(LOCAL->fd = fd,LOCK_SH);/* bind and lock the file */
   tenex_unlock (ld,tmp);	/* release shared parse permission */
   LOCAL->filesize = 0;		/* initialize parsed file size */
-  LOCAL->buf = (char *) fs_get (MAXMESSAGESIZE + 1);
-  LOCAL->buflen = MAXMESSAGESIZE;
+  LOCAL->filetime = 0;		/* time not set up yet */
+  LOCAL->mustcheck = LOCAL->shouldcheck = NIL;
   stream->sequence++;		/* bump sequence number */
 				/* parse mailbox */
   stream->nmsgs = stream->recent = 0;
@@ -553,7 +503,8 @@ ENVELOPE *tenex_fetchstructure (stream,msgno,body)
     text[textsize] = '\0';	/* make sure tied off */
     INIT (&bs,mail_string,(void *) text,textsize);
 				/* parse envelope and body */
-    rfc822_parse_msg (env,body ? b : NIL,hdr,hdrsize,&bs,lhostn,LOCAL->buf);
+    rfc822_parse_msg (env,body ? b : NIL,hdr,hdrsize,&bs,mylocalhost (),
+		      LOCAL->buf);
     fs_give ((void **) &text);
     fs_give ((void **) &hdr);
   }
@@ -727,14 +678,19 @@ void tenex_setflag (stream,sequence,flag)
 {
   MESSAGECACHE *elt;
   long i;
-  long uf;
+  unsigned long uf;
   short f;
+  struct stat sbuf;
+  if (LOCAL->filetime && !LOCAL->shouldcheck) {
+    fstat (LOCAL->fd,&sbuf);	/* get current write time */
+    if (LOCAL->filetime < sbuf.st_mtime) LOCAL->shouldcheck = T;
+  }
 				/* no-op if no flags to modify */
   if (!((f = tenex_getflags (stream,flag,&uf)) || uf)) return;
 				/* get sequence and loop on it */
   if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
-    if ((elt = tenex_elt (stream,i))->sequence) {
-				/* set all requested flags */
+    if ((elt = mail_elt (stream,i))->sequence) {
+      tenex_elt (stream,i);	/* set all requested flags */
       if (f&fSEEN) elt->seen = T;
       if (f&fDELETED) elt->deleted = T;
       if (f&fFLAGGED) elt->flagged = T;
@@ -743,8 +699,11 @@ void tenex_setflag (stream,sequence,flag)
 				/* recalculate status */
       tenex_update_status (stream,i,NIL);
     }
-				/* make sure the update takes */
-  if (!stream->readonly) fsync (LOCAL->fd);
+  if (!stream->readonly) {	/* make sure the update takes */
+    fsync (LOCAL->fd);
+    fstat (LOCAL->fd,&sbuf);	/* get current write time */
+    LOCAL->filetime = sbuf.st_mtime;
+  }
 }
 
 /* Tenex mail clear flag
@@ -760,14 +719,19 @@ void tenex_clearflag (stream,sequence,flag)
 {
   MESSAGECACHE *elt;
   long i;
-  long uf;
+  unsigned long uf;
   short f;
+  struct stat sbuf;
+  if (LOCAL->filetime && !LOCAL->shouldcheck) {
+    fstat (LOCAL->fd,&sbuf);	/* get current write time */
+    if (LOCAL->filetime < sbuf.st_mtime) LOCAL->shouldcheck = T;
+  }
 				/* no-op if no flags to modify */
   if (!((f = tenex_getflags (stream,flag,&uf)) || uf)) return;
 				/* get sequence and loop on it */
   if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
-    if ((elt = tenex_elt (stream,i))->sequence) {
-				/* clear all requested flags */
+    if ((elt = mail_elt (stream,i))->sequence) {
+      tenex_elt (stream,i);	/* clear all requested flags */
       if (f&fSEEN) elt->seen = NIL;
       if (f&fDELETED) elt->deleted = NIL;
       if (f&fFLAGGED) elt->flagged = NIL;
@@ -776,8 +740,11 @@ void tenex_clearflag (stream,sequence,flag)
 				/* recalculate status */
       tenex_update_status (stream,i,NIL);
     }
-				/* make sure the update takes */
-  if (!stream->readonly) fsync (LOCAL->fd);
+  if (!stream->readonly) {	/* make sure the update takes */
+    fsync (LOCAL->fd);
+    fstat (LOCAL->fd,&sbuf);	/* get current write time */
+    LOCAL->filetime = sbuf.st_mtime;
+  }
 }
 
 /* Tenex mail search for messages
@@ -793,7 +760,7 @@ void tenex_search (stream,criteria)
   char *d;
   search_t f;
 				/* initially all searched */
-  for (i = 1; i <= stream->nmsgs; ++i) tenex_elt (stream,i)->searched = T;
+  for (i = 1; i <= stream->nmsgs; ++i) mail_elt (stream,i)->searched = T;
 				/* get first criterion */
   if (criteria && (criteria = strtok (criteria," "))) {
 				/* for each criterion */
@@ -890,10 +857,23 @@ void tenex_search (stream,criteria)
 long tenex_ping (stream)
 	MAILSTREAM *stream;
 {
+  long i = 1;
   long r = NIL;
   int ld;
   char lock[MAILTMPLEN];
+  struct stat sbuf;
   if (stream && LOCAL) {	/* only if stream already open */
+    if (LOCAL->filetime && !(LOCAL->mustcheck || LOCAL->shouldcheck)) {
+      fstat (LOCAL->fd,&sbuf);	/* get current write time */
+      if (LOCAL->filetime < sbuf.st_mtime) LOCAL->shouldcheck = T;
+    }
+				/* check for changed message status */
+    if (LOCAL->mustcheck || LOCAL->shouldcheck) {
+      if (LOCAL->shouldcheck)	/* babble when we do this unilaterally */
+	mm_notify (stream,"[CHECK] Checking for flag updates",NIL);
+      while (i <= stream->nmsgs) tenex_elt (stream,i++);
+      LOCAL->mustcheck = LOCAL->shouldcheck = NIL;
+    }
 				/* get shared parse/append permission */
     if ((ld = tenex_lock (LOCAL->fd,lock,LOCK_SH)) >= 0) {
 				/* parse resulting mailbox */
@@ -923,12 +903,9 @@ long tenex_ping (stream)
 void tenex_check (stream)
 	MAILSTREAM *stream;
 {
-  long i = 1;
-  if (tenex_ping (stream)) {	/* ping mailbox */
-				/* get new message status */
-    while (i <= stream->nmsgs) tenex_elt (stream,i++);
-    mm_log ("Check completed",(long) NIL);
-  }
+				/* mark that a check is desired */
+  if (LOCAL) LOCAL->mustcheck = T;
+  if (tenex_ping (stream)) mm_log ("Check completed",(long) NIL);
 }
 
 /* Tenex mail snarf messages from bezerk inbox
@@ -945,17 +922,17 @@ void tenex_snarf (stream)
   char lock[MAILTMPLEN];
   MESSAGECACHE *elt;
   MAILSTREAM *bezerk = NIL;
-  int ld = tenex_lock (LOCAL->fd,lock,LOCK_EX);
-  if (ld < 0) return;		/* give up if can't get exclusive permission */
+  int ld;
+				/* give up if can't get exclusive permission */
+  if ((!strcmp (sysinbox (),stream->mailbox)) ||
+      ((ld = tenex_lock (LOCAL->fd,lock,LOCK_EX)) < 0)) return;
   mm_critical (stream);		/* go critical */
-				/* calculate name of bezerk file */
-  sprintf (LOCAL->buf,MAILFILE,myusername ());
-  stat (LOCAL->buf,&sbuf);	/* see if anything there */
+  stat (sysinbox (),&sbuf);	/* see if anything there */
   if (sbuf.st_size) {		/* non-empty? */
     fstat (LOCAL->fd,&sbuf);	/* yes, get current file size */
 				/* sizes match and can get bezerk mailbox? */
     if ((sbuf.st_size == LOCAL->filesize) &&
-	(bezerk = mail_open (bezerk,LOCAL->buf,OP_SILENT)) &&
+	(bezerk = mail_open (bezerk,sysinbox (),OP_SILENT)) &&
 	(!bezerk->readonly) && (r = bezerk->nmsgs)) {
 				/* yes, go to end of file in our mailbox */
       lseek (LOCAL->fd,sbuf.st_size,L_SET);
@@ -965,10 +942,10 @@ void tenex_snarf (stream)
 	iov[1].iov_base = bezerk_snarf (bezerk,i,&j);
 				/* calculate header line */
 	mail_date ((iov[0].iov_base = LOCAL->buf),elt = mail_elt (bezerk,i));
-	sprintf (LOCAL->buf + strlen (LOCAL->buf),",%d;0000000000%02o\n",
-		 iov[1].iov_len = j,fOLD + (fSEEN * elt->seen) +
-		 (fDELETED * elt->deleted) + (fFLAGGED * elt->flagged) +
-		 (fANSWERED * elt->answered));
+	sprintf (LOCAL->buf + strlen (LOCAL->buf),",%ld;0000000000%02o\n",
+		 iov[1].iov_len = j,(fOLD * !(elt->recent)) +
+		 (fSEEN * elt->seen) + (fDELETED * elt->deleted) +
+		 (fFLAGGED * elt->flagged) + (fANSWERED * elt->answered));
 	iov[0].iov_len = strlen (iov[0].iov_base);
 				/* copy message to new mailbox */
 	if (writev (LOCAL->fd,iov,2) < 0) {
@@ -978,9 +955,12 @@ void tenex_snarf (stream)
 	  r = 0;		/* flag that we have lost big */
 	}
       }
+      fsync (LOCAL->fd);	/* make sure changes dumped out */
+      fstat (LOCAL->fd,&sbuf);	/* yes, get current file size */
+      LOCAL->filetime = sbuf.st_mtime;
       if (r) {			/* delete all the messages we copied */
 	for (i = 1; i <= r; i++) mail_elt (bezerk,i)->deleted = T;
-	mail_expunge (bezerk);/* now expunge all those messages */
+	mail_expunge (bezerk);	/* now expunge all those messages */
       }
     }
     if (bezerk) mail_close (bezerk);
@@ -996,6 +976,7 @@ void tenex_snarf (stream)
 void tenex_expunge (stream)
 	MAILSTREAM *stream;
 {
+  struct stat sbuf;
   int ld;
   unsigned long i = 1;
   unsigned long j,k,m,recent;
@@ -1009,6 +990,10 @@ void tenex_expunge (stream)
     mm_log ("Expunge ignored on readonly mailbox",WARN);
     return;
   }
+  if (LOCAL->filetime && !LOCAL->shouldcheck) {
+    fstat (LOCAL->fd,&sbuf);	/* get current write time */
+    if (LOCAL->filetime < sbuf.st_mtime) LOCAL->shouldcheck = T;
+  }
   /* The cretins who designed flock() created a window of vulnerability in
    * upgrading locks from shared to exclusive or downgrading from exclusive
    * to shared.  Rather than maintain the lock at shared status at a minimum,
@@ -1063,6 +1048,9 @@ void tenex_expunge (stream)
     mm_log (LOCAL->buf,(long) NIL);
   }
   else mm_log ("No messages deleted, so no update needed",(long) NIL);
+  fsync (LOCAL->fd);		/* force disk update */
+  fstat (LOCAL->fd,&sbuf);	/* get new write time */
+  LOCAL->filetime = sbuf.st_mtime;
   mm_nocritical (stream);	/* release critical */
 				/* notify upper level of new mailbox size */
   mail_exists (stream,stream->nmsgs);
@@ -1104,6 +1092,7 @@ long tenex_move (stream,sequence,mailbox)
 	char *mailbox;
 {
   long i;
+  struct stat sbuf;
   MESSAGECACHE *elt;
   if (!(mail_sequence (stream,sequence) &&
 	tenex_copy_messages (stream,mailbox))) return NIL;
@@ -1114,8 +1103,11 @@ long tenex_move (stream,sequence,mailbox)
 				/* recalculate status */
       tenex_update_status (stream,i,NIL);
     }
-				/* make sure the update takes */
-  if (!stream->readonly) fsync (LOCAL->fd);
+  if (!stream->readonly) {	/* make sure the update takes */
+    fsync (LOCAL->fd);
+    fstat (LOCAL->fd,&sbuf);	/* get current write time */
+    LOCAL->filetime = sbuf.st_mtime;
+  }
   return LONGT;
 }
 
@@ -1126,25 +1118,54 @@ long tenex_move (stream,sequence,mailbox)
  * Returns: T if append successful, else NIL
  */
 
-long tenex_append (stream,mailbox,message)
+long tenex_append (stream,mailbox,flags,date,message)
 	MAILSTREAM *stream;
 	char *mailbox;
-	STRING *message;
+	char *flags;
+	char *date;
+	 		   STRING *message;
 {
   struct stat sbuf;
-  char c,*s,*t,tmp[MAILTMPLEN],lock[MAILTMPLEN];
   int fd,ld;
+  char c,*s,*t,tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN];
+  time_t tp[2];
+  MESSAGECACHE elt;
   long i = SIZE (message);
   long size = 0;
   long ret = LONGT;
+  unsigned long uf = 0;
+  short f = 0;
+  if (flags) {			/* get flags if given */
+    unsigned long ruf;
+    f = tenex_getflags (user_flags (&tenexproto),flags,&ruf);
+				/* reverse bits (dontcha wish we had CIRC?) */
+    while (ruf) uf |= 1 << (29 - find_rightmost_bit (&ruf));
+  }
+  if (date) {			/* want to preserve date? */
+				/* yes, parse date into an elt */
+    if (!mail_parse_date (&elt,date)) {
+      mm_log ("Bad date in append",ERROR);
+      return NIL;
+    }
+  }
 				/* N.B.: can't use LOCAL->buf for tmp */
 				/* make sure valid mailbox */
-  if (!tenex_isvalid (mailbox,tmp)) {
+  if (!tenex_isvalid (mailbox,tmp)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
+    return NIL;
+  case 0:			/* merely empty file? */
+    break;
+  case EINVAL:
+    sprintf (tmp,"Invalid Tenex-format mailbox name: %s",mailbox);
+    mm_log (tmp,ERROR);
+    return NIL;
+  default:
     sprintf (tmp,"Not a Tenex-format mailbox: %s",mailbox);
     mm_log (tmp,ERROR);
     return NIL;
   }
-  if ((fd = open (tenex_file (tmp,mailbox),O_RDWR|O_CREAT,
+  if ((fd = open (tenex_file (file,mailbox),O_RDWR|O_CREAT,
 		  S_IREAD|S_IWRITE)) < 0) {
     sprintf (tmp,"Can't open append mailbox: %s",strerror (errno));
     mm_log (tmp,ERROR);
@@ -1162,12 +1183,16 @@ long tenex_append (stream,mailbox,message)
   mm_critical (stream);		/* go critical */
   fstat (fd,&sbuf);		/* get current file size */
   lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
-  rfc822_date (tmp);		/* get the date in RFC822 format */
-  tmp[4] = '0';			/* in case 1-digit day */
-  t = tmp[7] == ' ' ? tmp+5 : tmp+4;
-  t[2] = t[6] = '-';		/* convert date string to Tenex format */
+				/* write preseved date */
+  if (date) mail_date (t = tmp,&elt);
+  else {
+    rfc822_date (tmp);		/* get current date in RFC822 format */
+				/* handle 1-digit date */
+    t = tmp[7] == ' ' ? tmp+5 : tmp+4;
+    t[2] = t[6] = '-';		/* convert date string to IMAP format */
+  }
 				/* add remainder of header */
-  sprintf (t+26,",%ld;000000000000\n",size);
+  sprintf (t+26,",%ld;%010lo%02o\n",size,uf,f);
 				/* write header */
   if ((write (fd,t,strlen (t)) < 0) || ((write (fd,s,size)) < 0)) {
     sprintf (tmp,"Message append failed: %s",strerror (errno));
@@ -1176,6 +1201,9 @@ long tenex_append (stream,mailbox,message)
     ret = NIL;
   }
   fsync (fd);			/* force out the update */
+  tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
+  tp[1] = sbuf.st_mtime;
+  utime (file,tp);		/* set the times */
   tenex_unlock (ld,lock);	/* release exclusive parse/append permission */
   close (fd);			/* close the file */
   mm_nocritical (stream);	/* release critical */
@@ -1216,7 +1244,9 @@ int tenex_lock (fd,lock,op)
   if (fstat (fd,&sbuf)) return -1;
 				/* make temporary file name */
   sprintf (lock,"/tmp/.%hx.%lx",sbuf.st_dev,sbuf.st_ino);
-  if ((ld = open (lock,O_RDWR|O_CREAT,0666)) < 0) return NIL;
+  if ((ld = open (lock,O_RDWR|O_CREAT,
+		  (int) mail_parameters (NIL,GET_LOCKPROTECTION,NIL))) < 0)
+    return NIL;
   flock (ld,op);		/* get this lock */
   return ld;			/* return locking file descriptor */
 }
@@ -1262,31 +1292,11 @@ char *tenex_file (dst,name)
 	char *dst;
 	char *name;
 {
-  struct passwd *pw;
-  char *s,*t,tmp[MAILTMPLEN];
-  switch (*name) {
-  case '*':			/* bboard? */
-    sprintf (tmp,"~ftp/%s",(name[1] == '/') ? name+2 : name+1);
-    dst = tenex_file (dst,tmp);/* recurse to get result */
-    break;
-  case '/':			/* absolute file path */
-    strcpy (dst,name);		/* copy the mailbox name */
-    break;
-  case '~':			/* home directory */
-    if (name[1] == '/') t = myhomedir ();
-    else {
-      strcpy (tmp,name + 1);	/* copy user name */
-      if (s = strchr (tmp,'/')) *s = '\0';
-      t = ((pw = getpwnam (tmp)) && pw->pw_dir) ? pw->pw_dir : "/NOSUCHUSER";
-    }
-    sprintf (dst,"%s%s",t,(s = strchr (name,'/')) ? s : "");
-    break;
-  default:			/* other name - INBOX becomes mail.TxT */
-    if (!strcmp (ucase (strcpy (dst,name)),"INBOX"))
-      name = tenex_isvalid ("~/mail.TxT",tmp) ? "mail.TxT" : "mail.txt";
-    sprintf (dst,"%s/%s",myhomedir (),name);
-  }
-  return dst;
+  char tmp[MAILTMPLEN];
+  char *s = mailboxfile (dst,name);
+				/* return our standard inbox */
+  return s ? s : mailboxfile (dst,tenex_isvalid ("~/mail.TxT",tmp) ?
+			      "mail.TxT" : "mail.txt");
 }
 
 /* Parse flag list
@@ -1299,10 +1309,10 @@ char *tenex_file (dst,name)
 long tenex_getflags (stream,flag,uf)
 	MAILSTREAM *stream;
 	char *flag;
-	long *uf;
+	unsigned long *uf;
 {
   char key[MAILTMPLEN];
-  char *t,*s;
+  char *t,*s,tmp[MAILTMPLEN];
   short f = 0;
   long i;
   short j;
@@ -1314,10 +1324,10 @@ long tenex_getflags (stream,flag,uf)
       return NIL;
     }
 				/* copy the flag string w/o list construct */
-    strncpy (LOCAL->buf,flag+i,(j = strlen (flag) - (2*i)));
-    LOCAL->buf[j] = '\0';	/* tie off tail */
+    strncpy (tmp,flag+i,(j = strlen (flag) - (2*i)));
+    tmp[j] = '\0';		/* tie off tail */
 				/* make uppercase, find first, parse */
-    if (t = strtok (ucase (LOCAL->buf)," ")) do {
+    if (t = strtok (ucase (tmp)," ")) do {
       i = 0;			/* no flag yet */
 				/* system flag, dispatch on first character */
       if (*t == '\\') switch (*++t) {
@@ -1348,8 +1358,6 @@ long tenex_getflags (stream,flag,uf)
       if (!i) {			/* didn't find a matching flag? */
 	sprintf (key,"Unknown flag: %.80s",t);
 	mm_log (key,ERROR);
-	*uf = NIL;		/* be sure no user flags returned */
-	return NIL;		/* return no system flags */
       }
 				/* parse next flag */
     } while (t = strtok (NIL," "));
@@ -1368,9 +1376,9 @@ long tenex_parse (stream)
 {
   struct stat sbuf;
   MESSAGECACHE *elt = NIL;
-  char *s,*t;
+  char c,*s,*t;
   char tmp[MAILTMPLEN];
-  long i,j,msiz;
+  unsigned long i,j,msiz;
   long curpos = LOCAL->filesize;
   long nmsgs = stream->nmsgs;
   long recent = stream->recent;
@@ -1380,12 +1388,18 @@ long tenex_parse (stream)
     tenex_close (stream);
     return NIL;
   }
-				/* while there is stuff to parse */
-  while (i = sbuf.st_size - curpos) {
+  while (sbuf.st_size - curpos){/* while there is stuff to parse */
 				/* get to that position in the file */
     lseek (LOCAL->fd,curpos,L_SET);
-    if (!((read (LOCAL->fd,tmp,64) > 0) && (s = strchr (tmp,'\012')))) {
-      sprintf (tmp,"Unable to read internal header line at %ld",curpos);
+    if ((i = read (LOCAL->fd,tmp,64)) <= 0) {
+      sprintf (tmp,"Unable to read internal header line at %ld: %s",curpos,
+	       i ? strerror (errno) : "no data read");
+      mm_log (tmp,ERROR);
+      tenex_close (stream);
+      return NIL;
+    }
+    if (!(s = strchr (tmp,'\012'))) {
+      sprintf (tmp,"Unable to find end of line at %ld",curpos);
       mm_log (tmp,ERROR);
       tenex_close (stream);
       return NIL;
@@ -1417,6 +1431,28 @@ long tenex_parse (stream)
       return NIL;
     }
 
+				/* start at first message byte */
+    lseek (LOCAL->fd,curpos + i,L_SET);
+				/* make sure didn't run off end of file */
+    if ((curpos += (msiz + i)) > sbuf.st_size) {
+      mm_log ("Last message runs past end of file",ERROR);
+      tenex_close (stream);
+      return NIL;
+    }
+    for (i = msiz; i;) {	/* for all bytes of the message */
+				/* read a buffer's worth */
+      read (LOCAL->fd,LOCAL->buf,j = min (i,LOCAL->buflen));
+      i -= j;			/* account for having read that much */
+				/* now count the CRLFs */
+      while (j) if (LOCAL->buf[--j] == '\n') elt->rfc822_size++;
+    }
+    c = t[10];			/* remember first system flags byte */
+    t[10] = '\0';		/* tie off flags */
+    j = strtol (t,NIL,8);	/* get user flags value */
+    t[10] = c;			/* restore first system flags byte */
+				/* set up all valid user flags (reversed!) */
+    while (j) if (((i = 29 - find_rightmost_bit (&j)) < NUSERFLAGS) &&
+		  stream->user_flags[i]) elt->user_flags |= 1 << i;
 				/* calculate system flags */
     if ((j = ((t[10]-'0') * 8) + t[11]-'0') & fSEEN) elt->seen = T;
     if (j & fDELETED) elt->deleted = T;
@@ -1427,26 +1463,13 @@ long tenex_parse (stream)
       recent++;			/* count up a new recent message */
 				/* mark it as old */
       tenex_update_status (stream,nmsgs,NIL);
-    }
-				/* start at first message byte */
-    lseek (LOCAL->fd,curpos += i,L_SET);
-    for (i = msiz; i;) {	/* for all bytes of the message */
-				/* read a buffer's worth */
-      read (LOCAL->fd,tmp,j = min (i,(long) MAILTMPLEN));
-      i -= j;			/* account for having read that much */
-				/* now count the CRLFs */
-      while (j) if (tmp[--j] == '\n') elt->rfc822_size++;
-    }
-				/* make sure didn't run off end of file */
-    if ((curpos += msiz) > sbuf.st_size) {
-      mm_log ("Last message runs past end of file",ERROR);
-      tenex_close (stream);
-      return NIL;
     }
   }
   fsync (LOCAL->fd);		/* make sure all the fOLD flags take */
-				/* update parsed file size */
+				/* update parsed file size and time */
   LOCAL->filesize = sbuf.st_size;
+  fstat (LOCAL->fd,&sbuf);	/* get status again to ensure time is right */
+  LOCAL->filetime = sbuf.st_mtime;
   mail_exists (stream,nmsgs);	/* notify upper level of new mailbox size */
   mail_recent (stream,recent);	/* and of change in recent messages */
   return LONGT;			/* return the winnage */
@@ -1464,23 +1487,30 @@ long tenex_copy_messages (stream,mailbox)
 	char *mailbox;
 {
   struct stat sbuf;
+  time_t tp[2];
   MESSAGECACHE *elt;
   unsigned long i,j,k;
+  long ret = LONGT;
   int fd,ld;
-  char lock[MAILTMPLEN];
+  char file[MAILTMPLEN],lock[MAILTMPLEN];
 				/* make sure valid mailbox */
-  if (!tenex_isvalid (mailbox,LOCAL->buf)) {
-    if (errno == ENOENT)	/* failed, was it no such file? */
-      mm_notify (stream,"[TRYCREATE] Must create mailbox before append",NIL);
-    else {
-      sprintf (LOCAL->buf,"Not a Tenex-format mailbox: %s",mailbox);
-      mm_log (LOCAL->buf,ERROR);
-    }
+  if (!tenex_isvalid (mailbox,LOCAL->buf)) switch (errno) {
+  case ENOENT:			/* no such file? */
+    mm_notify (stream,"[TRYCREATE] Must create mailbox before copy",NIL);
+    return NIL;
+  case 0:			/* merely empty file? */
+    break;
+  case EINVAL:
+    sprintf (LOCAL->buf,"Invalid Tenex-format mailbox name: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
+    return NIL;
+  default:
+    sprintf (LOCAL->buf,"Not a Tenex-format mailbox: %s",mailbox);
+    mm_log (LOCAL->buf,ERROR);
     return NIL;
   }
 				/* got file? */
-  if ((fd = open (tenex_file (LOCAL->buf,mailbox),O_RDWR|O_CREAT,
-		  S_IREAD|S_IWRITE)) < 0) {
+  if ((fd=open(tenex_file(file,mailbox),O_RDWR|O_CREAT,S_IREAD|S_IWRITE))<0) {
     sprintf (LOCAL->buf,"Unable to open copy mailbox: %s",strerror (errno));
     mm_log (LOCAL->buf,ERROR);
     return NIL;
@@ -1495,40 +1525,32 @@ long tenex_copy_messages (stream,mailbox)
   lseek (fd,sbuf.st_size,L_SET);/* move to end of file */
 
 				/* for each requested message */
-  for (i = 1; i <= stream->nmsgs; i++)
+  for (i = 1; ret && (i <= stream->nmsgs); i++)
     if ((elt = mail_elt (stream,i))->sequence) {
-      lseek (LOCAL->fd,tenex_header (stream,i,&k),L_SET);
-      j = 0;			/* mark first time through */
+      lseek (LOCAL->fd,elt->data1,L_SET);
+				/* number of bytes to copy */
+      k = (elt->data2 >> 24) + tenex_size (stream,i);
       do {			/* read from source position */
-	if (j) {		/* get another message chunk */
-	  k = min (j,LOCAL->buflen);
-	  read (LOCAL->fd,LOCAL->buf,(unsigned int) k);
-	}
-	else {			/* first time through */
-				/* make a header */
-	  mail_date (LOCAL->buf,elt);
-	  sprintf (LOCAL->buf + strlen (LOCAL->buf),",%d;000000000000\012",
-		   j = tenex_size (stream,i));
-				/* add size of header string */
-	  j += (k = strlen (LOCAL->buf));
-	}
-	if (write (fd,LOCAL->buf,(unsigned int) k) < 0) {
+	j = min (k,LOCAL->buflen);
+	read (LOCAL->fd,LOCAL->buf,(unsigned int) j);
+	if (write (fd,LOCAL->buf,(unsigned int) j) < 0) {
 	  sprintf (LOCAL->buf,"Unable to write message: %s",strerror (errno));
 	  mm_log (LOCAL->buf,ERROR);
 	  ftruncate (fd,sbuf.st_size);
-				/* release exclusive parse/append permission */
-	  tenex_unlock (ld,lock);
-	  close (fd);		/* punt */
-	  mm_nocritical (stream);
-	  return NIL;
+	  j = k;
+	  ret = NIL;		/* note error */
+	  break;
 	}
-      } while (j -= k);		/* until done */
+      } while (k -= j);		/* until done */
     }
   fsync (fd);			/* force out the update */
+  tp[0] = sbuf.st_atime;	/* preserve atime and mtime */
+  tp[1] = sbuf.st_mtime;
+  utime (file,tp);		/* set the times */
   tenex_unlock (ld,lock);	/* release exclusive parse/append permission */
   close (fd);			/* close the file */
   mm_nocritical (stream);	/* release critical */
-  return LONGT;
+  return ret;
 }
 
 /* Tenex get cache element with status updating from file
@@ -1586,6 +1608,7 @@ void tenex_update_status (stream,msgno,syncflag)
 	long syncflag;
 {
   MESSAGECACHE *elt = mail_elt (stream,msgno);
+  struct stat sbuf;
   unsigned long j,k = 0;
   if (!stream->readonly) {	/* not if readonly you don't */
     j = elt->user_flags;	/* get user flags */
@@ -1601,6 +1624,8 @@ void tenex_update_status (stream,msgno,syncflag)
     write (LOCAL->fd,LOCAL->buf,12);
 				/* sync if requested */
     if (syncflag) fsync (LOCAL->fd);
+    fstat (LOCAL->fd,&sbuf);	/* get new write time */
+    LOCAL->filetime = sbuf.st_mtime;
   }
 }
 
@@ -1964,8 +1989,8 @@ search_t tenex_search_string (f,d,n)
       *n = strtol (c+1,&c,10);	/* get its length */
       if (*c++ != '}' || *c++ != '\015' || *c++ != '\012' ||
 	  *n > strlen (*d = c)) return NIL;
-      c[*n] = '\255';		/* write new delimiter */
-      strtok (c,"\255");	/* reset the strtok mechanism */
+      c[*n] = DELIM;		/* write new delimiter */
+      strtok (c,DELMS);		/* reset the strtok mechanism */
       break;
     default:			/* atomic string */
       *n = strlen (*d = strtok (c," "));
