@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	5 November 1990
- * Last Edited:	8 October 1993
+ * Last Edited:	21 August 1994
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1994 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -35,6 +35,7 @@
 
 /* Parameter files */
 
+#include <sys/time.h>		/* must be before osdep.h */
 #include "mail.h"
 #include "osdep.h"
 #include <stdio.h>
@@ -45,13 +46,22 @@
 #include <pwd.h>
 #include <sys/file.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 #include "misc.h"
 
 
+/* Daemon files */
+#define ALERTFILE "/etc/imapd.alert"
+#define ANOFILE "/etc/anonymous.newsgroups"
+
+
 /* Autologout timer */
 #define TIMEOUT 60*30
 
+
+/* Login tries */
+#define LOGTRY 3
+
+
 /* Size of temporary buffers */
 #define TMPLEN 8192
 
@@ -65,15 +75,16 @@
 
 /* Global storage */
 
-char *version = "7.5(72)";	/* version number of this server */
-struct itimerval timer;		/* timeout state */
+char *version = "7.8(88)";	/* version number of this server */
+time_t alerttime = 0;		/* time of last alert */
 int state = LOGIN;		/* server state */
 int mackludge = 0;		/* MacMS kludge */
+int trycreate = 0;		/* saw a trycreate */
 int anonymous = 0;		/* non-zero if anonymous */
+int logtry = LOGTRY;		/* login tries */
 long kodcount = 0;		/* set if KOD has happened already */
 MAILSTREAM *stream = NIL;	/* mailbox stream */
 MAILSTREAM *tstream = NIL;	/* temporary mailbox stream */
-MAILSTREAM *create_policy;	/* current creation policy */
 long nmsgs = 0;			/* number of messages */
 long recent = 0;		/* number of recent messages */
 char *host = NIL;		/* local host name */
@@ -101,22 +112,16 @@ char *badarg = "%s BAD Argument given to %s when none expected\015\012";
 char *badseq = "%s BAD Bogus sequence in %s\015\012";
 char *badatt = "%s BAD Bogus attribute list in %s\015\012";
 char *badlit = "%s BAD Bogus literal count in %s\015\012";
-char *toobig = "* BAD Command line too long\015\012";
+char *toobig = "%s BAD Command line too long\015\012";
 char *nulcmd = "* BAD Null command\015\012";
 char *argrdy = "+ Ready for argument\015\012";
 
-/* Drivers we use */
-
-extern DRIVER bezerkdriver,tenexdriver,imapdriver,newsdriver,nntpdriver,
-  philedriver,dummydriver;
-
-
 /* Function prototypes */
 
 void main  ();
 void clkint  ();
 void kodint  ();
-void dorc  ();
+void slurp  ();
 char *snarf  ();
 void fetch  ();
 void fetch_body  ();
@@ -146,15 +151,10 @@ void main (argc,argv)
   long i;
   char *s,*t,*u,*v,tmp[MAILTMPLEN];
   struct hostent *hst;
+  struct stat sbuf;
   void (*f) () = NIL;
-  mail_link (&tenexdriver);	/* install the Tenex mail driver */
-  mail_link (&bezerkdriver);	/* install the Berkeley mail driver */
-  mail_link (&imapdriver);	/* install the IMAP driver */
-  mail_link (&newsdriver);	/* install the netnews driver */
-  mail_link (&nntpdriver);	/* install the NNTP driver */
-  mail_link (&philedriver);	/* install the file driver */
-  mail_link (&dummydriver);	/* install the dummy driver */
-  create_policy = mailstd_proto;/* default to creating system default */
+#include "linkage.c"
+  openlog ("imapd",LOG_PID,LOG_MAIL);
   gethostname (cmdbuf,TMPLEN-1);/* get local name */
   host = cpystr ((hst = gethostbyname (cmdbuf)) ? hst->h_name : cmdbuf);
   rfc822_date (cmdbuf);		/* get date/time now */
@@ -165,31 +165,26 @@ void main (argc,argv)
     pass = cpystr ("*");	/* and fake password */
     state = SELECT;		/* enter select state */
     t = "PREAUTH";		/* pre-authorized */
+    syslog (LOG_INFO,"Preauthenticated user=%s host=%s",user,
+	    tcp_clienthost (tmp));
   }
   else t = "OK";
   printf ("* %s %s IMAP2bis Service %s at %s\015\012",t,host,version,cmdbuf);
-  dorc ("/etc/imapd.conf");	/* run system init */
-				/* run user's init */
-  if (i) dorc (strcat (strcpy (tmp,myhomedir ()),"/.imaprc"));
   fflush (stdout);		/* dump output buffer */
   signal (SIGALRM,clkint);	/* prepare for clock interrupt */
   signal (SIGUSR2,kodint);	/* prepare for Kiss Of Death */
-				/* initialize timeout interval */
-  timer.it_interval.tv_sec = TIMEOUT;
-  timer.it_interval.tv_usec = 0;
   do {				/* command processing loop */
-				/* get a command under timeout */
-    timer.it_value.tv_sec = TIMEOUT; timer.it_value.tv_usec = 0;
-    setitimer (ITIMER_REAL,&timer,NIL);
-    if (!fgets (cmdbuf,TMPLEN-1,stdin)) _exit (1);
-				/* make sure timeout disabled */
-    timer.it_value.tv_sec = timer.it_value.tv_usec = 0;
-    setitimer (ITIMER_REAL,&timer,NIL);
+    slurp (cmdbuf,TMPLEN-1);	/* slurp command */
 				/* no more last error or literal */
     if (lsterr) fs_give ((void **) &lsterr);
     if (litbuf) fs_give ((void **) &litbuf);
 				/* find end of line */
-    if (!strchr (cmdbuf,'\012')) fputs (toobig,stdout);
+    if (!strchr (cmdbuf,'\012')) {
+      if (t = strchr (cmdbuf,' ')) *t = '\0';
+      printf (toobig,t ? cmdbuf : "*");
+				/* slurp until we find it then */
+      while (getchar () != '\012');
+    }
     else if (!(tag = strtok (cmdbuf," \015\012"))) fputs (nulcmd,stdout);
     else if (!(cmd = strtok (NIL," \015\012")))
       printf ("%s BAD Missing command\015\012",tag);
@@ -207,28 +202,23 @@ void main (argc,argv)
       }
 
 				/* kludge for MacMS */
-      else if (!strcmp (cmd,"VERSION")) {
-				/* single argument */
-	if (!(s = snarf (&arg))) response = misarg;
-	else if (arg) response = badarg;
-	else {
-	  if (!((i = atol (s)) && i > 0 && i <= 4))
-	    response = "%s BAD Unknown version\015\012";
-	  else if (mackludge = (i == 4))
-	    fputs ("* OK [MacMS] The MacMS kludge is enabled\015\012",stdout);
-	}
+      else if ((!strcmp (cmd,"VERSION")) && (s = snarf (&arg)) && (!arg) &&
+	       (i = atol (s)) && (i == 4)) {
+	mackludge = T;
+	fputs ("* OK [MacMS] The MacMS kludges are enabled\015\012",stdout);
       }
       else if (!strcmp (cmd,"NOOP")) {
 	if ((state == OPEN) && !mail_ping (stream)) {
 	  printf ("* BYE %s Fatal mailbox error: %s\015\012",host,
 		  lsterr ? lsterr : "<unknown>");
 	  state = LOGOUT;	/* go away */
+	  syslog (LOG_INFO,"Fatal mailbox error on host %s: %s",
+		  tcp_clienthost (tmp),lsterr ? lsterr : "???");
 	}
       }
       else switch (state) {	/* dispatch depending upon state */
       case LOGIN:		/* waiting to get logged in */
 	if (!strcmp (cmd,"LOGIN")) {
-	  struct stat sbuf;
 	  struct passwd *pwd;
 	  fs_give ((void **) &user);
 	  fs_give ((void **) &pass);
@@ -237,18 +227,32 @@ void main (argc,argv)
 		(pass = cpystr (snarf (&arg))))) response = misarg;
 	  else if (arg) response = badarg;
 				/* see if username and password are OK */
-	  else if (server_login (user,pass,&home,argc,argv)) state = SELECT;
+	  else if (server_login (user,pass,&home,argc,argv)) {
+	    state = SELECT;
+	    syslog (LOG_INFO,"Login user=%s host=%s",user,tcp_clienthost(tmp));
+	  }
 				/* nope, see if we allow anonymous */
-	  else if (!stat ("/etc/anonymous.newsgroups",&sbuf) &&
-		   !strcmp (user,"anonymous") && (pwd = getpwnam ("nobody"))) {
+	  else if (!stat (ANOFILE,&sbuf) && !strcmp (user,"anonymous") &&
+		   (pwd = getpwnam ("nobody"))) {
 	    anonymous = T;	/* note we are anonymous, login as nobody */
 	    setgid (pwd->pw_gid);
 	    setuid (pwd->pw_uid);
 	    state = SELECT;	/* make selected */
-				/* run user's init */
-	    dorc (strcat (strcpy (tmp,myhomedir ()),"/.imaprc"));
+	    syslog (LOG_INFO,"Login anonymous=%s host=%s",pass,
+		    tcp_clienthost(tmp));
+	  }
+	  else {
+	    response = "%s NO Bad %s user name and/or password\015\012";
+	    sleep (3);		/* slow the cracker down */
+	    if (!--logtry) {
+	      fputs ("* BYE Too many login failures\015\012",stdout);
+	      state = LOGOUT;
+	      syslog (LOG_INFO,"Excessive login failures user=%s host=%s",
+		      user,tcp_clienthost (tmp));
+	    }
+	    else syslog (LOG_INFO,"Login failure user=%s host=%s",
+			 user,tcp_clienthost (tmp));
 	  }
-	  else response = "%s NO Bad %s user name and/or password\015\012";
 	}
 	else response = "%s BAD Command unrecognized/login please: %s\015\012";
 	break;
@@ -339,10 +343,13 @@ void main (argc,argv)
 	}
 				/* copy message(s) */
 	else if (!(anonymous || strcmp (cmd,"COPY"))) {
+	  trycreate = NIL;	/* no trycreate status */
 	  if (!((s = snarf (&arg)) && (t = snarf (&arg)))) response = misarg;
 	  else if (arg) response = badarg;
-	  else if (!mail_copy (stream,s,t)) {
-	    response = lose;
+	  else if ((!mail_copy (stream,s,t)) &&
+		   !(mackludge && trycreate &&
+		     (mail_create (stream,t)) && (mail_copy (stream,s,t)))) {
+	    response = lose;	/* in case TRYCREATE failure */
 	    if (!lsterr) lsterr = cpystr ("No such destination mailbox");
 	  }
 	}
@@ -365,16 +372,14 @@ void main (argc,argv)
 		fflush (stdout);/* dump output buffer */
 				/* copy the literal */
 		while (i--) *t++ = getchar ();
-				/* start a timeout */
-		timer.it_value.tv_sec = TIMEOUT; timer.it_value.tv_usec = 0;
-		setitimer (ITIMER_REAL,&timer,NIL);
 				/* get new command tail */
-		if (!fgets (t,TMPLEN - (t-cmdbuf) - 1,stdin)) _exit (1);
-				/* clear timeout */
-		timer.it_value.tv_sec = timer.it_value.tv_usec = 0;
-		setitimer (ITIMER_REAL,&timer,NIL);
+		slurp (t,TMPLEN - (t-cmdbuf) - 1);
 				/* find end of line */
-		if (!(s = strchr (t,'\012'))) response = toobig;
+		if (!(s = strchr (t,'\012'))) {
+				/* slurp until we find it then */
+		  while (getchar () != '\012');
+		  response = toobig;
+		}
 		else {
 		  *s = NIL;	/* tie off line */
 		  if (*--s == '\015') *s = NIL;
@@ -398,8 +403,8 @@ void main (argc,argv)
 
       case SELECT:		/* valid whenever logged in */
 				/* select new mailbox */
-	if ((!(anonymous || strcmp (cmd,"SELECT"))) ||
-	    (!strcmp (cmd,"BBOARD")) || (!strcmp (cmd,"EXAMINE"))) {
+	if ((!(anonymous || (strcmp (cmd,"SELECT") && strcmp (cmd,"EXAMINE"))))
+	    || (!strcmp (cmd,"BBOARD"))) {
 				/* single argument */
 	  if (!(s = snarf (&arg))) response = misarg;
 	  else if (arg) response = badarg;
@@ -426,6 +431,8 @@ void main (argc,argv)
 	      response = stream->readonly ?
 		"%s OK [READ-ONLY] %s completed\015\012" :
 		  "%s OK [READ-WRITE] %s completed\015\012";
+	      if (anonymous) syslog (LOG_INFO,"Selected %s from %s",
+				     stream->mailbox,tcp_clienthost (tmp));
 	    }
 	    else {		/* nuke if still open */
 	      if (stream) mail_close (stream);
@@ -435,15 +442,37 @@ void main (argc,argv)
 	    }
 	  }
 	}
+
 				/* APPEND message to mailbox */
 	else if (!(anonymous || strcmp (cmd,"APPEND"))) {
-	  if (!((s = snarf (&arg)) && (t = snarf (&arg)))) response = misarg;
-	  else if (arg) response = badarg;
-	  else {		/* append the data */
-	    STRING st;
-	    INIT (&st,mail_string,(void *) t,strlen (t));
-	    mail_append (NIL,s,&st);
+	  u = v = NIL;		/* init flags/date */
+				/* parse mailbox name */
+	  if ((s = snarf (&arg)) && arg) {
+	    if (*arg == '(') {	/* parse optional flag list */
+	      u = ++arg;	/* pointer to flag list contents */
+	      while (*arg && (*arg != ')')) arg++;
+	      if (*arg) *arg++ = '\0';
+	      if (*arg == ' ') arg++;
+	    }
+				/* parse optional date */
+	    if (*arg == '"') v = snarf (&arg);
+				/* parse message */
+	    if (!arg || (*arg != '{'))
+	      response = "%s BAD Missing literal in %s\015\012";
+	    else if (!(isdigit (arg[1]) && (i = strtol (arg+1,NIL,10))))
+	      response = "%s NO Empty message to %s\015\012";
+	    else if (!(t = snarf (&arg))) response = misarg;
+	    else if (arg) response = badarg;
+	    else {		/* append the data */
+	      STRING st;
+	      INIT (&st,mail_string,(void *) t,i);
+	      if (!mail_append_full (NIL,s,u,v,&st)) {
+		response = lose;/* in case TRYCREATE failure */
+		if (!lsterr) lsterr = cpystr ("No such destination mailbox");
+	      }
+	    }
 	  }
+	  else response = misarg;
 	}
 
 				/* find mailboxes or bboards */
@@ -524,9 +553,7 @@ void main (argc,argv)
 	else if (!(anonymous || strcmp (cmd,"CREATE"))) {
 	  if (!(s = snarf (&arg))) response = misarg;
 	  else if (arg) response = badarg;
-				/* use specified creation policy */
-	  mail_create (tstream = create_policy,s);
-	  tstream = NIL;	/* drop prototype */
+	  mail_create (NIL,s);	/* create the sucker */
 	}
 				/* delete mailbox */
 	else if (!(anonymous || strcmp (cmd,"DELETE"))) {
@@ -540,23 +567,6 @@ void main (argc,argv)
 	  else if (arg) response = badarg;
 	  else mail_rename (NIL,s,t);
 	}
-
-				/* purge cache */
-	else if (!strcmp (cmd,"PURGE")) {
-	  response = "%s OK PURGE %s is unnecessary with this server\015\012";
-				/* get subcommand */
-	  if (!(arg && (s = strtok (arg," \015\012")) && (cmd = ucase (s))))
-	    response = misarg;	/* missing subcommand */
-	  arg = strtok (NIL,"\015\012");
-	  if (strcmp (cmd,"ALWAYS")) {
-	    if (!(arg && (s = snarf (&arg)))) response = misarg;
-	    else if (arg) response = badarg;
-	    if (strcmp (cmd,"STATUS") && strcmp (cmd,"STRUCTURE") &&
-		strcmp (cmd,"TEXTS"))
-	      response = "%s NO PURGE %s is unknown to this server\015\012";
-	  }
-	  else if (arg) response = badarg;
-	}
 
 	else response = "%s BAD Command unrecognized: %s\015\012";
 	break;
@@ -574,6 +584,21 @@ void main (argc,argv)
 	  fetch_flags (i,NIL);
 	  fputs (")\015\012",stdout);
 	}
+      }
+				/* output any new alerts */
+      if (!stat (ALERTFILE,&sbuf) && (sbuf.st_mtime > alerttime)) {
+	FILE *alf = fopen (ALERTFILE,"r");
+	char c,lc = '\n';
+	if (alf) {		/* only if successful */
+	  do {			/* loop outputting message */
+	    if (lc == '\n') fputs ("* OK [ALERT] ",stdout);
+	    putchar (c);	/* for each character */
+	    lc = c;		/* note previous character */
+	  } while ((c = getc (alf)) != EOF);
+	  fclose (alf);
+	  if (lc != '\n') putchar ('\n');
+	  alerttime = sbuf.st_mtime;
+	}
       }
 				/* get text for alternative win message now */
       if (response == altwin) cmd = lsterr;
@@ -582,6 +607,7 @@ void main (argc,argv)
     }
     fflush (stdout);		/* make sure output blatted */
   } while (state != LOGOUT);	/* until logged out */
+  syslog (LOG_INFO,"Logout from %s",tcp_clienthost (tmp));
   exit (0);			/* all done */
 }
 
@@ -590,7 +616,9 @@ void main (argc,argv)
 
 void clkint ()
 {
+  char tmp[MAILTMPLEN];
   fputs ("* BYE Autologout; idle for too long\015\012",stdout);
+  syslog (LOG_INFO,"Autologout from %s",tcp_clienthost (tmp));
   fflush (stdout);		/* make sure output blatted */
 				/* try to gracefully close the stream */
   if (state == OPEN) mail_close (stream);
@@ -620,38 +648,26 @@ void kodint ()
   }
 }
 
-/* Process rc file
- * Accepts: file name
+/* Slurp a command line
+ * Accepts: buffer pointer
+ *	    buffer size
  */
 
-void dorc (file)
-	char *file;
+void slurp (s,n)
+	char *s;
+	int n;
 {
-  char *s,*t,*k;
-  void *fs;
-  struct stat sbuf;
-  int fd = open (file,O_RDONLY, NIL);
-  if (fd < 0) return;		/* punt if no file */
-  fstat (fd,&sbuf);		/* yes, get size */
-				/* make readin area and read the file */
-  read (fd,(s = (char *) (fs = fs_get (sbuf.st_size + 1))),sbuf.st_size);
-  close (fd);			/* don't need the file any more */
-  s[sbuf.st_size] = '\0';	/* tie it off */
-				/* parse init file */
-  while (*s && (t = strchr (s,'\n'))) {
-    *t = '\0';			/* tie off line, find second space */
-    if ((k = strchr (s,' ')) && (k = strchr (++k,' '))) {
-      *k++ = '\0';		/* tie off two words*/
-      if (!strcmp (lcase (s),"set create-folder-format")) {
-	if (!strcmp (lcase (k),"same-as-inbox")) create_policy = NIL;
-	else if (!strcmp (lcase (k),"system-standard"))
-	  create_policy = mailstd_proto;
-	else printf ("* NO Unknown format in imaprc: %s\015\012",k);
-      }
+  alarm (TIMEOUT);		/* get a command under timeout */
+  errno = 0;			/* clear error */
+  while (!fgets (s,n,stdin)) {	/* read buffer */
+    if (errno==EINTR) errno = 0;/* ignore if some interrupt */
+    else {
+      syslog (LOG_INFO,"Connection broken while reading line from %s",
+	      tcp_clienthost (s));
+      _exit (1);
     }
-    s = ++t;			/* try next line */
   }
-  fs_give (&fs);		/* free buffer */
+  alarm (0);			/* make sure timeout disabled */
 }
 
 /* Snarf an argument
@@ -663,6 +679,7 @@ char *snarf (arg)
 	char **arg;
 {
   long i;
+  char tmp[TMPLEN];
   char *c = *arg;
   char *s = c + 1;
   char *t = NIL;
@@ -672,8 +689,15 @@ char *snarf (arg)
   case ' ':
     return NIL;
   case '"':			/* quoted string */
-    if (!(strchr (s,'"') && (c = strtok (c,"\"")))) return NIL;
-    break;
+    for (c = s; *c != '"'; c++){/* hunt for trailing quote */
+      if (*c == '\\') c++;	/* quote next character */
+      if (!*c) return NIL;	/* unterminated quoted string! */
+    }
+    *c++ = '\0';		/* tie off string */
+				/* update argument pointer */
+    if (*c) *arg = (*c == ' ') ? c + 1 : c;
+    else *arg = NIL;		/* no more arguments */
+    return s;
   case '{':			/* literal string */
     if (isdigit (*s)) {		/* be sure about that */
       i = strtol (s,&t,10);	/* get its length */
@@ -683,19 +707,17 @@ char *snarf (arg)
       fflush (stdout);		/* dump output buffer */
 				/* get a literal buffer */
       c = litbuf = (char *) fs_get (i+1);
-				/* start timeout */
-      timer.it_value.tv_sec = TIMEOUT; timer.it_value.tv_usec = 0;
-      setitimer (ITIMER_REAL,&timer,NIL);
+      alarm (TIMEOUT);		/* start timeout */
       while (i--) *c++ = getchar ();
+      alarm (0);		/* stop timeout */
       *c++ = NIL;		/* make sure string tied off */
       c = litbuf;		/* return value */
     				/* get new command tail */
-      if (!fgets ((*arg = t),TMPLEN - (t - cmdbuf) - 1,stdin)) _exit (1);
-				/* clear timeout */
-      timer.it_value.tv_sec = timer.it_value.tv_usec = 0;
-      setitimer (ITIMER_REAL,&timer,NIL);
+      slurp ((*arg = t),TMPLEN - (t - cmdbuf) - 1);
       if (!strchr (t,'\012')) {	/* have a newline there? */
 	response = toobig;	/* lost it seems */
+				/* slurp until we find it then */
+	while (getchar () != '\012');
 	return NIL;
       }
       break;
@@ -724,6 +746,7 @@ void fetch (s,t)
   char c,*v;
   long i,k;
   BODY *b;
+  int parse_envs = NIL;
   int parse_bodies = NIL;
   void (*f[MAXFETCH]) ();
   char *fa[MAXFETCH];
@@ -748,7 +771,8 @@ void fetch (s,t)
   k = 0;			/* initial index */
   if (s = strtok (t," ")) do {	/* parse attribute list */
     if (*s == 'B' && s[1] == 'O' && s[2] == 'D' && s[3] == 'Y') {
-      parse_bodies = T;		/* we will need to parse bodies */
+				/* we will need to parse bodies */
+      parse_envs = parse_bodies = T;
       switch (s[4]) {
       case '\0':		/* entire body */
 	f[k++] = fetch_body;
@@ -764,7 +788,10 @@ void fetch (s,t)
 	return;
       }
     }
-    else if (!strcmp (s,"ENVELOPE")) f[k++] = fetch_envelope;
+    else if (!strcmp (s,"ENVELOPE")) {
+      parse_envs = T;		/* we will need to parse envelopes */
+      f[k++] = fetch_envelope;
+    }
     else if (!strcmp (s,"FLAGS")) f[k++] = fetch_flags;
     else if (!strcmp (s,"INTERNALDATE")) f[k++] = fetch_internaldate;
     else if (!strcmp (s,"RFC822")) f[k++] = fetch_rfc822;
@@ -788,7 +815,7 @@ void fetch (s,t)
 				/* for each requested message */
   for (i = 1; i <= nmsgs; i++) if (mail_elt (stream,i)->sequence) {
 				/* parse envelope, set body, do warnings */
-    mail_fetchstructure (stream,i,parse_bodies ? &b : NIL);
+    if (parse_envs) mail_fetchstructure (stream,i,parse_bodies ? &b : NIL);
     printf ("* %d FETCH (",i);	/* leader */
     (*f[0]) (i,fa[0]);		/* do first attribute */
     for (k = 1; f[k]; k++) {	/* for each subsequent attribute */
@@ -832,8 +859,7 @@ void fetch_body_part (i,s)
   mail_fetchstructure (stream,i,&body);
   printf ("BODY[%s] ",s);	/* output attribute */
   if (body && (s = mail_fetchbody (stream,i,s,&j))) {
-				/* and literal string */
-    printf ("{%d}\015\012",j);
+    printf ("{%d}\015\012",j);	/* and literal string */
     while (j -= k) k = fwrite (s += k,1,j,stdout);
     changed_flags (i,f);	/* output changed flags */
   }
@@ -926,8 +952,8 @@ void fetch_rfc822 (i,s)
 	char *s;
 {
   int f = mail_elt (stream,i)->seen;
-  printf ("RFC822 {%d}\015\012%s",mail_elt (stream,i)->rfc822_size,
-	  mail_fetchheader (stream,i));
+  printf ("RFC822 {%d}\015\012",mail_elt (stream,i)->rfc822_size);
+  fputs (mail_fetchheader (stream,i),stdout);
   fputs (mail_fetchtext (stream,i),stdout);
   changed_flags (i,f);		/* output changed flags */
 }
@@ -1086,8 +1112,15 @@ void pstring (s)
   char c,*t;
   if (s) {			/* is there a string? */
 				/* must use literal string */
-    if (strpbrk (s,"\012\015\"%{\\")) printf ("{%d}\015\012%s",strlen (s),s);
-    else printf ("\"%s\"",s);	/* may use quoted string */
+    if (strpbrk (s,"\012\015\"%{\\")) {
+      printf ("{%d}\015\012",strlen (s));
+      fputs (s,stdout);		/* don't merge this with the printf() */
+    }
+    else {			/* may use quoted string */
+      putchar ('"');		/* don't even think of merging this into a */
+      fputs (s,stdout);		/*  printf().  Cretin VAXen can't do a */
+      putchar ('"');		/*  printf() of godzilla strings! */
+    }
   }
   else fputs ("NIL",stdout);	/* empty string */
 }
@@ -1247,8 +1280,11 @@ void mm_notify (s,string,errflg)
 	char *string;
 	long errflg;
 {
+  char tmp[MAILTMPLEN];
   if (!tstream || (s != tstream)) switch (errflg) {
   case NIL:			/* information message, set as OK response */
+    tmp[11] = '\0';		/* see if TRYCREATE for MacMS kludge */
+    if (!strcmp (ucase (strncpy (tmp,string,11)),"[TRYCREATE]")) trycreate = T;
   case PARSE:			/* parse glitch, output unsolicited OK */
     printf ("* OK %s\015\012",string);
     break;
@@ -1359,14 +1395,14 @@ long mm_diskerror (s,errcode,serious)
   if (serious) {		/* try your damnest if clobberage likely */
     fputs ("* BAD Retrying to fix probable mailbox damage!\015\012",stdout);
     fflush (stdout);		/* dump output buffer */
-				/* make damn sure timeout disabled */
-    timer.it_value.tv_sec = timer.it_value.tv_usec = 0;
-    setitimer (ITIMER_REAL,&timer,NIL);
+    syslog (LOG_ALERT,"Retrying after disk error %s",strerror (errcode));
+    alarm (0);			/* make damn sure timeout disabled */
     sleep (60);			/* give it some time to clear up */
     return NIL;
   }
 				/* otherwise die before more damage is done */
   printf ("* BYE Aborting due to disk error %s\015\012",strerror (errcode));
+  syslog (LOG_ALERT,"Fatal disk error %s",strerror (errcode));
   return T;
 }
 
