@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: context.c,v 4.25 1993/12/01 02:56:46 mikes Exp $";
+static char rcsid[] = "$Id: context.c,v 4.35 1994/08/13 20:43:05 mikes Exp $";
 #endif
 /*
  * Program:	Mailbox Context Management
@@ -15,7 +15,11 @@ static char rcsid[] = "$Id: context.c,v 4.25 1993/12/01 02:56:46 mikes Exp $";
  * Date:	27 February 1993
  * Last Edited:	27 February 1993
  *
- * Copyright 1993 by the University of Washington
+ * Copyright 1993-1994 by the University of Washington
+ * 
+ *
+ * USENET News reading additions in part by L Lundblade / NorthWestNet, 1993
+ * lgl@nwnet.net
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -45,10 +49,11 @@ char *current_context = NIL;	/* current context for FIND */
  * Accepts: Mailbox name as FQN
  */
 
-void mm_mailbox (string)
-char *string;
+void
+mm_mailbox (string)
+    char *string;
 {
-  context_fqn_to_an (context_mailbox,current_context,string);
+    context_fqn_to_an (context_mailbox,current_context,string);
 }
 
 
@@ -56,10 +61,11 @@ char *string;
  * Accepts: BBoard name as FQN
  */
 
-void mm_bboard (string)
-char *string;
+void
+mm_bboard (string)
+    char *string;
 {
-  context_fqn_to_an (context_bboard,current_context,string);
+    context_fqn_to_an (context_bboard,current_context,string);
 }
 
 /* Convert fully-qualified name to ambiguous name and send to function
@@ -68,9 +74,10 @@ char *string;
  *	    fully-qualified name
  */
 
-void context_fqn_to_an (f, context_string, fqn)
-find_return f;
-char *context_string,*fqn;
+void
+context_fqn_to_an (f, context_string, fqn)
+    find_return f;
+    char *context_string,*fqn;
 {
   char tmp[MAILTMPLEN];
   char *ret = fqn;
@@ -85,9 +92,13 @@ char *context_string,*fqn;
       context_digest(context_string, context, NULL, NULL, NULL) == NULL){
     delimiter = strstr(context, "%s");
 				/* determine prefix */
-    if (prefix = (pfxlen = delimiter - context) ? strncpy (tmp,context,pfxlen)
-	                                          : NIL)
-	prefix[pfxlen] = '\0';
+    pfxlen = delimiter - context;
+    if (pfxlen) {
+      strncpy (tmp,context,pfxlen);
+      prefix = tmp;
+      prefix[pfxlen] = '\0';
+    } else
+      prefix = NIL;
 				/* determine suffix */
     if (suffix = delimiter[2] ? delimiter+2 : NIL) sfxlen = strlen (suffix);
 
@@ -144,7 +155,7 @@ char *context_string,*fqn;
  */
 char *
 context_digest (context, scontext, host, rcontext, view)
-char *context, *scontext, *host, *rcontext, *view;
+    char *context, *scontext, *host, *rcontext, *view;
 {
     char *p, *viewp = view, tmp[MAILTMPLEN];
     int   i = 0;
@@ -254,7 +265,7 @@ char *context, *scontext, *host, *rcontext, *view;
  */
 void
 context_apply(b, c, n)
-char *b, *c, *n;
+    char *b, *c, *n;
 {
     char tmp[MAILTMPLEN];
 
@@ -270,25 +281,32 @@ char *b, *c, *n;
  * Returns: T if ambiguous, NIL if fully-qualified
  */
 
-short context_isambig (s)
-char *s;
+int
+context_isambig (s)
+    char *s;
 {
 				/* first character after possible bboard * */
-  char c = (*s == '*') ? s[1] : s[0];
-  if(c == '\0' && *s == '*')	/* special case! */
-    return(1);
+    char c = (*s == '*') ? *(s+1) : *s;
+    if(c == '\0' && *s == '*')	/* special case! */
+      return(1);
 #ifdef	DOS
-  else if (isalpha(*s) && *(s+1) == ':')
-    return(0);
+    else if (c != '{' && isalpha(*s) && *(s+1) == ':')
+      return(0);
 #endif
-  else
-    /*
-     * return ambiguous if alphanumeric OR something that starts
-     * with "dot".  Now, if this seems to indicate that the client
-     * assumes knowledge of the server's naming syntax, well, it does.
-     * THIS MUST BE REMOVED!!!!
-     */
-    return (short) (c == '.' || isalnum (c));
+    else
+      /*
+       * Since we can't know how the server handles foldernames or
+       * heirarchy, treat only what we know as being absolute folder
+       * specifiers.  We know '{' means a c-client remote folder, and
+       * that '/' breaks out of the current context under UNIX, and we're
+       * guessing that '[' and ':' are special under VMS (for which
+       * a couple of imapd's are known).  If you know this guess to be
+       * wrong or know of other servers, please send us diffs for this...
+       * 
+       * The good news is, THIS MUST AND WILL BE REMOVED
+       * WITH THE COMING IMAP4 SUPPORT!!!!
+       */
+      return(!(c == '{' || c == '/' || c == ':' || c == '[' || c == '#'));
 }
 
 /* Context Manager find list of subscribed mailboxes
@@ -297,8 +315,10 @@ char *s;
  *	    pattern to search
  */
 
-void context_find (context, stream, pat)
-char *context;MAILSTREAM *stream;char *pat;
+void
+context_find (context, stream, pat)
+    char       *context;
+    MAILSTREAM *stream;char *pat;
 {
   char tmp[MAILTMPLEN];
 				/* must be within context */
@@ -322,8 +342,11 @@ char *context;MAILSTREAM *stream;char *pat;
  *	    pattern to search
  */
 
-void context_find_bboards (context, stream, pat)
-char *context;MAILSTREAM *stream;char *pat;
+void
+context_find_bboards (context, stream, pat)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *pat;
 {
   char tmp[MAILTMPLEN];
 				/* must be within context */
@@ -346,8 +369,11 @@ char *context;MAILSTREAM *stream;char *pat;
  *	    pattern to search
  */
 
-void context_find_all (context, stream, pat)
-char *context;MAILSTREAM *stream;char *pat;
+void
+context_find_all (context, stream, pat)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *pat;
 {
   char tmp[MAILTMPLEN];
 				/* must be within context */
@@ -371,8 +397,11 @@ char *context;MAILSTREAM *stream;char *pat;
  *	    pattern to search
  */
 
-void context_find_all_bboard (context, stream, pat)
-char *context;MAILSTREAM *stream;char *pat;
+void
+context_find_all_bboard (context, stream, pat)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *pat;
 {
   char tmp[MAILTMPLEN];
 				/* must be within context */
@@ -388,6 +417,35 @@ char *context;MAILSTREAM *stream;char *pat;
 				/* try to do the find */
   mail_find_all_bboard (stream,tmp);
 }
+#ifdef NEWBB
+/* Context Manager find list of new bboards
+ * Accepts: context
+ *	    mail stream
+ *	    pattern to search
+ */
+
+void
+context_find_new_bboard (context, stream, pat, lasttime)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *pat, *lasttime;
+{
+  char tmp[MAILTMPLEN];
+				/* must be within context */
+  if (!context_isambig (pat)) {
+    sprintf (tmp,"Find of bboard outside context: %s",pat);
+    mm_log (tmp,WARN);
+    strcpy(tmp, pat);		/* allow find for now */
+  }
+  else{
+    current_context = context;	/* note current context */
+    context_apply (tmp,context,pat); /* build fully-qualified name */
+  }
+				/* try to do the find */
+  nntp_find_new_bboard (stream,tmp, lasttime);
+}
+
+#endif
 
 /* Context Manager create mailbox
  * Accepts: context
@@ -396,8 +454,11 @@ char *context;MAILSTREAM *stream;char *pat;
  * Returns: T on success, NIL on failure
  */
 
-long context_create (context, stream, mailbox)
-char *context;MAILSTREAM *stream;char *mailbox;
+long
+context_create (context, stream, mailbox)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *mailbox;
 {
   char tmp[MAILTMPLEN];		/* must be within context */
 
@@ -440,8 +501,12 @@ char *context;MAILSTREAM *stream;char *mailbox;
  * Returns: stream to use on success, NIL on failure
  */
 
-MAILSTREAM *context_open (context, old, name, opt)
-char *context;MAILSTREAM *old;char *name;long opt;
+MAILSTREAM *
+context_open (context, old, name, opt)
+    char       *context;
+    MAILSTREAM *old;
+    char       *name;
+    long	opt;
 {
   char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
   char host[MAILTMPLEN];
@@ -466,8 +531,11 @@ char *context;MAILSTREAM *old;char *name;long opt;
  * Returns: T on success, NIL on failure
  */
 
-long context_rename (context, stream, old, new)
-char *context;MAILSTREAM *stream;char *old, *new;
+long
+context_rename (context, stream, old, new)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *old, *new;
 {
   char *s,*t,tmp[MAILTMPLEN],tmp2[MAILTMPLEN];
 	/* build FQN from ambiguous name */
@@ -496,8 +564,11 @@ char *context;MAILSTREAM *stream;char *old, *new;
  * Returns: T on success, NIL on failure
  */
 
-long context_delete (context, stream, name)
-char *context;MAILSTREAM *stream;char *name;
+long
+context_delete (context, stream, name)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *name;
 {
   char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
   if (context_isambig (name)) context_apply (s = tmp,context,name);
@@ -519,8 +590,12 @@ char *context;MAILSTREAM *stream;char *name;
  * Returns: T on success, NIL on failure
  */
 
-long context_append (context, stream, name, msg)
-char *context;MAILSTREAM *stream;char *name;STRING *msg;
+long
+context_append (context, stream, name, msg)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *name;
+    STRING     *msg;
 {
   char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
   if (context_isambig (name)) context_apply (s = tmp,context,name);
@@ -534,6 +609,61 @@ char *context;MAILSTREAM *stream;char *name;STRING *msg;
 }
 
 
+/* Context Manager append message string with flags
+ * Accepts: context
+ *	    mail stream
+ *	    destination mailbox
+ *          flags to assign message being appended
+ *          date of message being appended
+ *	    stringstruct of message to append
+ * Returns: T on success, NIL on failure
+ */
+
+long context_append_full (context, stream, name, flags, date, msg)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *name, *flags, *date;
+    STRING     *msg;
+{
+  char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
+  if (context_isambig (name)) context_apply (s = tmp,context,name);
+  else if(name[0] != '{' && context[0] == '{' && (s = strindex(context, '}'))){
+      strncpy(tmp, context, (s - context) + 1);
+      strcpy(&tmp[(s - context)+1], name);
+      s = tmp;
+  }
+  else s = name;		/* fully-qualified name */
+  return mail_append_full (stream,s,flags,date,msg);
+}
+
+
+/* Mail copy message(s)
+ * Accepts: context
+ *	    mail stream
+ *	    sequence
+ *	    destination mailbox
+ */
+
+long context_copy (context, stream, sequence, name)
+    char *context;
+    MAILSTREAM *stream;
+    char *sequence;
+    char *name;
+{
+  char *s,tmp[MAILTMPLEN];		/* build FQN from ambiguous name */
+  if (context_isambig (name)) context_apply (tmp,context,name);
+  else if(name[0] != '{' && context[0] == '{' && (s = strindex(context, '}'))){
+      strncpy(tmp, context, (s - context) + 1);
+      strcpy(&tmp[(s - context)+1], name);
+  }
+  else strcpy(tmp,name);		/* fully-qualified name */
+  if((*(s=tmp) == '{' || (*s == '*' && *(s+1) == '{')) && (s=strindex(s, '}')))
+    s++;				/* strip imap host name ? */
+
+  return (s ? mail_copy(stream, sequence, s) : 0L);
+}
+
+
 /* Context Manager subscribe
  * Accepts: context
  *	    mail stream
@@ -541,8 +671,11 @@ char *context;MAILSTREAM *stream;char *name;STRING *msg;
  * Returns: T on success, NIL on failure
  */
 
-long context_subscribe (context, stream, name)
-char *context;MAILSTREAM *stream;char *name;
+long
+context_subscribe (context, stream, name)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *name;
 {
   char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
   if (context_isambig (name)) context_apply (s = tmp,context,name);
@@ -558,8 +691,11 @@ char *context;MAILSTREAM *stream;char *name;
  * Returns: T on success, NIL on failure
  */
 
-long context_unsubscribe (context, stream, name)
-char *context;MAILSTREAM *stream;char *name;
+long
+context_unsubscribe (context, stream, name)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *name;
 {
   char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
   if (context_isambig (name)) context_apply (s = tmp,context,name);
@@ -578,7 +714,9 @@ char *context;MAILSTREAM *stream;char *name;
  */
 MAILSTREAM *
 context_same_stream(context, name, stream)
-char *context;MAILSTREAM *stream; char *name;
+    char       *context;
+    MAILSTREAM *stream;
+    char       *name;
 {
   extern MAILSTREAM *same_stream();
   char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
@@ -591,3 +729,118 @@ char *context;MAILSTREAM *stream; char *name;
   else s = name;		/* fully-qualified name */
   return same_stream (s,stream);
 }
+
+
+/*
+ * Context manager driver compatibility test
+ * Accepts: context
+ *	    mail name
+ *	    mailbox name
+ *	    mail stream to test against mailbox name
+ * Returns: stream if drivers are the same, else NIL
+ */
+MAILSTREAM *
+context_same_driver(context, name, stream)
+    char       *context;
+    MAILSTREAM *stream;
+    char       *name;
+{
+  MAILSTREAM *proto_stream;
+  int rv;
+  char *s,tmp[MAILTMPLEN];	/* build FQN from ambiguous name */
+  if (context_isambig (name)) context_apply (s = tmp,context,name);
+  else if(name[0] != '{' && context[0] == '{' && (s = strindex(context, '}'))){
+      strncpy(tmp, context, (s - context) + 1);
+      strcpy(&tmp[(s - context)+1], name);
+      s = tmp;
+  }
+  else s = name;		/* fully-qualified name */
+
+  /* IF we're given a stream to compare *and* it's not remote *and*
+   * the given mailbox name isn't remote *and* we know the driver
+   * the given mailbox will be in *and* the drivers are the same
+   * return the given stream.
+   *
+   * Also, turn off any mm_log errors that might bubble up from 
+   * mail_open because anything other than "doesn't exist" will
+   * get encountered during whatever action follows this call...
+   */
+  ps_global->try_to_create = 1;
+  rv = (stream
+	&& !((*stream->mailbox == '{'
+	      || (*stream->mailbox == '*' && *(stream->mailbox + 1) == '{'))
+	     || (*s == '{' || (*s == '*' && *(s+1) == '{')))
+	&& ((proto_stream = mail_open(NIL,s,OP_PROTOTYPE))
+	    || (proto_stream = default_driver(context)))
+	&& !strcmp(stream->dtb->name, proto_stream->dtb->name));
+  ps_global->try_to_create = 0;
+
+  return(rv ? stream : NULL);
+}
+
+
+#ifdef NEWBB
+
+/*  This is code for inclusion in the c-client to support the command
+    to discover all the new newly created news groups, since the last
+    time the subscription screen was viewed. 
+
+    Some of this code goes un nntpunx.c and one line in nntpunx.h
+
+    It's not clear whether or not this code will be included as part
+    of the official c-client distribution so it's included here.
+*/
+
+#ifdef INCLUDE_THIS_CODE_AT_THE_END_OF_NNTPUNX_C
+
+/* NNTP mail find list of all bboards
+ * Accepts: mail stream
+ *	    pattern to search
+ *          last time that new groups were checked for in NNTP format
+ */
+
+void nntp_find_new_bboard (stream, pat, lasttime)
+    MAILSTREAM *stream;
+    char       *pat, *lasttime;
+{
+  char *s,*t,*bbd,*patx,tmp[MAILTMPLEN], checktime[20];
+				/* use .newsrc if a stream given */
+  if (stream && LOCAL && LOCAL->nntpstream) {
+				/* begin with a host specification? */
+    if (((*pat == '{') || ((*pat == '*') && (pat[1] == '{'))) &&
+	(t = strchr (pat,'}')) && *(patx = ++t)) {
+      if (*pat == '*') pat++;	/* yes, skip leading * (old Pine behavior) */
+      strcpy (tmp,pat);		/* copy host name */
+      bbd = tmp + (patx - pat);	/* where we write the bboards */
+    }
+    else {			/* no host specification */
+      bbd = tmp;		/* no prefix */
+      patx = pat;		/* use entire specification */
+    }
+				/* ask server for all active newsgroups */
+    if (!(smtp_send (LOCAL->nntpstream,"NEWGROUPS",lasttime) == NNTPNEWLIST))
+      return;
+				/* process data until we see final dot */
+    while ((s = tcp_getline (LOCAL->nntpstream->tcpstream)) && *s != '.') {
+				/* tie off after newsgroup name */
+      if (t = strchr (s,' ')) *t = '\0';
+      if (pmatch (s,patx)) {	/* report to main program if have match */
+	strcpy (bbd,s);		/* write newsgroup name after prefix */
+	mm_bboard(tmp);
+      }
+      fs_give ((void **) &s);	/* clean up */
+    }
+  }
+}
+
+     
+#endif
+
+
+#ifdef INCLUDE_THIS_CODE_NEAR_THE_TOP_OF_NNTPUNX_H
+
+#define NNTPNEWLIST (long) 231  /* NNTP list of new news groups */
+
+#endif
+
+#endif /* NEWBB */
