@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: filter.c,v 4.20 1993/10/30 00:27:42 mikes Exp $";
+static char rcsid[] = "$Id: filter.c,v 4.29 1994/08/18 17:39:51 mikes Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -16,7 +16,7 @@ static char rcsid[] = "$Id: filter.c,v 4.20 1993/10/30 00:27:42 mikes Exp $";
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1993  University of Washington
+   Copyright 1989-1994  University of Washington
 
     Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee to the University of
@@ -76,7 +76,6 @@ static char rcsid[] = "$Id: filter.c,v 4.20 1993/10/30 00:27:42 mikes Exp $";
 
 
 #include "headers.h"
-#include "../pico/pico.h"
 
 
 #ifdef ANSI			/* INTERNAL PROTOTYPES */
@@ -158,11 +157,11 @@ so_get(source, name, rtype)
 	 * networked environments.
 	 */
 	source   = TmpFileStar;
-	so->name = (char *)temp_nam(NULL, "pi");
+	so->name = temp_nam(NULL, "pi");
     }
 #else
     else if(source == TmpFileStar)		/* make one up! */
-      so->name = cpystr(temp_nam(NULL, "pine-tmp"));
+      so->name = temp_nam(NULL, "pine-tmp");
 #endif
 
     so->src = source;
@@ -181,21 +180,23 @@ so_get(source, name, rtype)
 	 */
 	if(so->name){
 	    if((so->txt = (void *)fopen(so->name, type)) == NULL){
+		dprint(1, (debugfile, "so_get error: %s : %s", so->name,
+			   error_description(errno)));
 		fs_give((void **)&so->name);
 		fs_give((void **)&so); 		/* so freed & set to NULL */
-		q_status_message(1, 3, 5,"ERROR creating message file");
 	    }
 	}
 	else{
 	    if((so->txt = (void *)create_tmpfile()) == NULL){
-		q_status_message(1, 3, 5,"ERROR creating message temp file");
+		dprint(1, (debugfile, "so_get error: tmpfile : %s",
+			   error_description(errno)));
 		fs_give((void **)&so);		/* so freed & set to NULL */
 	    }
 	}
     }
     else if(so->src == PicoText){
 	if((so->txt = pico_get()) == NULL){
-	    q_status_message(1, 3, 5,"ERROR allocating composer text space");
+	    dprint(1, (debugfile, "so_get error: alloc of pico text space"));
 	    if(so->name)
 	      fs_give((void **)&so->name);
 	    fs_give((void **)&so);		/* so freed & set to NULL */
@@ -256,15 +257,15 @@ so_writec(c, so)
     unsigned char ch = (unsigned char) c;
 
     if(so->src == CharStar){
-	if((so->used + 2) >= so->size){
+	if((so->used.l + 2) >= so->size){
 	    so->size += MSIZE_INC;		/* need to resize! */
 	    fs_resize(&so->txt, (size_t)so->size * sizeof(char));
-	    so->txtp = (unsigned char *)so->txt + so->used;
-	    memset(so->txtp, 0, (size_t)(so->size - so->used) * sizeof(char));
+	    so->txtp = (unsigned char *)so->txt + so->used.l;
+	    memset(so->txtp, 0, (size_t)(so->size - so->used.l)*sizeof(char));
 	}
 
 	*(so->txtp)++ = ch;
-	so->used++;
+	so->used.l++;
 	return(1);
     }
     else if(so->src == PicoText){
@@ -274,7 +275,7 @@ so_writec(c, so)
 	if(!so->txt && !so_reaquire(so))
 	  return(0);
 
-	return(fwrite(&ch, sizeof(char), (size_t)1, (FILE *)so->txt));
+	return(fwrite(&ch, sizeof(unsigned char), (size_t)1, (FILE *)so->txt));
     }
 }
 
@@ -290,7 +291,7 @@ so_readc(c, so)
     STORE_S       *so;
 {
     if(so->src == CharStar)
-      return((so->txtp - (unsigned char *)so->txt) < so->used ? 
+      return((so->txtp - (unsigned char *)so->txt) < so->used.l ? 
 		*c = *(so->txtp)++, 1 : 0);
     else if(so->src == PicoText)
       return(pico_readc(so->txt, c));
@@ -318,19 +319,19 @@ so_puts(so, s)
     if(so->src == CharStar){
 	int l = strlen(s);
 
-	if((so->used + l + 1) >= so->size){
-	    while(so->size <= (so->used + l + 1))
+	if((so->used.l + l + 1) >= so->size){
+	    while(so->size <= (so->used.l + l + 1))
 	      so->size += MSIZE_INC;		/* need to resize! */
 
 	    fs_resize(&so->txt, (size_t)so->size * sizeof(char));
-	    so->txtp = (unsigned char *)so->txt + so->used;
-	    memset(so->txtp, 0, (size_t)(so->size - so->used) * sizeof(char));
+	    so->txtp = (unsigned char *)so->txt + so->used.l;
+	    memset(so->txtp, 0, (size_t)(so->size - so->used.l)*sizeof(char));
 	}
 
 	while(*s != '\0')
 	  *(so->txtp)++ = *s++;
 
-	so->used += l;
+	so->used.l += l;
 	return(1);
     }
     else if(so->src == PicoText){
@@ -341,7 +342,8 @@ so_puts(so, s)
 	  return(0);
 
 	/* fwrite not necessary, but it helps */
-	return(fwrite(s, strlen(s)*sizeof(char), (size_t)1, (FILE *)so->txt));
+	return(*s ? fwrite(s,strlen(s)*sizeof(char),(size_t)1,(FILE *)so->txt)
+		  : 1);
     }
 }
 
@@ -358,26 +360,22 @@ int      orig;
     if(so->src == CharStar){
 	switch(orig){
 	    case 0 :				/* SEEK_SET */
-	      return(pos < so->used ? 
+	      return(pos < so->used.l ? 
 		so->txtp = (unsigned char *)so->txt + pos, 0 : -1);
 	    case 1 :				/* SEEK_CUR */
-	      return(pos < (so->used - (so->txtp-(unsigned char *)so->txt)) ? 
+	      return(pos < (so->used.l-(so->txtp-(unsigned char *)so->txt)) ? 
 		so->txtp += pos, 0 : -1);
 	    case 2 :				/* SEEK_END */
-	      return(pos < so->used ? 
-		so->txtp = (unsigned char *)so->txt+(so->used-pos), 0 : -1);
+	      return(pos < so->used.l ? 
+		so->txtp = (unsigned char *)so->txt+(so->used.l-pos), 0 : -1);
 	    default :
 	      return(-1);
 	}
     }
     else if(so->src == PicoText)
       return(pico_seek(so->txt, pos, orig));
-    else{			/* FileStar or TmpFileStar */
-	if(!so->txt && !so_reaquire(so))
-	  return(0);
-
-	return(fseek((FILE *)so->txt, pos, orig));
-    }
+    else			/* FileStar or TmpFileStar */
+      return((so->txt || so_reaquire(so)) && fseek((FILE *)so->txt,pos,orig));
 }
 
 
@@ -393,7 +391,7 @@ so_release(so)
 STORE_S *so;
 {
     if(so->txt && so->name && (so->src == FileStar || so->src == TmpFileStar)){
-	if(fget_pos((FILE *)so->txt, (fpos_t *)&(so->used)) == 0){
+	if(fget_pos((FILE *)so->txt, (fpos_t *)&(so->used.f)) == 0){
 	    fclose((FILE *)so->txt);		/* free the handle! */
 	    so->txt = NULL;
 	}
@@ -427,7 +425,7 @@ STORE_S *so;
 				error_description(errno));
 	    rv = 0;
 	}
-	else if(fset_pos((FILE *)so->txt, (fpos_t *)&(so->used))){
+	else if(fset_pos((FILE *)so->txt, (fpos_t *)&(so->used.f))){
 	    q_status_message2(1, 3, 5, "ERROR positioning in %s : %s", 
 				so->name, error_description(errno));
 	    rv = 0;
@@ -475,8 +473,9 @@ static	jmp_buf   gf_error_state;
 #define	HEX	2
 #define	WSPACE	3
 #define	CCR	4
-#define	CESCP	5
-#define	TOKEN	6
+#define	CLF	5
+#define	CESCP	6
+#define	TOKEN	7
 
 
 /*
@@ -609,7 +608,8 @@ int
 gf_fwritec(c)
     int c;
 {
-    return(fwrite(&c, sizeof(unsigned char), (size_t)1, gf_out.file));
+    unsigned char ch = (unsigned char)c;
+    return(fwrite(&ch, sizeof(unsigned char), (size_t)1, gf_out.file));
 }
 
 
@@ -1416,110 +1416,161 @@ gf_rich2plain_opt(plain)
 
 
 
-#ifdef notdef  /* not using this anymore */
 /*
- * ISO-ESCAPE CODE matching filter
+ * ENRICHED-TO-PLAIN text filter
  */
 
-/*
- * options to be used by iso_filter (NOTE: if this filter is ever 
- * used more than once in a pipe, all instances will have the same
- * option value)
- */
-static int gf_iso_match = 0;
+static int gf_enriched_plain = 0;
 
 
-/*
- * filter's out any non-displayable characters except those associated
- * with known charset shifts
- */
+/*----------------------------------------------------------------------
+      enriched text to plain text filter (ala rfc1523)
+    
+ Args: gc, pc -- functions to read and write chars with
+       width  --
+       plain  -- 
+
+  This basically removes all enriched formatting. A cute hack is used 
+  to get bold and underlining to work.
+
+  Further work could be done to handle things like centering and right
+  and left flush, but then it could no longer be done in place. This
+  operates on text *with* CRLF's.
+
+  WARNING: does not wrap lines!
+ ----*/
 void
-gf_iso_filter(f, c, flg)
+gf_enriched2plain(f, c, flg)
     FILTER_S *f;
     int       c, flg;
 {
-    register char *p;
 
     if(flg == GF_DATA){
 	switch(f->f1){
-	  case CESCP :			/* process chars after escape */
-	    if(c == '\033'){		/* dump whats there and start over */
-		for(p = f->line; p < f->linep; p++)
-		  (*f->next->f)(f->next, (*p == '\033') ? '_' : *p, GF_DATA);
+	  case TOKEN :			/* collect a richtext token */
+	    if(c == '>'){		/* what should we do with it? */
+		f->f1       = DFL;	/* return to default next time */
+		*(f->linep) = '\0';	/* cap off token */
+		if(!strcmp("param", f->line)){
+		    (f->f2)++;
+		}
+		else if(!strcmp("/param", f->line)){
+		    f->f2 = 0;
+		}
+		else if(!gf_enriched_plain){
+		    /* Following is a cute hack or two to get 
+		       bold and underline on the screen. 
+		       See Putline0n() where these codes are
+		       interpreted */
+		    if(!strcmp(f->line, "bold")) {
+			(*f->next->f)(f->next, '\01', GF_DATA);
+		    } else if(!strcmp(f->line, "/bold")) {
+			(*f->next->f)(f->next, '\02', GF_DATA);
+		    } else if(!strcmp(f->line, "italic")) {
+			(*f->next->f)(f->next, '\03', GF_DATA);
+		    } else if(!strcmp(f->line, "/italic")) {
+			(*f->next->f)(f->next, '\04', GF_DATA);
+		    } else if(!strcmp(f->line, "underline")) {
+			(*f->next->f)(f->next,'\03', GF_DATA);
+		    } else if(!strcmp(f->line, "/underline")) {
+			(*f->next->f)(f->next, '\04', GF_DATA);
+		    } 
+		}
+		/* else we just ignore the token! */
 
-		f->linep = f->line;
-		*(f->linep)++ = '\033';
+		f->linep = f->line;	/* reset token buffer */
 	    }
-	    else {			/* match or not too big ? */
-		*(f->linep)++ = c;
-		*(f->linep) = '\0';
-		if(match_iso2022(f->line)){
+	    else if(c == '<'){		/* literal '<'? */
+		if(f->linep == f->line){
+		    (*f->next->f)(f->next, '<', GF_DATA);
 		    f->f1 = DFL;
-		    for(p = f->line; p < f->linep; p++)
-		      (*f->next->f)(f->next, *p, GF_DATA);
-
-		    f->linep = f->line;
 		}
 		else{
-		    if(f->linep - f->line > 6){
-			/* search is over for now */
-			f->f1 = DFL;
-			for(p = f->line; p < f->linep; p++)
-			  (*f->next->f)(f->next, 
-					(*p == '\033') ? '_' : *p, 
-					GF_DATA);
-
-			f->linep = f->line;
-		    }
-		    /* else we wait for the next character */
+		    fs_give((void **)&(f->line));
+		    gf_error("Malformed Enriched text: unexpected '<'");
 		}
 	    }
-	    
+	    else{			/* add char to token */
+		if(f->linep - f->line > 60){ /* rfc1523 says 60 MAX! */
+		    fs_give((void **)&(f->line));
+		    gf_error("Malformed Enriched text: token too long");
+		}
+		
+		*(f->linep)++ = isupper(c) ? c - 'A' + 'a' : c;
+	    }
 	    break;
 
-	  case DFL    :			/* just */
-	  default     :
-	    if(gf_iso_match == 0 || c <= 0x80) {
-		if(c == '\033'){
-		    f->f1 = CESCP;
-		    f->linep = f->line;
-		    *(f->linep)++ = '\033';
-		}
-		else
-		  (*f->next->f)(f->next, c, GF_DATA);
-	    } 
-	    else 
-	      (*f->next->f)(f->next, '_', GF_DATA);
+	  case CCR   :
+	    if(c != '\012'){		/* treat as single space?    */
+		f->f1 = DFL;		/* lone cr? */
+		f->f2 = 0;
+		(*f->next->f)(f->next, '\015', GF_DATA);
+		goto df;
+	    }
 
+	    f->f1 = CLF;
+	    break;
+
+	  case CLF   :
+	    if(c == '\015'){		/* treat as single space?    */
+		f->f1 = CCR;		/* repeat crlf's mean real newlines */
+		f->f2 = 1;
+		(*f->next->f)(f->next, '\r', GF_DATA);
+		(*f->next->f)(f->next, '\n', GF_DATA);
+		break;
+	    }
+	    else{
+		f->f1 = DFL;
+		if(!f->f2)
+		  (*f->next->f)(f->next, ' ', GF_DATA);
+
+		f->f2 = 0;
+	    }
+
+	    /* fall thru to take care of 'c' */
+
+	  case DFL   :
+	  default :
+	  df : 
+	    if(c == '<')
+	      f->f1 = TOKEN;
+	    else if(c == '\015')
+	      f->f1 = CCR;
+	    else if(!f->f2)		/* not in param! */
+	      (*f->next->f)(f->next, c, GF_DATA);
+	    
 	    break;
 	}
     }
     else if(flg == GF_EOD){
-	f->n = f->linep - f->line;
-	while(f->n)
-	  (*f->next->f)(f->next, f->line[(f->n)--], GF_DATA);
+	if(f->f1 = (f->linep != f->line)){
+	    /* incomplete token!! */
+	    gf_error("Incomplete token in richtext");
+	}
 
 	fs_give((void **)&(f->line));
+
 	(*f->next->f)(f->next, 0, GF_EOD);
     }
     else if(flg == GF_RESET){
-	f->f1 = DFL;
-	f->linep = f->line = (char *)fs_get(10 * sizeof(char));
+	f->f1 = DFL;			/* state */
+	f->f2 = 0;			/* set means we're in a comment */
+	f->linep = f->line = (char *)fs_get(65 * sizeof(char));
     }
 
 }
 
 
 /*
- * function to call from other modules to set options for iso_filter
+ * function called from the outside to set
+ * richtext filter's options
  */
 void
-gf_iso_filter_opt(match)
-    int match;
+gf_enriched2plain_opt(plain)
+    int plain;
 {
-    gf_iso_match = match;
+    gf_enriched_plain = plain;
 }
-#endif  /* not using this anymore */
 
 
 
