@@ -1,4 +1,4 @@
-# $Id: makefile.hpp,v 4.8 1993/07/30 22:43:16 hubert Exp $
+# $Id: makefile.hpp,v 4.17 1994/07/25 22:21:20 mikes Exp $
 #
 #            T H E    P I N E    M A I L   S Y S T E M
 #
@@ -13,7 +13,7 @@
 #
 #   Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 #
-#   Copyright 1989-1993  University of Washington
+#   Copyright 1989-1994  University of Washington
 #
 #    Permission to use, copy, modify, and distribute this software and its
 #   documentation for any purpose and without fee to the University of
@@ -61,20 +61,25 @@ OPTIMIZE=    # -O
 PROFILE=     # -pg
 DEBUG=       -g -DDEBUG
 
+IMAPDIR=     ../c-client
+PICODIR=     ../pico
+
 LIBES=       -ltermcap -lnet 
-LOCLIBES=    ../pico/libpico.a ../c-client/c-client.a
+LOCLIBES=    $(PICODIR)/libpico.a $(IMAPDIR)/c-client.a
 
-CFLAGS= -DHPP $(OPTIMIZE) $(PROFILE) $(DEBUG)
+CFLAGS= -DHPP $(OPTIMIZE) $(PROFILE) $(DEBUG) -DSYSTYPE=\"HPP\"
 
-obj=    addrbook.o addrutil.o adrbklib.o args.o context.o filter.o \
-	folder.o help.o helptext.o imap.o init.o mailcmd.o mailindx.o \
-	mailview.o newmail.o other.o pine.o print.o reply.o screen.o \
-	send.o signals.o status.o strings.o ttyin.o ttyout.o os.o
+obj=    addrbook.o adrbklib.o args.o context.o filter.o \
+	folder.o help.o helptext.o imap.o init.o mailcap.o mailcmd.o \
+	mailindx.o mailpart.o mailview.o newmail.o other.o pine.o print.o \
+	reply.o screen.o send.o signals.o status.o strings.o ttyin.o \
+	ttyout.o os.o
 
-cfiles= addrbook.c addrutil.c adrbklib.c args.c context.c filter.c \
-	folder.c help.c helptext.c imap.c init.c mailcmd.c mailindx.c \
-	mailview.c newmail.c other.c pine.c print.c reply.c screen.c \
-	send.c signals.c status.c strings.c ttyin.c ttyout.c os.c
+cfiles= addrbook.c adrbklib.c args.c context.c filter.c \
+	folder.c help.c helptext.c imap.c init.c mailcap.c mailcmd.c \
+	mailindx.c mailpart.c mailview.c newmail.c other.c pine.c print.c \
+	reply.c screen.c send.c signals.c status.c strings.c ttyin.c \
+	ttyout.c os.c
 
 # The & means make in parallel
 pine: $& $(obj) $(LOCLIBES)
@@ -92,21 +97,22 @@ clean:
 
 
 HEADERS= headers.h pine.h os.h helptext.h context.h \
-	  ../c-client/mail.h ../c-client/osdep.h
+	  $(IMAPDIR)/mail.h $(IMAPDIR)/osdep.h
 
 addrbook.o:	addrbook.c  $(HEADERS) adrbklib.h
-addrutil.o:	addrutil.c  $(HEADERS) adrbklib.h
 adrbklib.o: 	adrbklib.c  $(HEADERS) adrbklib.h
 args.o:		args.c      $(HEADERS)
-context.o:	context.c   $(HEADERS) ../c-client/misc.h
-filter.o:	filter.c    $(HEADERS) ../pico/pico.h
+context.o:	context.c   $(HEADERS) $(IMAPDIR)/misc.h
+filter.o:	filter.c    $(HEADERS) $(PICODIR)/pico.h
 folder.o:	folder.c    $(HEADERS)
 help.o:		help.c      $(HEADERS)
 helptext.o:	helptext.c
 imap.o:	        imap.c      $(HEADERS)
 init.o:		init.c      $(HEADERS)
+mailcap.o:	mailcap.c   $(HEADERS)
 mailcmd.o:	mailcmd.c   $(HEADERS)
 mailindx.o:	mailindx.c  $(HEADERS)
+mailpart.o:	mailpart.c  $(HEADERS)
 mailview.o:	mailview.c  $(HEADERS)
 newmail.o:	newmail.c   $(HEADERS)
 os.o:		os.c 	    $(HEADERS)
@@ -115,7 +121,7 @@ pine.o:		pine.c      $(HEADERS)
 print.o:	print.c     $(HEADERS)
 reply.o:	reply.c     $(HEADERS)
 screen.o:	screen.c    $(HEADERS)
-send.o:	        send.c      $(HEADERS) ../c-client/smtp.h ../pico/pico.h
+send.o:	        send.c      $(HEADERS) $(IMAPDIR)/smtp.h $(PICODIR)/pico.h
 signals.o:	signals.c   $(HEADERS)
 status.o:	status.c    $(HEADERS)
 strings.o:	strings.c   $(HEADERS)
@@ -141,8 +147,9 @@ osdep/os-hpp.c:	osdep/bld_path osdep/canacces osdep/canonicl \
 		osdep/diskquot.hpp osdep/domnames osdep/err_desc \
 		osdep/expnfldr osdep/fgetpos osdep/filesize osdep/fltrname \
 		osdep/fnexpand osdep/header osdep/hostname.una \
-		osdep/jobcntrl osdep/lstcmpnt osdep/mimedisp \
+		osdep/jobcntrl osdep/lstcmpnt osdep/mimedisp osdep/pipe \
 		osdep/play_snd.non osdep/pw_stuff osdep/readfile \
 		osdep/rename osdep/srandom.dum osdep/tempfile osdep/tempnam \
-		osdep/writ_dir osdep/sunquota osdep/os-hpp.ic
+		osdep/sendmail osdep/execview osdep/writ_dir osdep/sunquota \
+		osdep/os-hpp.ic
 	cd osdep; $(MAKE) includer os-hpp.c; cd ..
