@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: newmail.c,v 4.22 1993/11/30 23:46:36 mikes Exp $";
+static char rcsid[] = "$Id: newmail.c,v 4.37 1994/08/18 06:33:44 mikes Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -16,7 +16,7 @@ static char rcsid[] = "$Id: newmail.c,v 4.22 1993/11/30 23:46:36 mikes Exp $";
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1993  University of Washington
+   Copyright 1989-1994  University of Washington
 
     Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee to the University of
@@ -101,18 +101,12 @@ new_mail(cursor_moved, force, time_for_check_point)
     static time_t last_check = 0;
     static time_t last_check_point_call = 0;
     time_t        now;
-    long          x, min_changed_sorted_msgno, old_current_raw_msgno, rv;
+    long          rv;
     MAILSTREAM   *stream;
     register struct pine *pine_state;
     int           checknow = 0;
-#ifdef	DOS
-    FILE	  *old_sort;
-    long           old_sort_val;
-#else
-    long          *old_sort;
-#endif
 
-    dprint(9, (debugfile, "new mail called (%x %d %d)\n",
+    dprint(9, (debugfile, "new mail called (%p %d %d)\n",
                cursor_moved, force, time_for_check_point));
     pine_state = ps_global;  /*  this gets called out of the composer which
                               *  doesn't know about pine_state
@@ -122,11 +116,11 @@ new_mail(cursor_moved, force, time_for_check_point)
     /*
      * only check every 15 seconds, unless we're compelled to
      */
-    if(!(stream = pine_state->mail_stream) || (now-last_check_point_call <= 15
-	&& time_for_check_point != 0 && !pine_state->mail_box_changed 
-	&& !force))
+    if(!(stream = pine_state->mail_stream)
+       || (now-last_check_point_call <= 15 && time_for_check_point != 0
+	   && !pine_state->mail_box_changed && !force))
       return(-1);
-    else if(force || now-last_check >= timeout-2){ /* 2 to check each timeout */
+    else if(force || now-last_check >= timeout-2){ /* 2: check each timeout */
 	checknow++;
         last_check = now;
     }
@@ -135,10 +129,13 @@ new_mail(cursor_moved, force, time_for_check_point)
 
     if(!check_point(time_for_check_point == 0 ? GoodTime:
 	      time_for_check_point == 1 ? BadTime : VeryBadTime) && checknow){
-	StartInverse();
-	PutLine0(0, 1, "*"); 		/* Show something indicate delay*/
-	MoveCursor(ps_global->ttyo->screen_rows -3, 0);
-	fflush(stdout);
+	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	    StartInverse();
+	    PutLine0(0, 1, "*");	/* Show something indicate delay*/
+	    MoveCursor(ps_global->ttyo->screen_rows -3, 0);
+	    fflush(stdout);
+	}
+
 #ifdef	DEBUG
 	now = time(0);
 	dprint(7, (debugfile, "Mail_Ping(mail_stream): %s\n", ctime(&now)));
@@ -152,20 +149,25 @@ new_mail(cursor_moved, force, time_for_check_point)
 	now = time(0);
 	dprint(7, (debugfile, "Ping complete: %s\n", ctime(&now)));
 #endif
-	PutLine0(0, 1, " ");
-	EndInverse();
-	fflush(stdout);
+	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	    PutLine0(0, 1, " ");
+	    EndInverse();
+	    fflush(stdout);
+	}
     }
 
-    if(cursor_moved)		/* if we got here, the cursor moved. */
-      *cursor_moved = 1;
+    if(F_ON(F_SHOW_DELAY_CUE, ps_global) && cursor_moved)
+      *cursor_moved = 1;		/* if we got here, the cursor moved. */
 
     if(checknow && pine_state->inbox_stream 
        && stream != pine_state->inbox_stream){
-	StartInverse();
-	PutLine0(0, 1, "*"); 		/* Show something indicate delay*/
-	MoveCursor(ps_global->ttyo->screen_rows -3, 0);
-	fflush(stdout);
+	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	    StartInverse();
+	    PutLine0(0, 1, "*"); 	/* Show something indicate delay*/
+	    MoveCursor(ps_global->ttyo->screen_rows -3, 0);
+	    fflush(stdout);
+	}
+
 #ifdef	DEBUG
 	now = time(0);
 	dprint(7, (debugfile, "Mail_Ping(inbox_stream): %s\n", ctime(&now)));
@@ -177,9 +179,11 @@ new_mail(cursor_moved, force, time_for_check_point)
 	now = time(0);
 	dprint(7, (debugfile, "Ping complete: %s\n", ctime(&now)));
 #endif
-	PutLine0(0, 1, " ");
-	EndInverse();
-	fflush(stdout);
+	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	    PutLine0(0, 1, " ");
+	    EndInverse();
+	    fflush(stdout);
+	}
     }
 
 
@@ -193,7 +197,7 @@ new_mail(cursor_moved, force, time_for_check_point)
                       "inbox" : "other",
                    pine_state->new_mail_count,
                    pine_state->expunge_count,
-                   pine_state->max_msgno));
+                   mn_get_total(pine_state->msgmap)));
 
         if(ps_global->new_mail_count > 0) {
             mailbox_mail_since_command += ps_global->new_mail_count;
@@ -202,133 +206,70 @@ new_mail(cursor_moved, force, time_for_check_point)
                       pine_state->mail_stream == pine_state->inbox_stream ?
                       NULL :  pine_state->cur_folder,
                       mailbox_mail_since_command,
-                      pine_state->max_msgno);
+		      mn_get_total(pine_state->msgmap));
         }
 
-        old_current_raw_msgno = (ps_global->current_sorted_msgno <= 0L) ? 0L :
-                         ps_global->sort[ps_global->current_sorted_msgno];
-        dprint(9, (debugfile,
-                  "old_current_raw_msgno: %ld  sort_allocated; %ld\n",
-                   old_current_raw_msgno, ps_global->sort_allocated));
+	/*
+	 * Straighten out any local flag problems here.  We can't
+	 * take care of them in the mm_exists or mm_expunged callbacks
+	 * since the flags themselves are in an MESSAGECACHE and we're
+	 * not allowed to reenter c-client from a callback...
+	 */
+	if(pine_state->new_mail_count > 0L
+	   && any_lflagged(pine_state->msgmap, MN_HIDE)){
+	    rv = mn_get_total(pine_state->msgmap)
+		   - pine_state->new_mail_count + 1L;
+	    while(rv > 0L && rv <= mn_get_total(pine_state->msgmap))
+	      set_lflag(pine_state->mail_stream, pine_state->msgmap, rv++,
+			MN_HIDE, 1);
+	}
 
-        if(ps_global->sort_allocated <= ps_global->max_msgno + 1){
-            /*--- Pad out the sort array if needed ----*/
-#ifdef	DOS
-	    /*
-	     * Blindly doubling the array size is too costly under
-	     * dos.
-	     */
-	    while(ps_global->sort_allocated < ps_global->max_msgno + 1)
-		ps_global->sort_allocated += 64;
-#else
-            ps_global->sort_allocated = max(ps_global->max_msgno + 1,
-                                            ps_global->sort_allocated * 2);
-#endif
-            fs_resize((void **)&(ps_global->sort),
-                      (size_t)ps_global->sort_allocated * sizeof(long));
-        }
+	/*
+	 * Also, worry about the case where expunged away all of the 
+	 * zoomed messages.  Unhide everything in that case...
+	 */
+	if(mn_get_total(pine_state->msgmap) > 0L){
+	    if(any_lflagged(pine_state->msgmap, MN_HIDE)
+					  >= mn_get_total(pine_state->msgmap)){
+		for(rv = 1L; rv <= mn_get_total(pine_state->msgmap); rv++)
+		  set_lflag(pine_state->mail_stream, pine_state->msgmap, rv,
+			    MN_HIDE, 0);
 
-        for(x  = ps_global->max_msgno - ps_global->new_mail_count + 1;
-            x <= ps_global->max_msgno; x++) {
-            ps_global->sort[x] = x;
-        }
+		mn_set_cur(pine_state->msgmap,
+			   mn_get_total(pine_state->msgmap));
+	    }
+	    else if(any_lflagged(pine_state->msgmap, MN_HIDE)){
+		/*
+		 * if we got here, there are some hidden messages and
+		 * some not.  Make sure the current message is one
+		 * that's not...
+		 */
+		for(rv = mn_get_cur(pine_state->msgmap);
+		    rv <= mn_get_total(pine_state->msgmap);
+		    rv++){
+		    if(!get_lflag(pine_state->mail_stream, pine_state->msgmap,
+				 rv, MN_HIDE)){
+			mn_set_cur(pine_state->msgmap, rv);
+			break;
+		    }
+		}
 
-	if(ps_global->sort_allocated) {
-#ifdef	DOS
-	    if(old_sort = create_tmpfile()){
-		for(x = 1; x <= ps_global->max_msgno; x++){
-		    if(fwrite((void *)&(ps_global->sort[x]), sizeof(long), 
-			      (size_t)1, old_sort) != 1){
-			fclose(old_sort);
-			old_sort = NULL;
+		for(rv = mn_get_cur(pine_state->msgmap); rv > 0L; rv--){
+		    if(!get_lflag(pine_state->mail_stream, pine_state->msgmap,
+				 rv, MN_HIDE)){
+			mn_set_cur(pine_state->msgmap, rv);
 			break;
 		    }
 		}
 	    }
-#else
-           old_sort = (long *)fs_get(sizeof(long)*(size_t)ps_global->sort_allocated);
-           for(x = 1; x <= ps_global->max_msgno; x++)
-             old_sort[x] = ps_global->sort[x];
-#endif
-        }  else {
-            old_sort = NULL;
-        }
-
-
-        sort_current_folder();
-
-        /* Figure out what the current message number is now */
-	if(old_current_raw_msgno && ps_global->max_msgno > 0){
-	    /*
-	     * Previously had messages AND still have messages
-	     * to pick from...
-	     */
-	    for(x = 1; x <= ps_global->max_msgno; x++) {
-		if(ps_global->sort[x] == old_current_raw_msgno) {
-		    ps_global->new_current_sorted_msgno = x;
-		    break;
-		}
-	    }
-	}
-	else if(ps_global->max_msgno > 0)
-	  /*
-	   * No previous messages, BUT there are new ones, so
-	   * pick one to highlight...
-	   */
-	  ps_global->new_current_sorted_msgno = 1L;
-	else
-	  /*
-	   * No old messages and no new messages so make sure 
-	   * functions down the line know that...
-	   */
-	  ps_global->new_current_sorted_msgno = -1L;
-
-
-        /* Figure out the lowest position message sorted too */
-        if(old_sort != NULL){
-#ifdef	DOS
-	    fseek(old_sort, 0L, 0);
-#endif
-	    for(x = 1; x <= ps_global->max_msgno; x++) {
-#ifdef	DOS
-		if(fread(&old_sort_val,sizeof(long),(size_t)1,old_sort) != 1){
-		    fclose(old_sort);
-		    old_sort = NULL;
-		    break;
-		}
-
-		if(ps_global->sort[x] != old_sort_val)
-		  break;
-#else
-		if(ps_global->sort[x] != old_sort[x])
-		  break;
-#endif
-	    }
 	}
 
-        /* Figure out the lowest position that changed */
-        if(pine_state->expunge_count > 0 || old_sort == NULL) {
-            min_changed_sorted_msgno = 0;
-        } else if(x > ps_global->max_msgno) {
-            min_changed_sorted_msgno = ps_global->max_msgno -
-              ps_global->new_mail_count;
-        } else {
-            min_changed_sorted_msgno = x - 1;
-        }
-        dprint(9, (debugfile, "new sorted_msgno: %ld  min_changed: %ld\n",
-                   ps_global->new_current_sorted_msgno,
-                   min_changed_sorted_msgno));
-
-        if(old_sort != NULL)
-#ifdef	DOS
-	  fclose(old_sort);
-#else
-          fs_give((void **)&old_sort);
-#endif
-
-	if((ps_global->current_sort_order != SortArrival) ||
-	                                      ps_global->current_reverse_sort)
-          clear_index_cache();
+	/*
+	 * Lastly, worry about sorting if we got something new, otherwise
+	 * it was taken care of inside mm_expunge...
+	 */
+	if(pine_state->new_mail_count > 0L)
+	  sort_current_folder();
     }
 
     if(pine_state->inbox_changed &&
@@ -339,24 +280,25 @@ new_mail(cursor_moved, force, time_for_check_point)
          "New mail, inbox, new_mail_count:%ld expunge: %ld,  max_msgno %ld\n",
                    pine_state->inbox_new_mail_count,
                    pine_state->inbox_expunge_count,
-                   pine_state->inbox_max_msgno));
+                   mn_get_total(pine_state->inbox_msgmap)));
 
         if(pine_state->inbox_new_mail_count > 0) {
             inbox_mail_since_command       += ps_global->inbox_new_mail_count;
             ps_global->inbox_new_mail_count = 0;
             new_mail_mess(pine_state->inbox_stream,NULL,
                           inbox_mail_since_command,
-                          pine_state->inbox_max_msgno);
+			  mn_get_total(pine_state->inbox_msgmap));
         }
     }
 
-    rv = pine_state->mail_box_changed ? min_changed_sorted_msgno : -1;
+    rv = pine_state->new_mail_count + pine_state->expunge_count;
     pine_state->inbox_changed    = 0;
     pine_state->mail_box_changed = 0;
-    ps_global->new_mail_count    = 0;
+    ps_global->new_mail_count    = 0L;
 
-    dprint(6, (debugfile, "******** new mail returning %ld  ********\n", rv));
-    return(rv);
+    dprint(6, (debugfile, "******** new mail returning %ld  ********\n", 
+	       rv ? rv : -1));
+    return(rv ? rv : -1);
 }
 
 
@@ -395,50 +337,42 @@ new_mail_mess(stream, folder, number, max_num)
 				"subject:"
     };
 
-    e = mail_fetchstructure(stream, max_num, NULL);
-
-    if(folder == NULL) {
+    if(!folder) {
         if(number > 1)
           sprintf(intro, "%ld new messages!", number);
         else
           strcpy(intro, "New mail!");
-    } else {
+    }
+    else {
         if(number > 1)
           sprintf(intro,"%ld messages saved to folder \"%s\"", number, folder);
         else
           sprintf(intro, "Mail saved to folder \"%s\"", folder);
     }
-     
-    if(number > 1) {
-        if(e != NULL && e->from != NULL) {
-            if(e->from->personal != NULL)
-              sprintf(from, " Most recent from %s", e->from->personal);
-            else
-              sprintf(from, " Most recent from %s@%s", e->from->mailbox,
-                      e->from->host);
-        } else {
-            from[0] = '\0';
-        }
-        q_status_message2(0, 2, 4, "\007%s%s", intro, from);
-    } else {
-        if(e != NULL && e->from != NULL) {
-            if(e->from->personal != NULL)
-              sprintf(from, " From %s", e->from->personal);
-            else
-              sprintf(from, " From %s@%s", e->from->mailbox, e->from->host);
-        } else {
-            from[0] = '\0';
-        }
-        if(e != NULL && e->subject == NULL)
-          strcpy(subject, " with no subject");
-        else
-          sprintf(subject, " %s %.199s", carray[(unsigned)random()%12], 
-		                         e->subject);
-        if(from == NULL)
+
+    if((e = mail_fetchstructure(stream, max_num, NULL)) && e->from)
+      sprintf(from, " %srom %s%s%s", 
+	     (number > 1L) ? "Most recent f" : "F",
+	     (e->from->personal) ? e->from->personal : e->from->mailbox,
+	     (e->from->personal) ? "" : "@",
+	     (e->from->personal) ? "" : e->from->host);
+    else
+      from[0] = '\0';
+
+    if(number <= 1L) {
+        if(e && e->subject)
+	  sprintf(subject, " %s %.150s", carray[(unsigned)random()%12],
+		  e->subject);
+	else
+	  strcpy(subject, " with no subject");
+
+        if(!from[0])
           subject[1] = toupper(subject[1]);
 
-        q_status_message3(0, 2,4, "\007%s%s%s", intro, from, subject);
+        q_status_message3(0, 2, 4, "\007%s%s%s", intro, from, subject);
     }
+    else
+      q_status_message2(0, 2, 4, "\007%s%s", intro, from);
 }
 
 
@@ -460,10 +394,10 @@ various pieces of pines state associated with the message count and such.
 Only need to checkpoint current stream because there are no changes going
 on with other streams when we're not manipulating them.
   ----*/
-static int check_count = 0;           /* number of changes since last chk_pt  */
-static long first_status_change = 0;  /* time of 1st change since last chk_pt */
-static long last_status_change = 0;   /* time of last change                  */
-static long check_count_ave = 10 * 10;
+static int check_count		= 0;  /* number of changes since last chk_pt */
+static long first_status_change = 0;  /* time of 1st change since last chk_pt*/
+static long last_status_change	= 0;  /* time of last change                 */
+static long check_count_ave	= 10 * 10;
 
 check_point(timing)
      CheckPointTime timing;
@@ -518,10 +452,13 @@ check_point(timing)
                      "Doing checkpoint: %s  Since 1st status change: %d\n",
                      ctime(&now), now - first_status_change));
 #endif
-        StartInverse();
-        PutLine0(0, 0, "**"); 		/* Show something indicate delay*/
-	MoveCursor(ps_global->ttyo->screen_rows -3, 0);
-        fflush(stdout);
+	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	    StartInverse();
+	    PutLine0(0, 0, "**");	/* Show something indicate delay*/
+	    MoveCursor(ps_global->ttyo->screen_rows -3, 0);
+	    fflush(stdout);
+	}
+
         mail_check(ps_global->mail_stream);
 					/* Causes mail file to be written */
 #ifdef	DEBUG
@@ -530,9 +467,12 @@ check_point(timing)
 #endif
         check_count = 0;
         first_status_change = time(0);
-        PutLine0(0, 0, "  ");
-        EndInverse();
-        fflush(stdout);
+	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	    PutLine0(0, 0, "  ");
+	    EndInverse();
+	    fflush(stdout);
+	}
+
         return(1);
     } else {
         return(0);
