@@ -1,5 +1,5 @@
 /*----------------------------------------------------------------------
-  $Id: os-lnx.h,v 1.6 1993/11/16 00:32:46 hubert Exp $
+  $Id: os-lnx.h,v 1.17 1994/08/23 02:47:02 hubert Exp $
 
             T H E    P I N E    M A I L   S Y S T E M
 
@@ -14,7 +14,7 @@
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1993  University of Washington
+   Copyright 1989-1994  University of Washington
 
     Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee to the University of
@@ -73,6 +73,16 @@
 
 
 
+/*----------------------------------------------------------------------
+   Define this if you want to allow the users to change their From header
+   line when they send out mail.  The users will still have to configure
+   either default-composer-hdrs or custumized-hdrs to get at the From
+   header, even if this is set.
+ ----*/
+/* #define ALLOW_CHANGING_FROM  /* comment out to not allow changing From */
+
+
+
 /*----------------------------------------------------------------------
     Turn this on if you want to disable the keyboard lock function.
  ----*/
@@ -123,6 +133,7 @@
 
 /*----- System-wide config file ----------------------------------------*/
 #define SYSTEM_PINERC             "/usr/local/lib/pine.conf"
+#define SYSTEM_PINERC_FIXED       "/usr/local/lib/pine.conf.fixed"
 
 
 
@@ -133,7 +144,8 @@
 #define DF_DEFAULT_FCC            "sent-mail"
 #define DEFAULT_SAVE              "saved-messages"
 #define POSTPONED_MAIL            "postponed-mail"
-#define INTERRUPTED_MAIL          "interrupted-mail"
+#define POSTPONED_MSGS            "postponed-msgs"
+#define INTERRUPTED_MAIL          ".pine-interrupted-mail"
 #define DF_MAIL_DIRECTORY         "mail"
 #define INBOX_NAME                "INBOX"
 #define DF_SIGNATURE_FILE         ".signature"
@@ -143,13 +155,15 @@
 #define DF_USE_ONLY_DOMAIN_NAME   "no"
 #define DF_FEATURE_LEVEL          "sapling"
 #define DF_SAVE_BY_SENDER         "no"
-#define DF_LOCAL_MAIL_FORMAT      "pine"
 #define DF_SORT_KEY               "arrival"
+#define DF_AB_SORT_RULE           "fullname-with-lists-last"
 #define DF_SAVED_MSG_NAME_RULE    "default-folder"
-#define	DF_DRIVER_PROTO		  (&bezerkproto)
+#define DF_FCC_RULE               "default-fcc"
 #define DF_STANDARD_PRINTER       "lpr"
 #define ANSI_PRINTER              "attached-to-ansi"
 #define DF_ADDRESSBOOK            ".addressbook"
+#define DF_BUGS_FULLNAME          "Pine Developers"
+#define DF_BUGS_ADDRESS           "pine-bugs@cac.washington.edu"
 
 /*----------------------------------------------------------------------
    The default printer when pine starts up for the first time with no printer
@@ -179,12 +193,12 @@
 /*-------- Standard ANSI functions usually defined in stdlib.h ---------*/
 #include <stdlib.h>  /* declarations of many common functions */
 /*---- Declare getenv() if not already declared ------------------------*/
-char *getenv();      /* often declared in stdlib.h */
-FILE *tmpfile();
-extern int errno;
+/* char *getenv();      /* often declared in stdlib.h */
+/* FILE *tmpfile(); */
+/* extern int errno; */
 /* Provide these two functions in os.c */
 
-int fgetpos(); int fsetpos();
+/* int fgetpos(); int fsetpos(); */
 
 
 /*------- Standard string functions ------------------------------------*/
@@ -207,9 +221,9 @@ int fgetpos(); int fsetpos();
 
 
 /*----------------- time.h ---------------------------------------------*/
-/* #include <time.h> */
+#include <time.h> /* */
 /* plain time.h isn't enough on some systems */
-#include <sys/time.h>  /* For struct timeval usually in time.h */ 
+/* #include <sys/time.h>  /* For struct timeval usually in time.h */ 
 
 
 
@@ -244,7 +258,7 @@ int fgetpos(); int fsetpos();
 
 
 /*--------- System V terminal driver -----------------------------------*/
-/* #define HAVE_TERMIO     /* this is for pure System V */
+#define HAVE_TERMIO     /* this is for pure System V */
 /* #define HAVE_TERMIOS    /* this is an alternative */
 /* #define USE_POLL        /* use the poll() system call instead of select() */
 /* #define USE_TERMINFO    /* use terminfo instead of termcap */
@@ -279,7 +293,7 @@ int fgetpos(); int fsetpos();
 #include <sys/param.h>          /* Get it from param.h if available */
 #undef MAXPATH                  /* Sometimes defined in param.h differently */
 #define MAXPATH MAXPATHLEN      /* Longest pathname we ever expect */
-/* #define MAXPATH        (512)    /* Longest pathname we ever expect */
+/* #define MAXPATH     (512)    /* Longest pathname we ever expect */
 #define MAXFOLDER      (64)     /* Longest foldername we ever expect */  
 
 
@@ -329,10 +343,10 @@ int fgetpos(); int fsetpos();
 /*----------------------------------------------------------------------
     Various maximum field lengths, probably shouldn't be changed.
  ----*/
-#define MAX_FULLNAME      (50) 
-#define MAX_NICKNAME      (20)
-#define MAX_ADDRESS      (100)
-#define MAX_NEW_LIST     (100)  /* Max addrs to be added when creating list */
+#define MAX_FULLNAME     (100) 
+#define MAX_NICKNAME      (40)
+#define MAX_ADDRESS      (200)
+#define MAX_NEW_LIST     (500)  /* Max addrs to be added when creating list */
 #define MAX_SEARCH       (100)  /* Longest string to search for             */
 #define MAX_ADDR_EXPN   (1000)  /* Longest expanded addr                    */
 #define MAX_ADDR_FIELD (10000)  /* Longest fully-expanded addr field        */
