@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: screen.c,v 4.28 1993/10/02 07:29:22 hubert Exp $";
+static char rcsid[] = "$Id: screen.c,v 4.49 1994/08/23 23:56:27 mikes Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -16,7 +16,7 @@ static char rcsid[] = "$Id: screen.c,v 4.28 1993/10/02 07:29:22 hubert Exp $";
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1993  University of Washington
+   Copyright 1989-1994  University of Washington
 
     Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee to the University of
@@ -64,48 +64,70 @@ static char rcsid[] = "$Id: screen.c,v 4.28 1993/10/02 07:29:22 hubert Exp $";
 
 #ifdef	ANSI
 char *format_titlebar(void);
-void  savebits(BITMAP, BITMAP, int);
-int   equalbits(BITMAP, BITMAP, int);
+void  savebits(bitmap_t, bitmap_t, int);
+int   equalbits(bitmap_t, bitmap_t, int);
+char *message_string(MSGNO_S *, char *);
+char *percentage(int, int, int);
+
 #else
 char *format_titlebar();
 void  savebits();
 int   equalbits();
+char *message_string();
+char *percentage();
 #endif
 
 
-/*------ 
-   Saved key menu drawing state 
-  ------*/
+/* Saved key menu drawing state */
 static struct {
     struct key_menu *km;
-    int              row, column;
-    int              blanked;
-    BITMAP           bitmap;
+    int              row,
+		     column,
+                     blanked;
+    bitmap_t         bitmap;
 } km_state;
 
 
-/*----------------------------------------------------------------------
-      Splat an already formatted key_menu to the screen
-
- Args:  km      - key_menu structure
-        bm      - bitmap, 0's mean don't draw this key
-        row     - the row on the screen to begin on, negative values
-                  are counted from the bottom of the screen up
-        column  - column on the screen to begin on
-
- Just write out the name and the label at the specified column.
-
- ----*/
+/*
+ * Write an already formatted key_menu to the screen
+ *
+ * Args: km     -- key_menu structure
+ *       bm     -- bitmap, 0's mean don't draw this key
+ *       row    -- the row on the screen to begin on, negative values
+ *                 are counted from the bottom of the screen up
+ *       column -- column on the screen to begin on
+ *
+ * The bits in the bitmap are used from least significant to most significant,
+ * not left to right.  So, if you write out the bitmap in the normal way, for
+ * example,
+ * bm[0] = 0x5, bm[1] = 0x8, bm[2] = 0x21, bm[3] = bm[4] = bm[5] = 0
+ *   0000 0101    0000 1000    0010 0001   ...
+ * means that menu item 0 (first row, first column) is set, item 1 (2nd row,
+ * first column) is not set, item 2 is set, items 3-10 are not set, item 11
+ * (2nd row, 6th and last column) is set.  In the second menu (the second set
+ * of 12 bits) items 0-3 are unset, 4 is set, 5-8 unset, 9 set, 10-11 unset.
+ * That uses up bm[0] - bm[2].
+ * Just to make sure, here it is drawn out for the first set of 12 items in
+ * the first keymenu (0-11)
+ *    bm[0] x x x x  x x x x   bm[1] x x x x  x x x x
+ *          7 6 5 4  3 2 1 0                 1110 9 8
+ */
 void
 output_keymenu(km, bm, row, column)
-    struct key_menu *km;
-    BITMAP bm;
-    int row, column;
+struct key_menu *km;
+bitmap_t	 bm;
+int		 row,
+		 column;
 {
     register struct key *k;
-    int                  i, c, ufk, real_row, fkey, max_column;
+    int                  i,
+			 c,
+			 ufk,        /* using function keys */
+			 real_row,
+			 fkey,
+			 max_column, /* number of columns on screen */
+			 off;        /* offset into keymap */
     char                 temp[20];
-    int			 off;
 #if	defined(DOS) && defined(MOUSE)
     extern void          register_key();
     int    j = -1;
@@ -116,7 +138,7 @@ output_keymenu(km, bm, row, column)
     max_column    = ps_global->ttyo->screen_cols;
 
     if(ps_global->ttyo->screen_rows < 4 || max_column <= 0)
-      return;
+        return;
 
     for(i = 0; i < 12; i++){
 	int e;
@@ -132,49 +154,51 @@ output_keymenu(km, bm, row, column)
 	register_key(i, NO_OP_COMMAND, "", 0, 0, 0);
 #endif
     }
-    ufk = F_ON(F_USE_FK,ps_global);
+
+    ufk = F_ON(F_USE_FK, ps_global);
     real_row = row > 0 ? row : ps_global->ttyo->screen_rows + row;
     dprint(9, (debugfile, "row: %d, real_row: %d, column: %d\n", row, 
-               real_row, column)) ;
-    for(i = 0; i < 2; i++) {
+               real_row, column));
+
+    for(i = 0; i < 2; i++){
 	int el;
         MoveCursor(real_row + i, column);
         CleartoEOLN();
         for(c = 0, el = off+i, k = &km->keys[el]; k < &km->keys[off+12];
-							k += 2, el += 2) {
+							k += 2, el += 2){
 #if	defined(DOS) && defined(MOUSE)
 	    j++;
 #endif
             if(ufk)
-              fkey =  1 + k - &km->keys[off];
+                fkey = 1 + k - &km->keys[off];
             if(!bitnset(el,bm) || k->name == NULL || k->column > max_column)
-              continue;
+                continue;
      
-            if(k->column < c) {
+            if(k->column < c){
                 /*--- over writing previous entry, a bit weird */
                 MoveCursor(row + i, k->column + column);
-            } else { 
+            }
+	    else{ 
                 Write_to_screen(repeat_char(k->column - c, ' '));
             }
 
             StartInverse();
-            if(ufk) {
+            if(ufk)
                 sprintf(temp, "F%d", fkey);
-            } else {
+            else
                 strcpy(temp, k->name);
-            }
             if(strlen(temp) + k->column > max_column)
-              temp[max_column - k->column] = '\0';
+                temp[max_column - k->column] = '\0';
             Write_to_screen(temp);
             EndInverse();
             c = k->column + strlen(temp);
 
             if(++c < max_column)
-              Writechar(' ', 0);
+                Writechar(' ', 0);
 
             strcpy(temp, k->label);
             if(c + strlen(temp) >= max_column)
-              temp[max(0,max_column - c)] = '\0';
+                temp[max(0, max_column - c)] = '\0';
             Write_to_screen(temp);
 #if	defined(DOS) && defined(MOUSE)
 	    register_key(j, (ufk) ? PF1 + fkey - 1 : 
@@ -194,12 +218,12 @@ output_keymenu(km, bm, row, column)
 }
 
 
-/*----------------------------------------------------------------------
-   Blank out the key menu and clear that part of the screen 
-  ----*/
+/*
+ * Clear the key menu lines.
+ */
 void
 blank_keymenu(row, column)
-     int row, column;
+int row, column;
 {
     km_state.blanked    = 1;
     km_state.row        = row;
@@ -212,59 +236,61 @@ blank_keymenu(row, column)
 }
         
 
-
-/*----------------------------------------------------------------------
-    Calculate formatting for key menu at bottom of screen
-
-   Args:  km -- The key_menu structure to format
-          bm -- Bitmap indicating which menu items should be displayed.  If
-		an item is NULL, that also means it shouldn't be displayed.
-		Sometimes the bitmap will be turned on in that case and just
-		rely on the NULL entry.
-          width -- the screen width to format it at
-	  what  -- Used to indicate which of the possible sets of 12 keys
-		   to show next
-	  which -- Only used when what == AParticularTwelve, in which case
-		   it indicates which particular twelve.
-
- If already formatted for this particular screen width, this set of twelve
- keys, and this part of the bitmap set this way, then return.
-
- The formatting results in the column field in the key_menu being filled in.
- The column field is the column to start the label at, the name of the key;
- after that is the label for or command that key will do. This would be simple
- except for a few long key names such as SPACE. All the calculations are to
- make this look nice. The basic idea is to line up the end of the names and 
- beginning of the labels. If the name is too long and shifting it left
- would run into previous label, then shift the whole menu right, or at least
- that entry of things following are short enough to fit back into the
- regular spacing. There's a special case to make 80 column screens look nice.
- This has to be calculated and not fixed so it can cope with screen resize.
-   ----*/
+/*
+ * Calculate formatting for key menu at bottom of screen
+ *
+ * Args:  km    -- The key_menu structure to format
+ *        bm    -- Bitmap indicating which menu items should be displayed.  If
+ *		   an item is NULL, that also means it shouldn't be displayed.
+ *		   Sometimes the bitmap will be turned on in that case and just
+ *		   rely on the NULL entry.
+ *        width -- the screen width to format it at
+ *	  what  -- Used to indicate which of the possible sets of 12 keys
+ *		   to show next
+ *	  which -- Only used when what == AParticularTwelve, in which case
+ *		   it indicates which particular twelve.
+ *
+ * If already formatted for this particular screen width, this set of twelve
+ * keys, and this part of the bitmap set this way, then return.
+ *
+ * The formatting results in the column field in the key_menu being
+ * filled in.  The column field is the column to start the label at, the
+ * name of the key; after that is the label for the key.  The basic idea
+ * is to line up the end of the names and beginning of the labels.  If
+ * the name is too long and shifting it left would run into previous
+ * label, then shift the whole menu right, or at least that entry of
+ * things following are short enough to fit back into the regular
+ * spacing.  This has to be calculated and not fixed so it can cope with
+ * screen resize.
+ *
+ * This seems to be working correctly now.  I'd like to take some time to
+ * try to figure out what the heck all the calculations are about.  It seems
+ * too complicated at first glance.
+ */
 void
 format_keymenu(km, bm, width, what, which)
-    struct key_menu *km;
-    BITMAP bm;
-    int width;
-    OtherMenu what;
-    int which;  /* only used if what == AParticularTwelve */
+struct key_menu *km;
+bitmap_t	 bm;
+int		 width;
+OtherMenu	 what;
+int		 which;	 /* only used if what == AParticularTwelve */
 {
     short   spacing[6];
     int     i, first, prev_end_top, prev_end_bot, ufk, top_column, bot_column;
     int     off;  /* offset into keymenu */
     struct key *keytop, *keybot, *prev_keytop, *prev_keybot;
 
-    if(what == FirstMenu) {
+    if(what == FirstMenu)
       km->which = 0;
-    }else if(what == NextTwelve) {
+    else if(what == NextTwelve)
       km->which = (km->which + 1) % (unsigned int)km->how_many;
-    }else if(what == AParticularTwelve) {
+    else if(what == AParticularTwelve)
       km->which = which;
-    } /* else SameTwelve */
+    /* else SameTwelve */
 
     /* it's already formatted. */
-    if(width == km->width[km->which] && equalbits(bm,km->bitmap,km->which))
-      return;
+    if(width == km->width[km->which] && equalbits(bm, km->bitmap, km->which))
+        return;
 
     /*
      * If we're in the initial command sequence we may be using function
@@ -272,52 +298,53 @@ format_keymenu(km, bm, width, what, which)
      * the formatting next time through.
      */
     if ((F_ON(F_USE_FK,ps_global) && ps_global->orig_use_fkeys) ||
-        (F_OFF(F_USE_FK,ps_global) && !ps_global->orig_use_fkeys)) {
+        (F_OFF(F_USE_FK,ps_global) && !ps_global->orig_use_fkeys)){
       km->width[km->which] = width;
       savebits(km->bitmap, bm, km->which);
     }
 
-    off = km->which * 12;
+    off = km->which * 12;  /* off is offset into keymenu */
 
-    ufk = F_ON(F_USE_FK,ps_global);
+    ufk = F_ON(F_USE_FK,ps_global);  /* ufk stands for using function keys */
 
     /* this spacing is column numbers for column key name ends in */
     /*       F1 Menu	   */
     /*        ^            */
     spacing[0] = 0;
     for(i = 1; i < 6; i++) 
-      spacing[i] = spacing[i-1] + (width == 80 ? (i > 2 ? 14 : 13) : width/6);
+        spacing[i] = spacing[i-1] + (width == 80 ? (i > 2 ? 14 : 13) : width/6);
 
 
     km->keys[off].column = km->keys[off+6].column = 0;
     if(ufk)
-      first = 2;
+        first = 2;
     else
-      first = max((bitnset(off,bm) && km->keys[off].name != NULL) ?
+        first = max((bitnset(off,bm) && km->keys[off].name != NULL) ?
 	        strlen(km->keys[off].name) : 0,
 	        (bitnset(off+1,bm) && km->keys[off+1].name != NULL) ?
 	        strlen(km->keys[off+1].name) : 0);
 
     prev_keytop = &km->keys[off];
     prev_keybot = &km->keys[off+1];
-    for(i = 1 ; i < 6; i++) {
-	int ktop, kbot;
+    for(i = 1 ; i < 6; i++){
+	int k_top, k_bot;
 
-	ktop = off + i*2;
-	kbot = ktop + 1;
-        keytop = &km->keys[ktop];
-        keybot = &km->keys[kbot];
+	k_top  = off + i*2;
+	k_bot  = k_top + 1;
+        keytop = &km->keys[k_top];
+        keybot = &km->keys[k_bot];
 
-        if(!bitnset(ktop,bm) || keytop->name == NULL) {
+        if(!bitnset(k_top,bm) || keytop->name == NULL){
             top_column = first + spacing[i];
-        } else {
+        }
+	else{
             prev_end_top = 
                  prev_keytop->column +
-                 ((!bitnset(ktop-2,bm) || prev_keytop->name == NULL) ? 0 : 
+                 ((!bitnset(k_top-2,bm) || prev_keytop->name == NULL) ? 0 : 
                    (ufk ? (i-1 >= 5 ? 3 : 2) : strlen(prev_keytop->name))) +
-                 1 + ((!bitnset(ktop-2,bm) || prev_keytop->label==NULL) ? 0 :
+                 1 + ((!bitnset(k_top-2,bm) || prev_keytop->label==NULL) ? 0 :
 		 strlen(prev_keytop->label)) +
-                 1 + ((!bitnset(ktop,bm) || keytop->name == NULL) ? 0 : 
+                 1 + ((!bitnset(k_top,bm) || keytop->name == NULL) ? 0 : 
                    (ufk ? (i >= 5 ? 3 : 2) : strlen(keytop->name)));
             top_column = max(prev_end_top, first + spacing[i]);
             dprint(9, (debugfile,
@@ -326,24 +353,25 @@ format_keymenu(km, bm, width, what, which)
                        spacing[i]));
         }
 
-        if(!bitnset(kbot,bm) || keybot->name == NULL) {
+        if(!bitnset(k_bot,bm) || keybot->name == NULL){
             bot_column = first + spacing[i];
-        } else {
+        }
+	else{
             prev_end_bot = 
                  prev_keybot->column +
-                 ((!bitnset(kbot-2,bm) || prev_keybot->name  == NULL) ? 0 :
+                 ((!bitnset(k_bot-2,bm) || prev_keybot->name  == NULL) ? 0 :
                   (ufk ? (i-1 >= 4 ? 3 : 2) :strlen(prev_keybot->name))) +
-                 1 + ((!bitnset(kbot-2,bm) || prev_keybot->label==NULL) ? 0 :
+                 1 + ((!bitnset(k_bot-2,bm) || prev_keybot->label==NULL) ? 0 :
 		 strlen(prev_keybot->label)) +
-                 1 + ((!bitnset(kbot,bm) || keybot->name  == NULL) ? 0 :
+                 1 + ((!bitnset(k_bot,bm) || keybot->name  == NULL) ? 0 :
                   (ufk ? (i >= 4 ? 3 : 2) :strlen(keybot->name)));
             bot_column = max(prev_end_bot, first + spacing[i]);
         }
         keytop->column = max(bot_column, top_column) -
-                          ((!bitnset(ktop,bm) || keytop->name == NULL) ? 0 : 
+                          ((!bitnset(k_top,bm) || keytop->name == NULL) ? 0 : 
                             (ufk ? (i >= 5 ? 3 : 2) : strlen(keytop->name)));
         keybot->column = max(bot_column, top_column) -
-                          ((!bitnset(kbot,bm) || keybot->name  == NULL) ? 0 :
+                          ((!bitnset(k_bot,bm) || keybot->name  == NULL) ? 0 :
                             (ufk ? (i >= 4 ? 3 : 2) : strlen(keybot->name)));
 
         prev_keytop = keytop;
@@ -352,36 +380,38 @@ format_keymenu(km, bm, width, what, which)
 }
 
 
-
-/*----------------------------------------------------------------------
-    Draw the key menu at bottom of screen
-
-   Args:  km -- key_menu structure
-          bitmap -- which fields are active
-          width -- the screen width to format it at
-	  row,column -- where to put it
-          what -- this is an enum telling us whether to display the
-		  first menu (first set of 12 keys) or to display the same
-		  one we displayed last time, or to display a particular
-		  one (which), or to display the next one.
-	  which -- the keys to display if (what == AParticularTwelve)
- 
-   Fields are inactive if either the corresponding bitmap entry is 0 or
-   the actual entry in the key_menu is NULL.  Therefore, it is sometimes
-   useful to just turn on all the bits in a bitmap and let the NULLs take
-   care of it.  On the other hand, the bitmap gives a convenient method
-   for turning some keys on or off dynamically or due to options.
-
-   Also saves the state for a possible redraw later.
-   ----*/
+/*
+ * Draw the key menu at bottom of screen
+ *
+ * Args:  km     -- key_menu structure
+ *        bitmap -- which fields are active
+ *        width  -- the screen width to format it at
+ *	  row    -- where to put it
+ *	  column -- where to put it
+ *        what   -- this is an enum telling us whether to display the
+ *		    first menu (first set of 12 keys), or to display the same
+ *		    one we displayed last time, or to display a particular
+ *		    one (which), or to display the next one.
+ *	  which  -- the keys to display if (what == AParticularTwelve)
+ *
+ * Fields are inactive if *either* the corresponding bitmap entry is 0 *or*
+ * the actual entry in the key_menu is NULL.  Therefore, it is sometimes
+ * useful to just turn on all the bits in a bitmap and let the NULLs take
+ * care of it.  On the other hand, the bitmap gives a convenient method
+ * for turning some keys on or off dynamically or due to options.
+ * Both methods are used about equally.
+ *
+ * Also saves the state for a possible redraw later.
+ */
 void
 draw_keymenu(km, bitmap, width, row, column, what, which)
-     struct key_menu *km;
-     int width;
-     BITMAP bitmap;
-     int row, column;
-     OtherMenu what;
-     int which;
+struct key_menu *km;
+bitmap_t	 bitmap;
+int	         width;
+int	         row,
+                 column;
+OtherMenu        what;
+int	         which;
 {
     format_keymenu(km, bitmap, width, what, which);
     output_keymenu(km, bitmap, row, column);
@@ -395,22 +425,18 @@ draw_keymenu(km, bitmap, width, row, column, what, which)
 }
 
 
-
-/*----------------------------------------------------------------------
-    
- */
 void
 redraw_keymenu()
 {
     if(km_state.blanked)
-        blank_keymenu(km_state.row, km_state.column);
+      blank_keymenu(km_state.row, km_state.column);
     else
-        draw_keymenu(km_state.km, km_state.bitmap, ps_global->ttyo->screen_cols,
-		km_state.row, km_state.column, SameTwelve, 0);
+      draw_keymenu(km_state.km, km_state.bitmap,
+		   ps_global->ttyo->screen_cols,
+		   km_state.row, km_state.column, SameTwelve, 0);
 }
     
 
-
 /*
  * some useful macros...
  */
@@ -420,27 +446,29 @@ redraw_keymenu()
 
 #define	BAR_STATUS(X)	(((X) & MS_DEL) ? "DEL"   \
 			 : ((X) & MS_ANS) ? "ANS"   \
-		           : (!IS_NEWS && ((X) & MS_NEW)) ? "NEW" : "   ")
+		           : (as.stream && !IS_NEWS(as.stream)  \
+			      && ((X) & MS_NEW)) ? "NEW" : "   ")
 
 
 
 static struct titlebar_state {
-    char *title,
-         *folder_name,
-         *context_name;
-    int   msg_state,
-          total_pages,
-          current_page,
-          total_lines,
-          current_line,
-          total_messages,
-          current_message,
-          cur_mess_col,
-          del_column, 
-          percent_column,
-          page_column,
-          screen_cols;
-    enum  {Normal, ReadOnly, Closed} stream_status;
+    MAILSTREAM	*stream;
+    char	*title,
+		*folder_name,
+		*context_name,
+		*current_message;
+    int		 msg_state,
+		 total_pages,
+		 current_page,
+		 total_lines,
+		 current_line,
+		 total_messages,
+		 cur_mess_col,
+		 del_column, 
+		 percent_column,
+		 page_column,
+		 screen_cols;
+    enum	 {Normal, ReadOnly, Closed} stream_status;
     TitleBarType style;
 } as, titlebar_stack;
 
@@ -456,7 +484,8 @@ static struct titlebar_state {
   add a few more flags
   ----------------------------------------------------------------------*/
 char *
-status_string(message_cache)
+status_string(stream, message_cache)
+     MAILSTREAM   *stream;
      MESSAGECACHE *message_cache;
 {
      static char string[3];
@@ -470,7 +499,8 @@ status_string(message_cache)
 
      string[0] = 'N';
 
-     if(message_cache == NULL || message_cache->seen || IS_NEWS)
+     if(message_cache == NULL || message_cache->seen
+	|| (stream && IS_NEWS(stream)))
        string[0] = ' ';
 
      if(message_cache->answered)
@@ -489,9 +519,10 @@ status_string(message_cache)
 void
 push_titlebar_state()
 {
-    titlebar_stack  = as;
-    as.folder_name  = NULL;	/* erase knowledge of malloc'd data */
-    as.context_name = NULL;
+    titlebar_stack     = as;
+    as.folder_name     = NULL;	/* erase knowledge of malloc'd data */
+    as.context_name    = NULL;
+    as.current_message = NULL;
 }
 
 
@@ -503,11 +534,28 @@ pop_titlebar_state()
 {
     fs_give((void **)&(as.folder_name)); /* free malloc'd values */
     fs_give((void **)&(as.context_name));
+    fs_give((void **)&(as.current_message));
     as = titlebar_stack;
 }
 
 
 
+/*--------
+------*/
+char *
+message_string(map, s)
+     MSGNO_S *map;
+     char    *s;
+{
+    sprintf(s, "%s%s%s",
+	    (mn_total_cur(map) > 1L) ? "(" : "",
+	    comatose((mn_total_cur(map)>1L)?mn_total_cur(map):mn_get_cur(map)),
+	    (mn_total_cur(map) > 1L) ? ")" : "");
+    return(s);
+}
+
+
+
 /*----------------------------------------------------------------------
       Sets up style and contents of current titlebar line
 
@@ -515,7 +563,7 @@ pop_titlebar_state()
           display_on_screen -- flag whether to display on screen or generate
                                 string
           style  -- The format/style of the titlebar line
-          current_message -- The current message number
+	  msgmap -- MSGNO_S * to selected message map
           current_pl -- The current page or line number
           total_pl   -- The total line or page count
 
@@ -531,10 +579,12 @@ It's OK to call this withy a bogus current message - it is only used
 to look up status of current message 
  
 Formats only change the right section (part3).
-  FolderName:    "<folder>"  xx Messages
-  MessageNumber: "<folder>" message x,xxx of x,xxx XXX
-  TextPercent:   "<folder>" message x,xxx of x,xxx  xx% XXX
-  PageNumber:    "page xx of xx
+  FolderName:      "<folder>"  xx Messages
+  MessageNumber:   "<folder>" message x,xxx of x,xxx XXX
+  TextPercent:     line xxx of xxx  xx%
+  MsgTextPercent:  "<folder>" message x,xxx of x,xxx  xx% XXX
+  FileTextPercent: "<filename>" line xxx of xxx  xx%
+  PageNumber:      "page xx of xx
 
 Several strings and column numbers are saved so later updates to the status 
 line for changes in message number or percentage can be done efficiently.
@@ -542,32 +592,35 @@ This code is some what complex, and might benefit from some improvements.
  ----*/
 
 char *
-set_titlebar(title, display_on_screen, style, current_message, current_pl,
-                                                                  total_pl)
+set_titlebar(title, stream, cntxt, folder, msgmap, display_on_screen, style,
+	     current_pl, total_pl)
      char        *title;
+     MAILSTREAM  *stream;
+     CONTEXT_S   *cntxt;
+     char        *folder;
+     MSGNO_S	 *msgmap;
      TitleBarType style;
      int          display_on_screen, current_pl, total_pl;
-     long         current_message;
 {
     char          *tb;
-    MESSAGECACHE *mc;
-    extern char   version_buff[];
+    MESSAGECACHE *mc = NULL;
 
-
-    dprint(9, (debugfile,
-"set_titlebar - style: %d  current_message:%ld  current_pl: %d  total_pl: %d\n",
-               style, current_message, current_pl, total_pl));
+    dprint(9, (debugfile, "set_titlebar - style: %d  current message cnt:%ld",
+	       style, mn_total_cur(msgmap)));
+    dprint(9, (debugfile, "  current_pl: %d  total_pl: %d\n", 
+	       current_pl, total_pl));
 
     as.style = style;
     as.title = title;
-    as.stream_status = (ps_global->mail_stream == NULL ||
-                                  ps_global->dead_stream) ? Closed :
-                       ps_global->mail_stream->readonly ? ReadOnly : Normal;
+    as.stream = stream;
+    as.stream_status = (!as.stream || (as.stream == ps_global->mail_stream
+				       && ps_global->dead_stream))
+			 ? Closed : as.stream->readonly ? ReadOnly : Normal;
 
     if(as.folder_name)
       fs_give((void **)&as.folder_name);
 
-    as.folder_name = cpystr(pretty_fn(ps_global->cur_folder));
+    as.folder_name = cpystr(pretty_fn(folder));
 
     if(as.context_name)
       fs_give((void **)&as.context_name);
@@ -575,42 +628,39 @@ set_titlebar(title, display_on_screen, style, current_message, current_pl,
     /*
      * Handle setting up the context if appropriate.
      */
-    if(ps_global->context_current && context_isambig(ps_global->cur_folder)
-       && ps_global->context_list->next && strucmp(as.folder_name, "inbox")){
+    if(cntxt && context_isambig(folder) && ps_global->context_list->next
+       && strucmp(as.folder_name, ps_global->inbox_name)){
 	/*
 	 * if there's more than one context and the current folder
 	 * is in it (ambiguous name), set the context name...
 	 */
-	as.context_name = cpystr(ps_global->context_current->label[0]);
+	as.context_name = cpystr(cntxt->label[0]);
     }
     else
       as.context_name = cpystr("");
 
-    if(current_message < 0)
-      current_message = 0;		/* BUG Don't like this */
+    if(mn_get_total(msgmap) < 1L)
+      mn_set_cur(msgmap, 0L);		/* BUG Don't like this */
 
-    if(style != FolderName && current_message > 0) {
-        (void)mail_fetchstructure(ps_global->mail_stream,
-				 (long)ps_global->sort[current_message],
-				 NULL);
-        mc  = mail_elt(ps_global->mail_stream,
-                       (long)ps_global->sort[current_message]);
-    } else {
-	mc  = (MESSAGECACHE *)NULL;
+    if(as.stream && style != FolderName && mn_get_cur(msgmap) > 0L) {
+        (void)mail_fetchstructure(as.stream,
+				  mn_m2raw(msgmap, mn_get_cur(msgmap)), NULL);
+        mc  = mail_elt(as.stream, mn_m2raw(msgmap, mn_get_cur(msgmap)));
     }
-
-    
     
     switch(style) {
       case FolderName:
-        as.total_messages = ps_global->max_msgno;
+        as.total_messages = mn_get_total(msgmap);
 	break;
 
       case MessageNumber:
         as.total_lines = total_pl;
         as.current_line = current_pl;
-        as.current_message = current_message;
-        as.total_messages = ps_global->max_msgno;
+	if(as.current_message)
+	  fs_give((void **)&as.current_message);
+
+	as.current_message = cpystr(message_string(msgmap, tmp_20k_buf));
+        as.total_messages  = mn_get_total(msgmap);
         as.msg_state = !mc ? 0
 			   : mc->deleted ? MS_DEL
 			     : mc->answered ? MS_ANS
@@ -618,10 +668,15 @@ set_titlebar(title, display_on_screen, style, current_message, current_pl,
 	break;
 
       case TextPercent:
+      case MsgTextPercent:
+      case FileTextPercent :
         as.total_lines = total_pl;
         as.current_line = current_pl;
-        as.current_message = current_message;
-        as.total_messages = ps_global->max_msgno;
+	if(as.current_message)
+	  fs_give((void **)&as.current_message);
+
+	as.current_message = cpystr(message_string(msgmap, tmp_20k_buf));
+        as.total_messages  = mn_get_total(msgmap);
         as.msg_state = !mc ? 0
 			   : mc->deleted ? MS_DEL
 			     : mc->answered ? MS_ANS
@@ -655,8 +710,6 @@ redraw_titlebar()
 
 
 #define NUM_LEN(x) ((x) < 10 ? 1 : (x) < 100 ? 2 : (x) < 1000 ? 3 : 4)
-#define CALC_PERCENT(part, tot) ((tot) == 0 ? 0 : \
-				  (int)((100L * (long)(part))/(long)(tot)))
 
 
 /*----------------------------------------------------------------------
@@ -690,11 +743,10 @@ char *
 format_titlebar()
 {
     static  char titlebar_line[MAX_SCREEN_COLS+1];
-    char    version[50], comatmp[20], fold_tmp[MAXPATH],
+    char    version[50], fold_tmp[MAXPATH],
            *loc_label, *ss_string;
     int     sc, tit_len, ver_len, loc_len, fold_len, num_len, ss_len, 
             is_context;
-    extern  char version_buff[];
 
     /* blank the line */
     memset((void *)titlebar_line, ' ', MAX_SCREEN_COLS*sizeof(char));
@@ -707,9 +759,10 @@ format_titlebar()
     as.percent_column = -1;
     as.page_column    = -1;
     is_context        = strlen(as.context_name);
-    sprintf(version, "PINE %s", version_buff); 
+    sprintf(version, "PINE %s", pine_version); 
     ss_string         = as.stream_status == Closed ? "(CLOSED)" :
-                        (as.stream_status == ReadOnly && !IS_NEWS)
+                        (as.stream_status == ReadOnly
+			 && !IS_NEWS(as.stream))
                            ? "(READONLY)" : "";
     ss_len            = strlen(ss_string);
 
@@ -741,34 +794,36 @@ format_titlebar()
 		loc_label, plural(as.total_messages));
 	break;
       case MessageNumber :	       	/* "<loc_label> xxx of xxx DEL"  */
-	num_len  = NUM_LEN(as.current_message) + NUM_LEN(as.total_messages);
-	if(as.current_message >= 1000)
-	  num_len++;
-	if(as.total_messages >= 1000)
-	  num_len++;
-	loc_len += num_len + 9;		/* add spaces and "DEL" */
-	as.cur_mess_col = sc - 10 - num_len;
-	as.del_column   = sc - 5;
-	strcpy(comatmp, comatose((long)as.current_message));
+	num_len		 = strlen(comatose(as.total_messages));
+	loc_len		+= (2 * num_len) + 9;	/* add spaces and "DEL" */
+	as.cur_mess_col  = sc - (2 * num_len) - 10;
+	as.del_column    = as.cur_mess_col + num_len 
+			    + strlen(as.current_message) + 5;
 	sprintf(tmp_20k_buf, "%s %s of %s %s", loc_label, 
-		comatmp, comatose((long)as.total_messages),
+		as.current_message, comatose((long)as.total_messages),
+	       (!isdigit(*as.current_message))?"   ":BAR_STATUS(as.msg_state));
+	break;
+      case MsgTextPercent :		/* "<loc_label> xxx of xxx xxx% DEL" */
+	num_len		   = strlen(comatose(as.total_messages));
+	loc_len		  += (2 * num_len) + 13; /* add spaces, %, and "DEL" */
+	as.cur_mess_col    = sc - 16 - (2 * num_len);
+	as.percent_column  = as.cur_mess_col + num_len
+			      + strlen(as.current_message) + 7;
+	as.del_column      = as.percent_column + 4;
+	sprintf(tmp_20k_buf, "%s %s of %s %s %s", loc_label, 
+		as.current_message, comatose((long)as.total_messages),
+		percentage(as.current_line, as.total_lines, 1),
 		BAR_STATUS(as.msg_state));
 	break;
-      case TextPercent :		/* "<loc_label> xxx of xxx xxx% DEL" */
-	num_len  = NUM_LEN(as.current_message) + NUM_LEN(as.total_messages);
-	if(as.current_message >= 1000)
-	  num_len++;
-	if(as.total_messages >= 1000)
-	  num_len++;
-	loc_len += num_len + 14;	/* add spaces, percent, and "DEL" */
-	as.del_column     = sc - 5;
-	as.percent_column = sc - 10;
-	as.cur_mess_col   = sc - 16 - num_len;
-	strcpy(comatmp, comatose((long)as.current_message));
-	sprintf(tmp_20k_buf, "%s %s of %s %3d%% %s", loc_label, 
-		comatmp, comatose((long)as.total_messages),
-		CALC_PERCENT(as.current_line, as.total_lines),
-		BAR_STATUS(as.msg_state));
+      case TextPercent :
+	/* NOTE: no fold_tmp setup below for TextPercent style */
+      case FileTextPercent :
+	as.page_column = sc - (14 + 2*(num_len = NUM_LEN(as.total_lines)));
+	loc_len        = 17 + 2*num_len;
+	sprintf(tmp_20k_buf, "Line %*d of %*d %s    ",
+		num_len, as.current_line, 
+		num_len, as.total_lines,
+		percentage(as.current_line, as.total_lines, 1));
 	break;
     }
 
@@ -785,7 +840,20 @@ format_titlebar()
 
     /* figure folder_length and what's to be displayed */
     fold_tmp[0] = '\0';
-    if(as.style != PageNumber){
+    if(as.style == FileTextPercent || as.style == TextPercent){
+	if(as.style == FileTextPercent && !ps_global->anonymous){
+	    char *fmt    = "File: %s%s";
+	    int   avail  = sc - (ver_len + tit_len + loc_len + 10);
+	    fold_len     = strlen(as.folder_name);
+	    if(fold_len + 6 < avail) 	/* all of folder fit? */
+	      sprintf(fold_tmp, fmt, "", as.folder_name);
+	    else if((fold_len/2) + 9 < avail)
+	      sprintf(fold_tmp, fmt, "...",
+		      as.folder_name + fold_len - (avail - 9));
+	}
+	/* else leave folder/file name blank */
+    }
+    else if(as.style != PageNumber){
 	int ct_len,
 	    avail  = sc - (ver_len + tit_len + loc_len + 10);
 	fold_len   = strlen(as.folder_name);
@@ -845,7 +913,7 @@ format_titlebar()
 /*----------------------------------------------------------------------
     Update the titlebar line if the message number changed
 
-   Args: new_message_number -- message number to display
+   Args: state - pine state containing pine 
 
 This is a bit messy because the length of the number displayed might 
 change which repositions everything after it, so we adjust all the saved 
@@ -853,50 +921,48 @@ columns and shorten tail, the string holding the rest of the line.
   ----*/
 
 void
-update_titlebar_message(new_message_number)
-     int new_message_number;
+update_titlebar_message(msgs)
+    MSGNO_S *msgs;
 {
-    char  s_new[20], s_old[20];
-    int   l_change;         
+    int l_change;
 
-    if(as.cur_mess_col < 0) {
+    if(as.cur_mess_col < 0)
       return;
-    }
 
-    strcpy(s_new, comatose((long)new_message_number));
-    strcpy(s_old, comatose((long)as.current_message));
-    l_change = strlen(s_new) - strlen(s_old);
-    
+    l_change = strlen(message_string(msgs, tmp_20k_buf))
+		 - strlen(as.current_message);
 
     StartInverse();
-    if(l_change == 0) {
-        PutLine0(0, as.cur_mess_col, s_new);
-    } else {
-        if(as.style == TextPercent) {
-            PutLine5(0, as.cur_mess_col, "%s of %s %3d%% %s%s",
-                     s_new, comatose((long)as.total_messages),
-                     (void *)CALC_PERCENT(as.current_line, as.total_lines),
+    if(l_change) {
+	if(as.current_message)
+	  fs_give((void **)&as.current_message);
+
+	as.current_message = cpystr(tmp_20k_buf);
+
+        if(as.style == MsgTextPercent){
+            PutLine5(0, as.cur_mess_col, "%s of %s %s %s%s",
+                     as.current_message, comatose((long)as.total_messages),
+                     percentage(as.current_line, as.total_lines, 0),
                      BAR_STATUS(as.msg_state),
-                     repeat_char(max(0, -l_change),' '));
-            if(as.del_column >= 0) {
-                as.del_column += l_change;
-            }
-            if(as.percent_column >= 0) {
-                as.percent_column += l_change;
-            }
+                     repeat_char(max(0, -l_change), ' '));
+	    as.del_column     += l_change;
+	    as.percent_column += l_change;
         } else {
             PutLine4(0, as.cur_mess_col, "%s of %s %s%s",
-                     s_new, comatose((long)as.total_messages),
-                     BAR_STATUS(as.msg_state),
+                     as.current_message,
+		     comatose((long)as.total_messages),
+                     (mn_total_cur(msgs) > 1L)?"   ":BAR_STATUS(as.msg_state),
                      repeat_char(max(0, -l_change),' '));
-            if(as.del_column >= 0) {
-                as.del_column += l_change;
-            }
+	    as.del_column += l_change;
         }
     }
+    else if(strcmp(as.current_message, tmp_20k_buf)){
+	strcpy(as.current_message, tmp_20k_buf);
+	PutLine0(0, as.cur_mess_col, as.current_message);
+    }
+
     EndInverse();
     fflush(stdout);
-    as.current_message = new_message_number;
 }
 
 
@@ -909,31 +975,76 @@ update_titlebar_message(new_message_number)
   ----*/
 
 void
-update_titlebar_percent(new_line_number)
-     int new_line_number;
+update_titlebar_percent(line)
+     int line;
 {
-    int new_percent, old_percent;
-    
-    if(as.percent_column < 0) {
-        return;
-    }
-
-    new_percent     = CALC_PERCENT(new_line_number, as.total_lines);
-    old_percent     = CALC_PERCENT(as.current_line, as.total_lines);
-    as.current_line = new_line_number;
-
-
-   if(old_percent == new_percent)
-     return;
+    if(as.percent_column < 0)
+      return;
 
     StartInverse();
-    PutLine1(0, as.percent_column, "%3d", (void *)new_percent);
+    PutLine0(0, as.percent_column,
+	     percentage(as.current_line = line, as.total_lines, 0));
     EndInverse();
     fflush(stdout);
 }
 
 
 
+/*---------------------------------------------------------------------- 
+    Update the percentage AND line number shown in the titlebar line
+
+  Args: new_line_number -- line number to calculate new percentage
+   
+  ----*/
+
+void
+update_titlebar_lpercent(new_line_number)
+     int new_line_number;
+{
+    if(as.page_column < 0 || new_line_number == as.current_line)
+      return;
+
+    as.current_line = new_line_number;
+
+    sprintf(tmp_20k_buf, "%*d of %*d %s    ",
+	    NUM_LEN(as.total_lines), as.current_line, 
+	    NUM_LEN(as.total_lines), as.total_lines,
+	    percentage(as.current_line, as.total_lines, 0));
+    StartInverse();
+    PutLine0(0, as.page_column, tmp_20k_buf);
+    EndInverse();
+    fflush(stdout);
+}
+
+
+
+/*----------------------------------------------------------------------
+    Return static buf containing portion of lines displayed
+
+  Args:  part -- how much so far
+	 total -- how many total
+
+  ---*/
+char *
+percentage(part, total, suppress_top)
+    int part, total, suppress_top;
+{
+    static char percent[4];
+
+    if(total == 0L || (total <= ps_global->ttyo->screen_rows - 5))
+      strcpy(percent, "ALL");
+    else if(part <= ps_global->ttyo->screen_rows - 5 && !suppress_top)
+      strcpy(percent, "TOP");
+    else if(part >= total)
+      strcpy(percent, "END");
+    else
+      sprintf(percent, "%2d%%", (100L * part)/total);
+
+    return(percent);
+}
+
+
+
 /*----------------------------------------------------------------------
     Update the page number. Changes here change only the tail length
 
@@ -970,7 +1081,7 @@ update_titlebar_page(new_page)
 update_titlebar_status(mc)
     MESSAGECACHE *mc;
 {
-    if(!mc || as.del_column < 0)
+    if(!mc || as.del_column < 0 || *as.current_message == '(')
       return(0);
 
     if(mc->deleted){			/* deleted takes precedence */
@@ -1002,15 +1113,16 @@ update_titlebar_status(mc)
 
 /*
  * Save the bits from the which'th set of twelve bits in from into to.
- * An assumption is that BITMAPs are 48 bits longs.  (There must be a
+ * An assumption is that bitmap_t's are 48 bits longs.  (There must be a
  * clever macro way to do this.)
  */
 void
 savebits(to, from, which)
-    BITMAP to, from;
-    int which;
+bitmap_t to,
+	 from;
+int	 which;
 {
-    char c1, c2;
+    unsigned char c1, c2;
     int x;
 
     switch(which){
@@ -1018,16 +1130,16 @@ savebits(to, from, which)
       case 2:
 	x       =  which == 0 ? 0 : 3;
 	to[x]   =  from[x];
-	c1      =  from[x+1]  &  0xf0;
-	c2      =  to[x+1]    &  0xf;
+	c1      =  from[x+1]  &  0x0f;
+	c2      =  to[x+1]    &  0xf0;
 	to[x+1] =  c1         |  c2;
 	break;
 
       case 1:
       case 3:
 	x       =  which == 1 ? 1 : 4;
-	c1      =  to[x]    &  0xf0;
-	c2      =  from[x]  &  0xf;
+	c1      =  to[x]    &  0x0f;
+	c2      =  from[x]  &  0xf0;
 	to[x]   =  c1       |  c2;
 	to[x+1] =  from[x+1];
 	break;
@@ -1044,8 +1156,8 @@ savebits(to, from, which)
  */
 int
 equalbits(bm1, bm2, which)
-    BITMAP bm1, bm2;
-    int which;
+    bitmap_t bm1, bm2;
+    int	     which;
 {
     int x;
 
@@ -1053,12 +1165,12 @@ equalbits(bm1, bm2, which)
       case 0:
       case 2:
 	x = which == 0 ? 0 : 3;
-	return ((bm1[x] == bm2[x]) && ((bm1[x+1] & 0xf0) == (bm2[x+1] & 0xf0)));
+	return ((bm1[x] == bm2[x]) && ((bm1[x+1] & 0x0f) == (bm2[x+1] & 0x0f)));
 
       case 1:
       case 3:
 	x = which == 1 ? 1 : 4;
-	return (((bm1[x] & 0xf) == (bm2[x] & 0xf)) && (bm1[x+1] == bm2[x+1]));
+	return (((bm1[x] & 0xf0) == (bm2[x] & 0xf0)) && (bm1[x+1] == bm2[x+1]));
 
       default:
 	panic("Can't happen in equalbits()\n");
