@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: status.c,v 4.13 1993/11/19 04:10:59 mikes Exp $";
+static char rcsid[] = "$Id: status.c,v 4.27 1994/08/23 06:49:55 mikes Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -16,7 +16,7 @@ static char rcsid[] = "$Id: status.c,v 4.13 1993/11/19 04:10:59 mikes Exp $";
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1993  University of Washington
+   Copyright 1989-1994  University of Washington
 
     Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee to the University of
@@ -64,7 +64,7 @@ static char rcsid[] = "$Id: status.c,v 4.13 1993/11/19 04:10:59 mikes Exp $";
 #ifdef ANSI
 int  output_message(char *);
 void radio_help(int, int, HelpType);
-void draw_radio_prompt(int, int, char *, int, HelpType);
+void draw_radio_prompt(int, int, char *);
 
 #else
 int  output_message();
@@ -278,25 +278,43 @@ mark_status_dirty()
 
      
 /*----------------------------------------------------------------------
-        Find out how many messages are on queue to be displayed
+     Pause a suitable amount of time if a message is on the display
 
-  Args:   none
+     Note: This flags the message line as having nothing displayed.
+           The idea is that it's a function called by routines that want
+	   the message line for a prompt or something, and that they're
+	   going to obliterate the message anyway.
+ ----*/
+void
+pause_for_current_message()
+{
+    if(message_is_displayed){
+	long w = (displayed_time+currently_displayed.min_display_time)-time(0);
+	if(w > 0L)
+	  sleep(w);
 
-  Result: time to display currently displayed message if there
-          are more messages waiting to be displayed. 1 even if time
-          for current status message is 0.
+	message_is_displayed = 0;
+    }
+}
+
+
+     
+/*----------------------------------------------------------------------
+        Find out how many messages are queued for display
+
+  Args:   dtime -- will get set to minimum display time for current message
+
+  Result: number of messages in the queue.
 
-  This is used to know how long to set the timer for the next message
-  to be displayed. If the message time is 0 return 1 instead so the alarm
-  will work and all. Could go to more precise sub second timers, but try it
-  this way first.
   ---------*/
 int
-messages_queued()
+messages_queued(dtime)
+    long *dtime;
 {
-    return(queue_length <= 0 ? 0 : 
-           currently_displayed.min_display_time == 0 ?
-                                  1: currently_displayed.min_display_time);
+    if(dtime && message_is_displayed)
+      *dtime = max(currently_displayed.min_display_time, 1L);
+
+    return((ps_global->in_init_seq || queue_length < 1) ? 0 : queue_length);
 }
 
 
@@ -306,9 +324,12 @@ messages_queued()
 
    Arg: command -- The command that is about to be executed
 
-  Result:  status line cleared or
+  Result: status line cleared or
              next message queued is displayed or
              current message is redisplayed.
+	     if next message displayed, it's min display time
+	     is returned else if message already displayed, it's
+	     time remaining on the display is returned, else 0.
 
    This is called when ready to display the next message, usually just
 before reading the next command from the user. We pass in the nature
@@ -328,15 +349,15 @@ of time it takes to paint the screen when the whole screen is being painted
 is greater than the second or two delay so the time stamps set here have
 nothing to do with when the user actually sees the message.
 ----------------------------------------------------------------------*/
-void
+int
 display_message(command)
      int command;
 {
-    long now, diff;
+    long now, diff, rv = 0;
 
     if(ps_global == NULL  || ps_global->ttyo == NULL ||
        ps_global->ttyo->screen_rows <= 1 || ps_global->in_init_seq)
-      return;
+      return(0);
 
     now = time(0);
 
@@ -344,7 +365,8 @@ display_message(command)
 	/* message is showing */
 	if(command == ctrl('L')) {
 	    output_message(currently_displayed.text);
-	} else {
+	}
+	else {
 	    now = time(0);
 	    /* must be displayed for at least 2 seconds */
 	    diff = (displayed_time +
@@ -352,22 +374,14 @@ display_message(command)
                        currently_displayed.max_display_time : 
                          currently_displayed.min_display_time)
                     ) - now;
-            dprint(9, (debugfile, "diff: %d, displayed_time: %d, now: %d\n", diff, displayed_time, now));
+            dprint(9, (debugfile, "diff: %d, displayed_time: %d, now: %d\n",
+		       diff, displayed_time, now));
             if(diff > 0)
-              return;
-	    if(queue_length > 0){
-                do {
-                     /* Dequeue message and display next */
-                    currently_displayed = message_queue[0];
-                    d_q_status_message();
-                    output_message(currently_displayed.text);
-                    message_is_displayed = 1;
-                    displayed_time       = time(0);
-                } while(queue_length>0 &&
-                          currently_displayed.min_display_time==0);
-    	    } else {
+              return((int) diff);
+
+	    if(queue_length <= 0) {
     	        if(command != NO_OP_COMMAND && command != NO_OP_IDLE) {
-    		    /* Clear the current message and dequeue it */
+    		    /* Clear the message line */
                     dprint(9, (debugfile, "Clearing status line\n"));
 		    MoveCursor(ps_global->ttyo->screen_rows - 3, 0);
 		    CleartoEOLN();
@@ -376,31 +390,32 @@ display_message(command)
 		}
 	    }
 	}
-    } else {
-	/* No message is currently displayed */
-	if(queue_length > 0) {
-            do{
-                currently_displayed = message_queue[0];
-                d_q_status_message();
-	        message_is_displayed = 1;
-    	        output_message(currently_displayed.text);
-	        displayed_time       = time(0);
-                dprint(9, (debugfile, "displayed_time: %d\n", displayed_time));
-            } while(queue_length > 0 &&
-                    currently_displayed.min_display_time == 0);
-	} else {
-	    if(command == ctrl('L') || needs_clearing) {
-	        MoveCursor(ps_global->ttyo->screen_rows - 3, 0);
-	        CleartoEOLN();
-	    }
+    }
+    else if(queue_length <= 0) {
+	if(command == ctrl('L') || needs_clearing) {
+	    MoveCursor(ps_global->ttyo->screen_rows - 3, 0);
+	    CleartoEOLN();
 	}
     }
-    needs_clearing = 0; /* always cleared or overwritten */
+
+    if(queue_length > 0){		/* Dequeue message and display next */
+	do {
+	    currently_displayed = message_queue[0];
+	    d_q_status_message();
+	    output_message(currently_displayed.text);
+	    message_is_displayed = 1;
+	    displayed_time       = time(0);
+	} while(queue_length > 0
+		&& !(rv = currently_displayed.min_display_time));
+    }
+
+    needs_clearing = 0;			/* always cleared or overwritten */
     dprint(9, (debugfile,
                "STATUS DISPE command:%d,  disp:%d,  length:%d,  Time:%s",
 	       command, message_is_displayed,
 	       queue_length, ctime(&now)));
     fflush(stdout);
+    return(rv);
 }
 
 
@@ -410,6 +425,8 @@ display_message(command)
 void
 flush_status_messages()
 {
+    pause_for_current_message();
+
     while(ps_global->ttyo->screen_rows > 1 && queue_length > 0) {
         currently_displayed = message_queue[0];
         d_q_status_message();
@@ -417,6 +434,7 @@ flush_status_messages()
 	displayed_time = time(0);
         sleep(currently_displayed.min_display_time);
     }
+
     message_is_displayed = 1;
 }
 
@@ -472,6 +490,16 @@ output_message(message)
 }
 
 
+/*
+ * want_to's array passed to radio_buttions...
+ */
+static ESCKEY_S yorn[] = {
+    {'y', 'y', "Y", "Yes"},
+    {'n', 'n', "N", "No"},
+    {-1, 0, NULL, NULL}
+};
+
+
 /*----------------------------------------------------------------------
      Ask a yes/no question in the status line
 
@@ -486,17 +514,13 @@ output_message(message)
          returns y, n, dflt, or dflt_C
   ---*/
 int
-want_to(question, dflt, on_ctrl_C, help, display_help)
+want_to(question, dflt, on_ctrl_C, help, display_help, flush)
      char      *question;
      HelpType   help;
-     int    dflt, on_ctrl_C, display_help;
+     int    dflt, on_ctrl_C, display_help, flush;
 {
-    char *yorn[3], *q2;
-    int   rv;
-
-    yorn[0] = "Yes";
-    yorn[1] = "No";
-    yorn[2] = NULL;
+    char     *q2;
+    int	      rv;
 
     /*----
        One problem with adding the (y/n) here is that shrinking the 
@@ -504,19 +528,17 @@ want_to(question, dflt, on_ctrl_C, help, display_help)
        off. It would be better to truncate the question passed in
        hear and leave the full "(y/n) [x] : " on.
       ----*/
-    q2 = fs_get(strlen(question) + 17);
-    sprintf(q2, "%.*s? (y/n%s) [%c]: ", ps_global->ttyo->screen_cols - 17,
-	    question, (on_ctrl_C) ? "/^C" : "", dflt);
-    
+    q2 = fs_get(strlen(question) + 6);
+    sprintf(q2, "%.*s? ", ps_global->ttyo->screen_cols - 6, question);
     rv = radio_buttons(q2, ps_global->ttyo->screen_rows > 4 ? -3 : -1, 0,
-                       "yn", yorn, dflt, on_ctrl_C, help, display_help);
-
+                       yorn, dflt, on_ctrl_C, flush, help);
     fs_give((void **)&q2);
 
     return(rv);
 }
 
 
+
 /*----------------------------------------------------------------------
     Prompt user for a choice among alternatives
 
@@ -529,7 +551,6 @@ Args --  prompt: The prompt for the question/selection
 		   be one of the chars in buttons)
          on_ctrl_C: The selection when ^C is pressed
          help_text: Text to be displayed on bottom two lines
-         help_on:   If set the help will displayed without the user asking 
 
 Result -- Returns the letter pressed. Will be one of the characters in the
           buttons argument or one of the two deefaults.
@@ -540,63 +561,136 @@ handled along with ^Z suspension. Typing ^G will toggle the help text on and
 off. Character types that are not buttons wil result in a beep.
   ----*/
 int
-radio_buttons(prompt, line, column, buttons, labels, deefault, on_ctrl_C,
-							help_text, help_on)
-     char *prompt, *buttons, **labels;
-     HelpType help_text;
-     int  line, column, help_on, deefault, on_ctrl_C;
+radio_buttons(prompt,line,column,esc_list,deefault,on_ctrl_C,flush,help_text)
+     char	*prompt;
+     ESCKEY_S   *esc_list;
+     HelpType	 help_text;
+     int	 line, column, deefault, on_ctrl_C, flush;
 {
-    register int  ch, real_line;
-    char         *q, *p, **ll;
-    int           cursor_moved, max_label;
+    register int     ch, real_line;
+    char            *q, *ds = NULL;
+    int              cursor_moved, max_label, i, start, fkey_table[12];
+    struct key	     rb_keys[12];
+    struct key_menu  rb_keymenu;
+    bitmap_t	     bitmap;
 
-    /*----
-       Give the user time to see the status message that was
-      just put up if there was one.
-      ----*/
-    if(messages_queued())
-      sleep(3);
-
-    needs_clearing = 1;
-
-    mark_status_dirty();
+    pause_for_current_message();	/* give user time to see status */
+    mark_status_dirty();		/* clear message next display call */
     real_line = line > 0 ? line : ps_global->ttyo->screen_rows + line;
     MoveCursor(real_line, column);
     CleartoEOLN();
 
     /*---- Find longest label ----*/
     max_label = 0;
-    for(ll = labels; *ll != NULL; ll++)
-      max_label = max(max_label, strlen(*ll));
-    
+    for(i = 0; esc_list && esc_list[i].ch != -1 && i < 12; i++){
+      if(esc_list[i].ch == -2) /* -2 means to skip this key and leave blank */
+	continue;
+      max_label = max(max_label, strlen(esc_list[i].name));
+    }
+
     q = cpystr(prompt); /* So we can truncate string below if need be */
     if(strlen(q) + max_label > ps_global->ttyo->screen_cols) 
         q[ps_global->ttyo->screen_cols - max_label] = '\0';
 
-    draw_radio_prompt(real_line, column, q, help_on, help_text);
+    /*---- Init structs for keymenu ----*/
+    for(i = 0; i < 12; i++)
+      memset((void *)&rb_keys[i], 0, sizeof(struct key));
+
+    memset((void *)&rb_keymenu, 0, sizeof(struct key_menu));
+    rb_keymenu.how_many = 1;
+    rb_keymenu.keys     = rb_keys;
+
+    /*---- Setup key menu ----*/
+    start = 0;
+    clrbitmap(bitmap);
+    memset(fkey_table, NO_OP_COMMAND, 12 * sizeof(int));
+    if(help_text != NO_HELP){		/* if shown, always at position 0 */
+	rb_keymenu.keys[0].name  = "^G";
+	rb_keymenu.keys[0].label = "Help";
+	setbitn(0, bitmap);
+	fkey_table[0] = ctrl('G');
+	start++;
+    }
+
+    if(on_ctrl_C){			/* if shown, always at position 1 */
+	rb_keymenu.keys[1].name  = "^C";
+	rb_keymenu.keys[1].label = "Cancel";
+	setbitn(1, bitmap);
+	fkey_table[1] = ctrl('C');
+	start++;
+    }
+
+    start = (start) ? 2 : 0;
+    /*---- Show the usual possible keys ----*/
+    for(i=start; esc_list && esc_list[i-start].ch != -1 && i < 12; i++){
+	if(esc_list[i-start].ch != -2)
+	    setbitn(i, bitmap);
+	fkey_table[i] = esc_list[i-start].ch;
+	rb_keymenu.keys[i].name  = esc_list[i-start].name;
+	if(esc_list[i-start].rval == deefault){
+	    ds = (char *)fs_get((strlen(esc_list[i-start].label) + 3)
+				* sizeof(char));
+	    sprintf(ds, "[%s]", esc_list[i-start].label);
+	    rb_keymenu.keys[i].label = ds;
+	}
+	else
+	  rb_keymenu.keys[i].label = esc_list[i-start].label;
+    }
+
+    for(; i < 12; i++)
+      rb_keymenu.keys[i].name = NULL;
+
+    ps_global->mangled_footer = 1;
+
+    draw_radio_prompt(real_line, column, q);
 
     do {
         fflush(stdout);
         fflush(stdin);
 
+	/*---- Paint the keymenu ----*/
+	EndInverse();
+	draw_keymenu(&rb_keymenu, bitmap, ps_global->ttyo->screen_cols,
+		     -2, 0, FirstMenu, 0);
+	StartInverse();
+	MoveCursor(real_line, column + strlen(q));
+
+	if(flush)
+	  flush_input();
+
 	/* Timeout 5 min to keep imap mail stream alive */
         ch = read_char(600);
         dprint(2, (debugfile,
                    "Want_to read: %s (%d)\n", pretty_command(ch), ch));
         ch = isupper(ch) ? tolower(ch) : ch;
 
+	if(F_ON(F_USE_FK,ps_global)
+	   && ((isalpha(ch) && !strchr("YyNn", ch)) /* y & n : special case */
+	       || ((ch >= PF1 && ch <= PF12)
+		   && (ch = fkey_table[ch - PF1]) == NO_OP_COMMAND))){
+	    /*
+	     * The funcky test above does two things.  First, it maps
+	     * esc_list character commands to function keys, *and* prevents
+	     * character commands from input while in function key mode.
+	     * NOTE: this breaks if we ever need more than the first
+	     * twelve function keys...
+	     */
+	    Writechar('\007', 0);
+	    continue;
+	}
+
         switch(ch) {
 
           default:
-            for(p = buttons; ch!=(int)*p && *p!='\0'; p++);/* Which button */
-            if(*p == '\0') {
-                Writechar('\007', 0);
-                break;
-            }
-            dprint(9, (debugfile, "label %d %s\n", p - buttons,
-                       labels[p-buttons]));
-            Write_to_screen(labels[p - buttons]);
-            goto out_of_loop;
+	    for(i = 0; esc_list && esc_list[i].ch != -1; i++)
+	      if(ch == esc_list[i].ch){
+		  Write_to_screen(esc_list[i].label);
+		  ch = esc_list[i].rval;
+		  goto out_of_loop;
+	      }
+
+	    Writechar('\007', 0);
+	    break;
 
           case ctrl('M'):
           case ctrl('J'):
@@ -617,15 +711,14 @@ radio_buttons(prompt, line, column, buttons, labels, deefault, on_ctrl_C,
           case ctrl('G'):
             if(help_text == NO_HELP)
               break;
-            if(help_on) {
-                MoveCursor(real_line + 1, column);
-                CleartoEOLN();
-                MoveCursor(real_line + 2, column);
-                CleartoEOLN();
-            } else {
-                radio_help(real_line, column, help_text);
-            }
-            help_on = ! help_on;
+
+	    MoveCursor(real_line + 1, column);
+	    CleartoEOLN();
+	    MoveCursor(real_line + 2, column);
+	    CleartoEOLN();
+	    radio_help(real_line, column, help_text);
+	    sleep(5);
+
             MoveCursor(real_line, column + strlen(q));
             break;
             
@@ -647,7 +740,7 @@ radio_buttons(prompt, line, column, buttons, labels, deefault, on_ctrl_C,
             if(ps_global->redrawer != NULL)
               (*ps_global->redrawer)();
             redraw_keymenu();
-            draw_radio_prompt(real_line, column, q, help_on, help_text);
+            draw_radio_prompt(real_line, column, q);
             break;
 
             
@@ -658,7 +751,7 @@ radio_buttons(prompt, line, column, buttons, labels, deefault, on_ctrl_C,
                 EndInverse();
                 do_suspend(ps_global);
                 clear_cursor_pos();
-                draw_radio_prompt(real_line, column, q, help_on, help_text);
+                draw_radio_prompt(real_line, column, q);
             }
             break;
 
@@ -667,6 +760,9 @@ radio_buttons(prompt, line, column, buttons, labels, deefault, on_ctrl_C,
 
   out_of_loop:
     fs_give((void *)&q);
+    if(ds)
+      fs_give((void *)&ds);
+
     EndInverse();
     fflush(stdout);
     return(ch);
@@ -694,10 +790,14 @@ radio_help(line, column, help)
     EndInverse();
     MoveCursor(line + 1, column);
     CleartoEOLN();
-    PutLine0(line + 1, column, text[0]);
+    if(text[0])
+      PutLine0(line + 1, column, text[0]);
+
     MoveCursor(line + 2, column);
     CleartoEOLN();
-    PutLine0(line + 2, column, text[1]);
+    if(text[1])
+      PutLine0(line + 2, column, text[1]);
+
     StartInverse();
 #ifdef	DOS
     free_help_text(text);
@@ -710,10 +810,9 @@ radio_help(line, column, help)
    Paint the screen with the radio buttons prompt
   ----*/
 void
-draw_radio_prompt(l, c, q, help_on, help_text)
-     int l, c, help_on;
+draw_radio_prompt(l, c, q)
+     int       l, c;
      char     *q;
-     HelpType  help_text;
 {
     int x;
 
@@ -723,8 +822,6 @@ draw_radio_prompt(l, c, q, help_on, help_text)
     MoveCursor(l, x);
     while(x++ < ps_global->ttyo->screen_cols)
       Writechar(' ', 0);
-    if(help_on) 
-      radio_help(l, c, help_text);
     MoveCursor(l, c + strlen(q));
     fflush(stdout);
 }
