@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: ttyin.c,v 4.26 1993/11/18 01:43:50 mikes Exp $";
+static char rcsid[] = "$Id: ttyin.c,v 4.52 1994/08/18 17:40:15 mikes Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -16,7 +16,7 @@ static char rcsid[] = "$Id: ttyin.c,v 4.26 1993/11/18 01:43:50 mikes Exp $";
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1993  University of Washington
+   Copyright 1989-1994  University of Washington
 
     Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee to the University of
@@ -161,11 +161,11 @@ int state;
 
     if (state == 0 && _inraw) {
         /*----- restore state to original -----*/
-#ifdef HAVE_TERMIOS
+#ifdef	HAVE_TERMIOS
 	if (tcsetattr (STDIN_FD, TCSADRAIN, &_original_tty) < 0)
 		return -1;
-#else  /* HAVE_TERMIOS */
-#ifdef HAVE_TERMIO
+#else	/* HAVE_TERMIOS */
+#ifdef	HAVE_TERMIO
         if(ioctl(STDIN_FD, TCSETAW, &_original_tty) < 0)
           return(-1);
 #else
@@ -174,21 +174,22 @@ int state;
 	(void) ioctl(STDIN_FD, TIOCSLTC, &_original_ltchars);
 	(void) ioctl(STDIN_FD, TIOCSETC, &_original_tchars);
         (void) ioctl(STDIN_FD, TIOCLSET, &_original_lmode);
-#endif /* HAVE_TERMIO */
-#endif /* HAVE_TERMIOS */
+#endif	/* HAVE_TERMIO */
+#endif	/* HAVE_TERMIOS */
         _inraw = 0;
     } else if (state == 1 && ! _inraw) {
         /*----- Go into raw mode (cbreak actually) ----*/
 
-#ifdef HAVE_TERMIOS
+#ifdef	HAVE_TERMIOS
 	if (tcgetattr (STDIN_FD, &_original_tty) < 0)
 		return -1;
 	tcgetattr (STDIN_FD, &_raw_tty);
-	_raw_tty.c_lflag &= ~(ICANON | ECHO | IEXTEN);	/* noecho raw mode  */
+	_raw_tty.c_lflag &= ~(ICANON | ECHO | IEXTEN); /* noecho raw mode    */
 
- 	_raw_tty.c_lflag &= ~ISIG;            /* disable signals */
- 	_raw_tty.c_iflag &= ~ICRNL;           /* turn off CR->NL on input */
- 	_raw_tty.c_oflag &= ~ONLCR;           /* turn off NL->CR on output */
+ 	_raw_tty.c_lflag &= ~ISIG;		/* disable signals           */
+ 	_raw_tty.c_iflag &= ~IXON;		/* turn off ^S/^Q on input   */
+ 	_raw_tty.c_iflag &= ~ICRNL;		/* turn off CR->NL on input  */
+ 	_raw_tty.c_oflag &= ~ONLCR;		/* turn off NL->CR on output */
  	_raw_tty.c_oflag &= ~OPOST;
 
 	/* Only go into 8 bit mode if we are doing something other
@@ -198,31 +199,32 @@ int state;
 	 */
         if(ps_global->VAR_CHAR_SET && ps_global->VAR_CHAR_SET[0] &&
 	   strucmp(ps_global->VAR_CHAR_SET, "us-ascii"))
-	  _raw_tty.c_iflag &= ~ISTRIP; 	       /* turn off hi bit stripping */
+	  _raw_tty.c_iflag &= ~ISTRIP;		/* turn off hi bit stripping */
 
-	_raw_tty.c_cc[VMIN]  = '\01'; /* minimum # of chars to queue  */
-	_raw_tty.c_cc[VTIME] = '\0';  /* minimum time to wait for input */
+	_raw_tty.c_cc[VMIN]  = '\01';		/* min # of chars to queue  */
+	_raw_tty.c_cc[VTIME] = '\0';		/* min time to wait for input*/
 	ps_global->low_speed = cfgetospeed(&_raw_tty) < B4800;
 	tcsetattr (STDIN_FD, TCSADRAIN, &_raw_tty);
 
 #else
-#ifdef HAVE_TERMIO
+#ifdef	HAVE_TERMIO
         if(ioctl(STDIN_FD, TCGETA, &_original_tty) < 0)
           return(-1);
 	(void) ioctl(STDIN_FD, TCGETA, &_raw_tty);    /** again! **/
 
 	_raw_tty.c_lflag &= ~(ICANON | ECHO);	/* noecho raw mode  */
 
- 	_raw_tty.c_lflag &= ~ISIG;            /* disable signals */
- 	_raw_tty.c_iflag &= ~ICRNL;           /* turn off CR->NL on input */
- 	_raw_tty.c_oflag &= ~ONLCR;           /* turn off NL->CR on output */
+ 	_raw_tty.c_lflag &= ~ISIG;		/* disable signals */
+ 	_raw_tty.c_iflag &= ~IXON;		/* turn off ^S/^Q on input   */
+ 	_raw_tty.c_iflag &= ~ICRNL;		/* turn off CR->NL on input  */
+ 	_raw_tty.c_oflag &= ~ONLCR;		/* turn off NL->CR on output */
  	_raw_tty.c_oflag &= ~OPOST;
-	_raw_tty.c_cc[VMIN]  = 1;  /* minimum # of chars to queue  */
-	_raw_tty.c_cc[VTIME] = 0;  /* minimum time to wait for input */
-        ps_global->low_speed = (_raw_tty.c_cflag & CBAUD) < (unsigned int)B4800;
+	_raw_tty.c_cc[VMIN]  = 1;		/* min # of chars to queue   */
+	_raw_tty.c_cc[VTIME] = 0;		/* min time to wait for input*/
+        ps_global->low_speed = (_raw_tty.c_cflag&CBAUD) < (unsigned int)B4800;
 	(void) ioctl(STDIN_FD, TCSETAW, &_raw_tty);
 
-#else/* HAVE_TERMIO */
+#else	/* HAVE_TERMIO */
         if(ioctl(STDIN_FD, TIOCGETP, &_original_tty) < 0)
           return(-1);
 	(void) ioctl(STDIN_FD, TIOCGETP, &_raw_tty);   
@@ -235,17 +237,25 @@ int state;
 
 	_raw_tty.sg_flags &= ~(ECHO);	/* echo off */
 	_raw_tty.sg_flags |= CBREAK;	/* raw on    */
-        _raw_tty.sg_flags &= ~CRMOD;    /* Turn off CR -> LF mapping */
+        _raw_tty.sg_flags &= ~CRMOD;	/* Turn off CR -> LF mapping */
 
-	_raw_tchars.t_intrc = -1; /* Turn off ^C and ^D */
+	_raw_tchars.t_intrc = -1;	/* Turn off ^C and ^D */
 	_raw_tchars.t_eofc  = -1;
 
-	_raw_ltchars.t_lnextc = -1;   /* Turn off ^V so we can use it */
-	_raw_ltchars.t_dsuspc = -1;   /* Turn off ^Y so we can use it */
-	_raw_ltchars.t_suspc  = -1;   /* Turn off ^Z; we just read 'em */
-	_raw_ltchars.t_werasc = -1;   /* Turn off ^w word erase */
-	_raw_ltchars.t_rprntc = -1;   /* Turn off ^R reprint line */
-        _raw_ltchars.t_flushc = -1;   /* Turn off ^O output flush */
+	_raw_tchars.t_startc = -1;	/* Turn off ^S/^Q */
+	_raw_tchars.t_stopc  = -1;
+
+#ifdef	DEBUG
+	if(debug < 9)			/* only on if full debugging set */
+#endif
+	_raw_tchars.t_quitc = -1;	/* Turn off ^\ */
+
+	_raw_ltchars.t_lnextc = -1;	/* Turn off ^V so we can use it */
+	_raw_ltchars.t_dsuspc = -1;	/* Turn off ^Y so we can use it */
+	_raw_ltchars.t_suspc  = -1;	/* Turn off ^Z; we just read 'em */
+	_raw_ltchars.t_werasc = -1;	/* Turn off ^w word erase */
+	_raw_ltchars.t_rprntc = -1;	/* Turn off ^R reprint line */
+        _raw_ltchars.t_flushc = -1;	/* Turn off ^O output flush */
 
 	/* Only go into 8 bit mode if we are doing something other
 	 * than plain ASCII. This will save the folks that have
@@ -255,7 +265,7 @@ int state;
         if(ps_global->VAR_CHAR_SET && ps_global->VAR_CHAR_SET[0] &&
 	   strucmp(ps_global->VAR_CHAR_SET, "us-ascii")) {
 	    _raw_lmode |= LPASS8;
-#ifdef NXT  /* Hope there aren't many specific like this! */
+#ifdef	NXT				/* Hope there aren't many like this! */
             _raw_lmode |= LPASS8OUT;
 #endif
 	}
@@ -326,9 +336,11 @@ read_with_timeout(time_out)
 #ifdef USE_POLL
 	pollfd.fd = STDIN_FD;
 	pollfd.events = POLLIN;
-	dprint(9,(debugfile,"poll event %d, timeout %d\n", pollfd.events, time_out));
+	dprint(9,(debugfile,"poll event %d, timeout %d\n", pollfd.events,
+		  time_out));
 	res = poll (&pollfd, 1, time_out * 1000);
-	dprint(9, (debugfile, "poll on tty returned %d, events %d\n", res, pollfd.revents));
+	dprint(9, (debugfile, "poll on tty returned %d, events %d\n",
+		   res, pollfd.revents));
 #else
          if(STDIN_FD == 0)
            readfds    = 1;
@@ -337,7 +349,8 @@ read_with_timeout(time_out)
          tmo.tv_sec  = time_out;
          tmo.tv_usec = 0; 
 
-         dprint(9,(debugfile,"Select readfds:%d timeval:%d,%d\n",readfds, tmo.tv_sec,tmo.tv_usec));
+         dprint(9, (debugfile,"Select readfds:%d timeval:%d,%d\n", readfds,
+		   tmo.tv_sec,tmo.tv_usec));
 
          res = select(STDIN_FD+1, &readfds, 0, &readfds, &tmo);
 
@@ -418,7 +431,6 @@ read_with_timeout(time_out)
   escape sequence. The timeout will effectively cancel the escape sequence.
 
   ---*/
-#ifndef OLD_READ_CHAR
 int
 read_char(time_out)
      int time_out;
@@ -484,7 +496,7 @@ get_one_char:
               case 'B': return(KEY_DOWN);
               case 'C': return(KEY_RIGHT);
               case 'D': return(KEY_LEFT);
-              default: return(KEY_JUNK);
+	      /* default: return junk below */
             }
         } else if(ch == '[') {
             /* For dec terminals, vt200s, and some weird Sun stuff */
@@ -516,8 +528,10 @@ get_one_char:
                   case 'j': return(PF10);
                   case 'k': return(PF11);
                   case 'l': return(PF12);
-                  default: return(KEY_JUNK);
+		  /* default: return junk below */
                 }
+		break;
+
               case '1': /* Sun keys */
                   rx = KEY_JUNK; goto swallow_till_z;
     
@@ -569,8 +583,7 @@ get_one_char:
                       ch &= 0x7f;
                   } 
                   return(rx);			 
-                      
-                         
+
               default:
                 /* DEC function keys */
                 num_keys = 0;
@@ -582,7 +595,7 @@ get_one_char:
 		      goto done;
                     ch &= 0x7f;
                 } while (num_keys++ < 6 && ch != '~');
-                return(KEY_JUNK);
+	       /* return junk below */
             }   
         } else if(ch == '?') {
              /* DEC vt52 application keys, and some Zenith 19 */
@@ -595,7 +608,7 @@ get_one_char:
                case 't': return(KEY_LEFT);
                case 'v': return(KEY_RIGHT);
                case 'x': return(KEY_UP);
-               default: return(KEY_JUNK);
+	       /* default: return junk below */
              }    
         } else if(ch == '\033'){
 	     /* special hack to get around comm devices eating 
@@ -610,18 +623,19 @@ get_one_char:
 
 	     return((isalpha(ch) || ch == '@' || (ch >= '[' && ch <= '_'))
 		    ? ctrl(ch) : ch);
-	    
 	} else {
-             /* This gets most Z19 codes, and some VT52 modes */
-             switch(ch) {
-               case 'A': return(KEY_UP);
-               case 'B': return(KEY_DOWN);
-               case 'C': return(KEY_RIGHT);
-               case 'D': return(KEY_LEFT);
-               default:  return(KEY_JUNK);
-             }    
+	    /* This gets most Z19 codes, and some VT52 modes */
+	    /* NOTE: A == KEY_UP, B == KEY_DOWN... */
+	    return((ch < 'A' || ch > 'D') ? KEY_JUNK : (KEY_UP + (ch - 'A')));
         }
 
+	/*
+	 * If we got here we didn't have anything useful to return, so
+	 * just let the caller know we got junk...
+	 */
+	return(KEY_JUNK);
+
+
      default:
      done:
         dprint(9, (debugfile, "Read char returning: %d %s\n",
@@ -631,168 +645,6 @@ get_one_char:
 }
 
 
-#else
-int
-read_char(time_out)
-     int time_out;
-{
-    register int  rx, ch, num_keys;
-
-    rx = 0; /* avoid ref before set errors */
-    ch = read_with_timeout(time_out);
-    if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND || ch == KEY_RESIZE) goto done;
-    ch &= 0x7f;
-    switch(ch) {
-    
-      case '\033':
-        ch = read_with_timeout(time_out);
-        if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND || ch == KEY_RESIZE)
-          goto done;
-        ch &= 0x7f;
-        if(ch == 'O') {
-            /* For DEC terminals, vt100s */
-            ch = read_with_timeout(time_out);
-            if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND || ch == KEY_RESIZE)
-              goto done;
-            ch &= 0x7f;
-            switch(ch) {
-              case 'P': return(PF1);
-              case 'Q': return(PF2);
-              case 'R': return(PF3);
-              case 'S': return(PF4);
-              case 'p': return(PF5);
-              case 'q': return(PF6);
-              case 'r': return(PF7);
-              case 's': return(PF8);
-              case 't': return(PF9);
-              case 'u': return(PF10);
-              case 'v': return(PF11);
-              case 'w': return(PF12);
-              case 'A': return(KEY_UP);
-              case 'B': return(KEY_DOWN);
-              case 'C': return(KEY_RIGHT);
-              case 'D': return(KEY_LEFT);
-              default: return(KEY_JUNK);
-            }
-        } else if(ch == '[') {
-            /* For dec terminals, vt200s, and some weird Sun stuff */
-            ch = read_with_timeout(time_out);
-            if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND || ch == KEY_RESIZE)
-              goto done;
-            ch &= 0x7f;
-            switch(ch) {
-              case 'A': return(KEY_UP);
-              case 'B': return(KEY_DOWN);
-              case 'C': return(KEY_RIGHT);
-              case 'D': return(KEY_LEFT);
-    
-              case '=': /* ansi terminal function keys */
-                ch = read_with_timeout(time_out);
-                if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND || ch == KEY_RESIZE)
-                  goto done;
-                ch &= 0x7f;
-                switch(ch) {
-                  case 'a': return(PF1);
-                  case 'b': return(PF2);
-                  case 'c': return(PF3);
-                  case 'd': return(PF4);
-                  case 'e': return(PF5);
-                  case 'f': return(PF6);
-                  case 'g': return(PF7);
-                  case 'h': return(PF8);
-                  case 'i': return(PF9);
-                  case 'j': return(PF10);
-                  case 'k': return(PF11);
-                  case 'l': return(PF12);
-                  default: return(KEY_JUNK);
-                }
-              case '1': /* Sun keys */
-                  rx = KEY_JUNK; goto swallow_till_z;
-    
-              case '2': /* Sun keys */
-                  ch = read_with_timeout(time_out);
-                  if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND ||
-                     ch == KEY_RESIZE)
-                    goto done;
-                  ch &= 0x7f;
-                  if(ch == '1') {
-                      ch = read_with_timeout(time_out);
-                      if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND ||
-                         ch == KEY_RESIZE) goto done;
-                      ch &= 0x7f;
-                      switch (ch) {
-                        case '5':  rx = KEY_UP; break;
-                        case '7':  rx = KEY_LEFT; break;
-                        case '9':  rx = KEY_RIGHT; break;
-                        default:   rx = KEY_JUNK;
-                      }
-                  } else if (ch == '2') {
-                      ch = read_with_timeout(time_out);
-                      if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND ||
-                         ch == KEY_RESIZE) goto done;
-                      ch &= 0x7f;
-                      if(ch == '1')
-                        rx = KEY_DOWN;
-                      else
-                        rx = KEY_JUNK;
-                  }
-                    swallow_till_z:		  
-    
-		  num_keys = 0;
-                  do {
-                      ch = read_with_timeout(time_out);
-                      if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND ||
-                         ch == KEY_RESIZE) goto done;
-                      ch &= 0x7f;
-                  } while (num_keys++ < 4 && ch != 'z');
-                  return(rx);			 
-                      
-                         
-              default:
-                /* DEC function keys */
-                num_keys = 0;
-                do {
-                    ch = read_with_timeout(time_out);
-                    if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND ||
-                       ch == KEY_RESIZE) goto done;
-                    ch &= 0x7f;
-                } while (num_keys++ < 6 && ch != '~');
-                return(KEY_JUNK);
-            }   
-        } else if(ch == '?') {
-             /* DEC vt52 application keys, and some Zenith 19 */
-             ch = read_with_timeout(time_out);
-             if(ch == NO_OP_IDLE || ch == NO_OP_COMMAND ||
-                ch == KEY_RESIZE) goto done;
-             ch &= 0x7f;
-             switch(ch) {
-               case 'r': return(KEY_DOWN);
-               case 't': return(KEY_LEFT);
-               case 'v': return(KEY_RIGHT);
-               case 'x': return(KEY_UP);
-               default: return(KEY_JUNK);
-             }    
-        } else {
-             /* This gets most Z19 codes, and some VT52 modes */
-             switch(ch) {
-               case 'A': return(KEY_UP);
-               case 'B': return(KEY_DOWN);
-               case 'C': return(KEY_RIGHT);
-               case 'D': return(KEY_LEFT);
-               default:  return(KEY_JUNK);
-             }    
-        }
-
-     default:
-     done:
-        dprint(9, (debugfile, "Read char returning: %d %s\n",
-                   ch, pretty_command(ch)));
-        return(ch);
-    }
-}
-#endif
-
-
 
 /*----------------------------------------------------------------------
          Highest level read for reading Pine commands.    (UNIX)
@@ -807,9 +659,14 @@ things.
 int
 read_command()
 {
-    int ch, tm;
+    int ch, tm = 0;
+
+    if(timeout > 0){
+	long dtime;
+
+	tm = messages_queued(&dtime) ? (int)dtime : timeout;
+    }
 
-    tm = timeout > 0 ? messages_queued() ? messages_queued() : timeout : 0;
     ch = read_char(tm);
     dprint(9, (debugfile, "Read command returning: %d %s\n", ch,
               pretty_command(ch)));
@@ -820,7 +677,7 @@ read_command()
 
 
 
-extern char termname[]; /* termname from ttyout.c-- affect keyboard*/
+extern char term_name[]; /* term_name from ttyout.c-- affect keyboard*/
 /* -------------------------------------------------------------------
      Set up the keyboard -- usually enable some function keys  (UNIX)
 
@@ -836,9 +693,9 @@ void
 init_keyboard(use_fkeys)
      int use_fkeys;
 {
-    if(use_fkeys && (!strucmp(termname,"vt102") || !strucmp(termname,"vt100")))
+    if(use_fkeys && (!strucmp(term_name,"vt102")
+		     || !strucmp(term_name,"vt100")))
       printf("\033\133\071\071\150");
-
 }
 
 
@@ -854,11 +711,42 @@ void
 end_keyboard(use_fkeys)
      int use_fkeys;
 {
-    if(use_fkeys && (!strcmp(termname, "vt102") || !strcmp(termname, "vt100"))){
-      printf("\033\133\071\071\154");
-      fflush(stdout);
+    if(use_fkeys && (!strcmp(term_name, "vt102")
+		     || !strcmp(term_name, "vt100"))){
+	printf("\033\133\071\071\154");
+	fflush(stdout);
     }
+}
 
+    
+/*----------------------------------------------------------------------
+   Discard any pending input characters				(UNIX)
+
+   Args:  none
+
+ Result: pending input buffer flushed
+  ----*/
+void
+flush_input()
+{
+#ifdef	HAVE_TERMIOS
+    tcflush(STDIN_FD, TCIFLUSH);
+#else
+#ifdef	HAVE_TERMIO
+    ioctl(STDIN_FD, TCFLSH, 0);
+#else
+#ifdef	TIOCFLUSH
+#ifdef	FREAD
+    int i = FREAD;
+#else
+    int i = 1;
+#endif
+
+    ioctl(STDIN_FD, TIOCFLUSH, &i);
+#else
+#endif	/* TIOCFLUSH */
+#endif	/* HAVE_TERMIO */
+#endif	/* HAVE_TERMIOS */
 }
 
     
@@ -872,10 +760,20 @@ static int enhanced = 0;
 /* global function to execute while waiting for character input */
 void   (*while_waiting)() = NULL;
 
-/* 
-   This code compiled once, but it out of date with respect to the rest of
-  Pine. Substantial work is needed to make it run
- */
+#ifdef _WINDOWS
+/* global to tell us if the window was resized. */
+static int DidResize = FALSE;
+
+
+/*----------------------------------------------------------------------
+    Flag the fact the window has resized.
+*/
+int
+pine_window_resize_callback ()
+{
+    DidResize = TRUE;
+}
+#endif
 
 
 
@@ -889,11 +787,16 @@ void   (*while_waiting)() = NULL;
 init_tty_driver(pine)
      struct pine *pine;
 {
-    union REGS in;
+#ifdef _WINDOWS
+    extern int init_mouse ();
 
+    mswin_setresizecallback (pine_window_resize_callback);
+    init_mouse ();			/* always a mouse under windows? */
+#else
     /* detect enhanced keyboard */
     enhanced = enhanced_keybrd();	/* are there extra keys? */
-    pine     = pine;			/* Get rid of unused parm warning */
+#endif
+    pine = pine;			/* Get rid of unused parm warning */
     return(Raw(1));
 }
 
@@ -911,6 +814,9 @@ end_tty_driver(pine)
      struct pine *pine;
 {
     dprint(2, (debugfile, "about to end_tty_driver\n"));
+#ifdef _WINDOWS
+    mswin_clearresizecallback (pine_window_resize_callback);
+#endif
 }
 
 /*----------------------------------------------------------------------
@@ -924,6 +830,32 @@ extended_code(kc)
 unsigned  kc;
 {
     switch(kc){
+#ifdef	_WINDOWS
+	case MSWIN_KEY_F1: return(PF1);
+	case MSWIN_KEY_F2: return(PF2);
+	case MSWIN_KEY_F3: return(PF3);
+	case MSWIN_KEY_F4: return(PF4);
+	case MSWIN_KEY_F5: return(PF5);
+	case MSWIN_KEY_F6: return(PF6);
+	case MSWIN_KEY_F7: return(PF7);
+	case MSWIN_KEY_F8: return(PF8);
+	case MSWIN_KEY_F9: return(PF9);
+	case MSWIN_KEY_F10: return(PF10);
+	case MSWIN_KEY_F11: return(PF11);
+	case MSWIN_KEY_F12: return(PF11);
+
+	case MSWIN_KEY_UP: return(KEY_UP);
+	case MSWIN_KEY_DOWN: return(KEY_DOWN);
+	case MSWIN_KEY_LEFT: return(KEY_LEFT);
+	case MSWIN_KEY_RIGHT: return(KEY_RIGHT);
+	case MSWIN_KEY_HOME: return(KEY_HOME);
+	case MSWIN_KEY_END: return(KEY_END);
+	case MSWIN_KEY_PREVPAGE: return(KEY_PGUP);
+	case MSWIN_KEY_NEXTPAGE: return(KEY_PGDN);
+	case MSWIN_KEY_DELETE: return(KEY_DEL);
+
+	case MSWIN_KEY_NODATA:	return (NO_OP_COMMAND);
+#else
 	case 0x3b00 : return(PF1);
 	case 0x3c00 : return(PF2);
 	case 0x3d00 : return(PF3);
@@ -936,6 +868,7 @@ unsigned  kc;
 	case 0x4400 : return(PF10);
 	case 0x8500 : return(PF11);
 	case 0x8600 : return(PF12);
+
 	case 0x4800 : return(KEY_UP);
 	case 0x5000 : return(KEY_DOWN);
 	case 0x4b00 : return(KEY_LEFT);
@@ -954,6 +887,7 @@ unsigned  kc;
 	case 0x49e0 : return(KEY_PGUP);			/* grey key version */
 	case 0x51e0 : return(KEY_PGDN);			/* grey key version */
 	case 0x53e0 : return(KEY_DEL);			/* grey key version */
+#endif
 	default   : return(NO_OP_COMMAND);
     }
 }
@@ -985,11 +919,12 @@ int
 read_char(tm)
 int tm;
 {
-    unsigned ch = 0;
+    unsigned ch = 0, intrupt = 0;
     long timein;
     static int firsttime = 1;
-
+    extern int win_multiplex();
 #ifdef	MOUSE
+    /* Imported from os_dos.c or os_win.c. */
     extern int mexist;
     extern int checkmouse();
     extern void mouseon(), mouseoff();
@@ -1017,29 +952,57 @@ int tm;
         return(ctrl('L'));
       }
     }
+
 get_one_char:
 
+#ifdef _WINDOWS
+    if (DidResize) {
+	 DidResize = FALSE;
+	 return (get_windsize (ps_global->ttyo));
+    }
+#endif
+
 #ifdef	MOUSE
     mouseon();
 #endif
 
     if(tm){
 	timein = time(0L);
+#ifdef _WINDOWS
+	/* mswin_charavail() Yeilds control to other window apps. */
+	while (!mswin_charavail()) {
+#else
 	while(!_bios_keybrd(enhanced ? _NKEYBRD_READY : _KEYBRD_READY)){
+#endif
 	    if(time(0L) >= timein+tm){
 		ch = NO_OP_COMMAND;
 		goto gotone;
 	    }
+#ifdef _WINDOWS
+	    if (DidResize) {
+		DidResize = FALSE;
+		return( get_windsize (ps_global->ttyo));
+	    }
+#endif
 #ifdef	MOUSE
  	    if(checkmouse(&ch))
 	      goto gotone;
 #endif
 	    if(while_waiting)
 	      (*while_waiting)();
+
+#ifndef	_WINDOWS
+	    if(!intrupt++)		/* surrender CPU to windows */
+	      win_multiplex();
+#endif
 	}
     }
 
+#ifdef _WINDOWS
+    ch = mswin_getc_fast();
+#else
     ch = _bios_keybrd(enhanced ? _NKEYBRD_READ : _KEYBRD_READ);
+#endif
 
 gotone:
 #ifdef	MOUSE
@@ -1052,8 +1015,16 @@ gotone:
      * by the mouse event catcher.  For now, just special case them
      * since they don't conflict with any of the DOS special keys.
      */
+#ifdef _WINDOWS
+
+    if (ch >= MSWIN_RANGE_START && ch <= MSWIN_RANGE_END)
+	    return (extended_code (ch));
+
+    return (ch&0xff);
+#else /* DOS */
     return((ch >= PF1 && ch <= PF12) ? ch : 
 	((ch&0xff) && ((ch&0xff) != 0xe0)) ? (ch&0xff) : extended_code(ch));
+#endif
 }
 
 
@@ -1085,12 +1056,31 @@ void
 end_keyboard(use_fkeys)
      int use_fkeys;
 {
-
 }
 
 
 
+/*----------------------------------------------------------------------
+   Discard any pending input characters				(DOS)
 
+   Args:  none
+
+ Result: pending input buffer flushed
+  ----*/
+void
+flush_input()
+{
+#ifdef _WINDOWS
+    while (mswin_charavail ())
+        (void) mswin_getc ();
+#else
+    while(_bios_keybrd(enhanced ? _NKEYBRD_READY : _KEYBRD_READY))
+      (void) _bios_keybrd(enhanced ? _NKEYBRD_READ : _KEYBRD_READ);
+#endif
+}
+
+
+    
 /*----------------------------------------------------------------------
         Read a character from keyboard with timeout               (DOS)
  Input:  none
@@ -1104,9 +1094,14 @@ int
 read_command()
 {
     unsigned ch;
-    int tm;
-    
-    tm = timeout > 0 ? messages_queued() ? messages_queued() : timeout : 0;
+    int tm = 0;
+
+    if(timeout > 0){
+	long dtime; 
+
+	tm = messages_queued(&dtime) ? (int)dtime : timeout;
+    }
+
     ch = read_char(tm);
     if(ch != NO_OP_COMMAND && ch != NO_OP_IDLE && ch != KEY_RESIZE)
       zero_new_mail_count();
@@ -1147,16 +1142,16 @@ static struct display_line {
 
 
 static struct key oe_keys[] =
-    {{"^G","Help",0},     {"RETURN","Enter",0},     {"^C","Cancel",0},
-     {"^T","xxx",0},      {NULL,NULL,0},            {NULL,NULL,0},
-     {NULL,NULL,0},       {NULL,NULL,0},            {NULL,NULL,0},
-     {NULL,NULL,0},       {NULL,NULL,0},            {NULL,NULL,0}};
+    {{"^G","Help",0},		{"^C","Cancel",0},	{"^T","xxx",0},
+     {"Ret","Accept",0},	{NULL,NULL,0},		{NULL,NULL,0},
+     {NULL,NULL,0},		{NULL,NULL,0},		{NULL,NULL,0},
+     {NULL,NULL,0},		{NULL,NULL,0},		{NULL,NULL,0}};
 static struct key_menu oe_keymenu =
 	{sizeof(oe_keys)/(sizeof(oe_keys[0])*12), 0, 0,0,0,0, oe_keys};
-#define OE_HELP_KEY 0
-#define OE_ENTER_KEY 1
-#define OE_CANCEL_KEY 2
-#define OE_CTRL_T_KEY 3
+#define	OE_HELP_KEY	0
+#define	OE_CANCEL_KEY	1
+#define	OE_CTRL_T_KEY	2
+#define	OE_ENTER_KEY	3
 
 
 /*---------------------------------------------------------------------- 
@@ -1180,7 +1175,7 @@ static struct key_menu oe_keymenu =
   Result:  editing input string
             returns -1 unexpected errors
             returns 0  normal entry typed (editing and return or PF2)
-            returns 1  typed ^C or PF3 (cancel)
+            returns 1  typed ^C or PF2 (cancel)
             returns 3  typed ^G or PF1 (help)
             returns 4  typed ^L for a screen redraw
 
@@ -1210,11 +1205,12 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 {
     register char *s2;
     register int   field_pos;
-    int            j, return_v, cols, ch, prompt_len, too_thin, real_y_base,
-                   cursor_moved;
+    int            i, j, return_v, cols, ch, prompt_len, too_thin,
+                   real_y_base, cursor_moved;
     char          *saved_original = NULL, *k, *kb;
     char          *kill_buffer = NULL;
-    char          **help_text;
+    char         **help_text;
+    int		   fkey_table[12];
 
     dprint(5, (debugfile, "=== optionally_enter called ===\n"));
     dprint(9, (debugfile, "string:\"%s\"  y:%d  x:%d  length: %d append: %d\n",
@@ -1234,39 +1230,60 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
           real_y_base = ps_global->ttyo->screen_rows;
     }
 
+    pause_for_current_message();
     mark_status_dirty();
+    if(append_current)			/* save a copy in case of cancel */
+      saved_original = cpystr(string);
 
-    if(append_current) {
-        /*---- save a copy in case of cancel -----*/
-	saved_original = fs_get(strlen(string) + 1);
-	strcpy(saved_original, string);
+    /*
+     * build the function key mapping table, skipping predefined keys...
+     */
+    memset(fkey_table, NO_OP_COMMAND, 12 * sizeof(int));
+    for(i = 0, j = 0; escape_list && escape_list[i].ch != -1 && i+j < 12; i++){
+	if(i+j == OE_HELP_KEY)
+	  j++;
+
+	if(i+j == OE_CANCEL_KEY)
+	  j++;
+
+	if(i+j == OE_ENTER_KEY)
+	  j++;
+
+	fkey_table[i+j] = escape_list[i].ch;
     }
 
-
 #ifdef	DOS
     help_text = (help != NO_HELP) ? get_help_text(help, NULL) : (char **)NULL;
 #else
     help_text = help;
 #endif
-    if(help_text != (char **)NULL) {
-        /*---- Show help text -----*/
-        MoveCursor(real_y_base + 1, x_base);
-        CleartoEOLN();
-        PutLine0(real_y_base+1, x_base, help_text[0]);
-        MoveCursor(real_y_base + 2, x_base);
-        CleartoEOLN();
-        if(help_text[1] != NULL) 
-          PutLine0(real_y_base + 2, x_base, help_text[1]);
+    if(help_text){			/*---- Show help text -----*/
+	int width = ps_global->ttyo->screen_cols - x_base;
+	for(j = 0; j < 2 && help_text[j]; j++){
+	    MoveCursor(real_y_base + 1 + j, x_base);
+	    CleartoEOLN();
+
+	    if(width < strlen(help_text[j])){
+		char *tmp = fs_get((width + 1) * sizeof(char));
+		strncpy(tmp, help_text[j], width);
+		tmp[width] = '\0';
+		PutLine0(real_y_base + 1 + j, x_base, tmp);
+		fs_give((void **)&tmp);
+	    }
+	    else
+	      PutLine0(real_y_base + 1 + j, x_base, help_text[j]);
+	}
+
 #ifdef	DOS
 	free_help_text(help_text);
 #endif
 
     } else {
-	BITMAP bitmap;
 	struct key_menu *km;
+	bitmap_t	 bitmap;
 
 	clrbitmap(bitmap);
-	km = &oe_keymenu;
+	clrbitmap((km = &oe_keymenu)->bitmap);		/* force formatting */
 	setbitn(OE_HELP_KEY, bitmap);
 	setbitn(OE_ENTER_KEY, bitmap);
         if(!disallow_cancel)
@@ -1274,17 +1291,24 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 	setbitn(OE_CTRL_T_KEY, bitmap);
 
         /*---- Show the usual possible keys ----*/
-	for(j=3; escape_list && j < 12; j++){
-	    if(escape_list[j-3].ch == -1)
-	      break;
+	for(i=0,j=0; escape_list && escape_list[i].ch != -1 && i+j < 12; i++){
+	    if(i+j == OE_HELP_KEY)
+	      j++;
 
-	    oe_keymenu.keys[j].label = escape_list[j-3].label;
-	    oe_keymenu.keys[j].name  = escape_list[j-3].name;
-	    setbitn(j, bitmap);
+	    if(i+j == OE_CANCEL_KEY)
+	      j++;
+
+	    if(i+j == OE_ENTER_KEY)
+	      j++;
+
+	    oe_keymenu.keys[i+j].label = escape_list[i].label;
+	    oe_keymenu.keys[i+j].name = escape_list[i].name;
+	    setbitn(i+j, bitmap);
 	}
 
-	for(; j < 12; j++)
-	  oe_keymenu.keys[j].name = NULL;
+	for(i = i+j; i < 12; i++)
+	  if(!(i == OE_HELP_KEY || i == OE_ENTER_KEY || i == OE_CANCEL_KEY))
+	    oe_keymenu.keys[i].name = NULL;
 
 	draw_keymenu(km, bitmap, cols, -2, 0, FirstMenu, 0);
     }
@@ -1321,6 +1345,10 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
     dline.vused = (int)(&string[field_pos] - string);
     line_paint(field_pos, &passwd);
 
+#ifdef	_WINDOWS
+    mswin_allowpaste(MSWIN_PASTE_LINE);
+#endif
+
     /*----------------------------------------------------------------------
       The main loop
    
@@ -1335,6 +1363,28 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 	/* Timeout 5 min to keep imap mail stream alive */
         ch = read_char(600);
 
+	/*
+	 * Don't want to intercept all characters if typing in passwd.
+	 * We select an ad hoc set that we will catch and let the rest
+	 * through.  We would have caught the set below in the big switch
+	 * but we skip the switch instead.  Still catch things like ^K,
+	 * DELETE, ^C, RETURN.
+	 */
+	if(passwd)
+	  switch(ch) {
+            case ctrl('F'):  
+	    case KEY_RIGHT:
+            case ctrl('B'):
+	    case KEY_LEFT:
+            case ctrl('U'):
+            case ctrl('A'):
+	    case KEY_HOME:
+            case ctrl('E'):
+	    case KEY_END:
+	    case TAB:
+	      goto ok_for_passwd;
+	  }
+
         if(too_thin && ch != KEY_RESIZE && ch != ctrl('Z'))
           goto bleep;
 
@@ -1359,24 +1409,13 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 	    break;
 
           /*--------------------  RETURN --------------------*/
-	  case PF2:
+	  case PF4:
 	    if(F_OFF(F_USE_FK,ps_global)) goto bleep;
 	  case ctrl('J'): 
 	  case ctrl('M'): 
 	    return_v = 0;
 	    break;
 
-#ifdef DEBUG
-          case ctrl('X'):
-            if(debug < 9)
-              goto bleep;
-            printf("field_pos: %d \n", field_pos);
-	    for(s2 = string; s2 < &string[field_len] ; s2++) 
-	      printf("%s%d --%c--\n", s2 == &string[field_pos] ?"-->" :"   ",
-                     *s2, *s2);
-            break;
-#endif
-
           /*-------------------- Destructive backspace --------------------*/
 	  case '\177': /* DEL */
 	  case ctrl('H'):
@@ -1421,11 +1460,14 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 
             /* Make string so it will fit */
             kb = cpystr(kill_buffer);
-            dprint(2, (debugfile, "Undelete: %d %d\n", strlen(string), field_len));
+            dprint(2, (debugfile,
+		       "Undelete: %d %d\n", strlen(string), field_len));
             if(strlen(kb) + strlen(string) > field_len) 
                 kb[field_len - strlen(string)] = '\0';
-            dprint(2, (debugfile, "Undelete: %d %d\n", field_len - strlen(string),
-                       strlen(kb)));
+            dprint(2, (debugfile,
+		       "Undelete: %d %d\n", field_len - strlen(string),
+		       strlen(kb)));
+                       
             if(string[field_pos] == '\0') {
                 /*--- adding to the end of the string ----*/
                 for(k = kb; *k; k++)
@@ -1447,22 +1489,16 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 	  case ctrl('C'): /* ^C */ 
 	    if(F_ON(F_USE_FK,ps_global) || disallow_cancel) goto bleep;
 	    goto cancel;
-	  case PF3:
+	  case PF2:
 	    if(F_OFF(F_USE_FK,ps_global) || disallow_cancel) goto bleep;
 	  cancel:
 	    return_v = 1;
-	    if(saved_original != NULL)
+	    if(saved_original)
 	      strcpy(string, saved_original);
+
 	    break;
 	    
 
-          case ctrl('R'):
-            /*----------- Repaint current line --------------*/
-            PutLine0(real_y_base, x_base, prompt);
-	    memset((void *)dline.dl, 0, dline.dlen);
-	    line_paint(field_pos, &passwd);
-	    break;
-
           case ctrl('A'):
 	  case KEY_HOME:
             /*-------------------- Start of line -------------*/
@@ -1515,8 +1551,8 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
                 too_thin = 1;
             } else {
 		dline.col   = x_base + prompt_len;
-		dline.dlen  = cols - x_base - prompt_len - 1;
-		fs_resize((void **)&dline.dl, (size_t)dline.dlen);
+		dline.dlen  = cols - (x_base + prompt_len + 1);
+		fs_resize((void **)&dline.dl, (size_t)dline.dlen + 1);
 		memset((void *)dline.dl, 0, (size_t)dline.dlen);
 		line_paint(field_pos, &passwd);
             }
@@ -1540,6 +1576,18 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
                 goto bleep;
             }
 
+	  case PF3 :		/* input to potentially remap */
+	  case PF5 :
+	  case PF6 :
+	  case PF7 :
+	  case PF8 :
+	  case PF9 :
+	  case PF10 :
+	  case PF11 :
+	  case PF12 :
+	      if(F_ON(F_USE_FK,ps_global)
+		 && fkey_table[ch - PF1] != NO_OP_COMMAND)
+		ch = fkey_table[ch - PF1]; /* remap function key input */
   
           default:
 	    if(escape_list){		/* in the escape key list? */
@@ -1560,6 +1608,7 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 		continue;
 	    }
 
+       ok_for_passwd:
 	    /*--- Insert a character -----*/
 	    if(dline.vused >= field_len)
 	      goto bleep;
@@ -1574,8 +1623,11 @@ optionally_enter(string, y_base, x_base, field_len, append_current, passwd,
 	}   /*---- End of switch on char ----*/
     }
 
+#ifdef	_WINDOWS
+    mswin_allowpaste(MSWIN_PASTE_DISABLE);
+#endif
     fs_give((void **)&dline.dl);
-    if(append_current) 
+    if(saved_original) 
       fs_give((void *)&saved_original);
 
     if(kill_buffer)
@@ -1689,17 +1741,14 @@ line_paint(offset, passwd)
 	  pbp--, vbp--;
     }
 
-    if(pfp != pbp || *pfp != *vfp){			/* anything to paint? */
+    if(pfp != pbp || *pfp != *vfp){			/* anything to paint?*/
 	MoveCursor(dline.row, dline.col + (int)(pfp - dline.dl));
 
-	while(pfp <= pbp){
-	    if(vfp <= vbp && *vfp)
-              Writechar(*pfp = *vfp++, 0);
-	    else
-	      Writechar(*pfp = ' ', 0);
-
-	    pfp++;
-	}
+	do
+	  Writechar((vfp <= vbp && *vfp)
+		      ? ((*pfp = *vfp++) == TAB) ? ' ' : *pfp
+		      : (*pfp = ' '), 0);
+	while(++pfp <= pbp);
     }
 
     MoveCursor(dline.row, dline.col + (offset - dline.vbase));
